﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.IO;
using System.Security.Cryptography;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM;
using LogSig.DM.DihedralGroup;
using LogSig.DM.MultiZnGroup;
using LogSig.DM.WreathProductGroup;
using LogSig.DM.ZnGroup;
using LogSig.DM.LsGen;
using LogSig.Interfaces;
using LogSig.Util;

namespace LogSig.AppModules
{
	public sealed class AppModExp : AppMod
	{
		private enum BenchmarkType
		{
			LPSolve = 0,
			CreateLPFiles,
			AbFac,
			AbFacEx
		}

		public AppModExp()
		{
			AddCommand("rc", RandomizeCanonical);
			AddCommand("gr", GenRandom);
			AddCommand("c2p", Count2PowLogSigs);
			AddCommand("count", CountLogSigs);
			AddCommand("t23s", Test23Swaps);
			AddCommand("tlsgen", TestNewLsGen);
			AddCommand("testilp", TestILP);
			AddCommand("benchmarkilp", BenchmarkILP);
			AddCommand("createlpfiles", CreateLPFiles);
			AddCommand("benchmarkabfac", BenchmarkAbFac);
			AddCommand("benchmarkabfacex", BenchmarkAbFacEx);
			AddCommand("testwr24", TestWr24);
			AddCommand("countdiffmodsets", CountDiffModSets);
			AddCommand("testmultdecid", TestMultDecid);
			AddCommand("testsizepermmeven", TestSizePermMEven);
			AddCommand("testzset", TestZSet);
			AddCommand("getzset", GetZSet);
			AddCommand("testlogsigs", TestLogSigs);
		}

		/* private static void GenHuge2PowAndFactorize(int nOrder)
		{
			while(true)
			{
				int nOrderEx = ((nOrder > 0) ? nOrder : (2 * 256 * 1024 * 1024));
				DmDihedralGroup dgHuge = new DmDihedralGroup(nOrderEx);
				var ls = dgHuge.CreateRandomSig();
				Console.WriteLine(ls.ToString());
				Console.WriteLine();
				FactorizeSomeElements(dgHuge, ls);

				Console.WriteLine("Simplified logarithmic signature:");
				var lsSimplified = dgHuge.Simplify2PowCanonical(ls);
				// Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(lsSimplified, dgHuge));
				Console.WriteLine(lsSimplified.ToString());
				Console.WriteLine();
				FactorizeSomeElements(dgHuge, lsSimplified);
			}
		}

		private static void FactorizeSomeElements(DmDihedralGroup dg,
			DmBlockSeq<DmDihedralGroupElement> ls)
		{
			Console.WriteLine("Element => rec. calls:");
			long lTotal = 0;
			for(int e = 0; e < 20; ++e)
			{
				DmDihedralGroupElement r = dg.GenerateRandomElement();
				DmTraceInfo ti = new DmTraceInfo();

				List<int> vFact = dg.FactorizePrPow(ls, 2, r, ti);

				if(ls.MultiplyElements(vFact, dg).CompareTo(r) != 0)
					Console.WriteLine("Failed to factorize: " + r.ToString());
				else
					Console.Write(((e > 0) ? ", " : string.Empty) +
						r.ToString() + " => " + ti.RecursiveCallCount);
				lTotal += ti.RecursiveCallCount;
			}
			Console.WriteLine();
			Console.WriteLine("Avg. rec. calls: " + (lTotal / 20).ToString());
			Console.WriteLine();
		} */

		private static void RandomizeCanonical(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeRandomizeCanonical), e);
		}

		internal static class AmeRandomizeCanonical
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e)
				where T : IDmGroupElement<T>
			{
				bool bTauReduce = (e.Args["taured"] != null);
				bool bBlocksOnNewLine = (e.Args["blocksonnewline"] != null);
				bool bNonCanonical = (e.Args["noncanonical"] != null);

				HashSet<string> hsWritten = new HashSet<string>();
				var ls = AppModGen.CreateLogSig(grp, e);

				while(true)
				{
					if(bNonCanonical) DmLogSigUtil.Randomize(ls, grp, e.Random);
					else DmLogSigUtil.RandomizeCanonical(ls, grp, e.Random);

					if(DmLogSigUtil.GeneratesWholeGroup(ls, grp))
					{
						string strSig = ls.ToString();
						if(hsWritten.Contains(strSig)) continue;

						Console.WriteLine(strSig);
						hsWritten.Add(strSig);

						if(bTauReduce)
						{
							ls = DmDihedralGroup.Simplify2PowCanonicalEx(ls, grp);
							Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, grp));

							string strSmp = ls.ToString();
							if(bBlocksOnNewLine) strSmp = StrUtil.BlocksOnNewLines(strSmp, 1);

							Console.WriteLine("=> " + strSmp);
							Console.WriteLine();
						}
						else if(bBlocksOnNewLine)
						{
							string strSmp = ls.ToString();
							strSmp = StrUtil.BlocksOnNewLines(strSmp, 1);
							Console.WriteLine("=> " + strSmp);
							Console.WriteLine();
						}
					}
				}
			}
		}

		/* private static void GenRandomCanonical(int nOrder, bool bEx)
		{
			DmDihedralGroup dg = new DmDihedralGroup(nOrder);
			HashSet<string> hsWritten = new HashSet<string>();
			var ls = dg.CreateDefaultSig();

			while(true)
			{
				if(!bEx) ls = dg.CreateRandomSig();
				else DmLogSigUtil.RandomizeCanonicalEx(ls, dg);

				string strSig = ls.ToString();
				if(hsWritten.Contains(strSig)) continue;

				Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));
				Console.WriteLine(strSig);
				hsWritten.Add(strSig);
			}
		} */

		/* public static void Main7(string[] args)
		{
			DmDihedralGroup dg = new DmDihedralGroup(2 * 2 * 2 * 2 * 2 * 2);
			while(true)
			{
				var ls = dg.CreateDefaultSig();
				DmLogSigUtil.RandomizeCanonicalEnsureGenerating(ls, dg, Program.CryptoRandom);

				Console.WriteLine(ls.ToString());
				Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));
				FactorizeAllElements(dg, 2, ls, false);
			}
		}

		public static void Main5(string[] args)
		{
			// DmDihedralGroup dg = new DmDihedralGroup(2 * 3 * 3 * 5 * 5 * 7 * 7);
			DmDihedralGroup dg = new DmDihedralGroup(2 * 3 * 3 * 3 * 3 * 3 * 3 * 3 * 3);
			var ls = dg.CreateDefaultSig();
			Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));

			Console.WriteLine(ls.ToString());
			FactorizeAllElements(dg, 3, ls, false);
		}

		public static void Main2(string[] args)
		{
			DmDihedralGroup dg = new DmDihedralGroup(2 * 3 * 3 * 3);
			var ls = DmBlockSeq<DmDihedralGroupElement>.CreateEmpty(dg);
			ls.Blocks[0].Elements[1] = dg.Tau;
			ls.Blocks[1].Elements[1] = new DmDihedralGroupElement(1, false);
			ls.Blocks[1].Elements[2] = new DmDihedralGroupElement(2, false);
			ls.Blocks[2].Elements[1] = new DmDihedralGroupElement(3, false);
			ls.Blocks[2].Elements[2] = new DmDihedralGroupElement(6, false);
			ls.Blocks[3].Elements[1] = new DmDihedralGroupElement(9, false);
			ls.Blocks[3].Elements[2] = new DmDihedralGroupElement(18, false);
			Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));

			DmDihedralGroupElement el = new DmDihedralGroupElement(1, true);
			// dg.TrivialFactorize(ls, el, dg.Identity, 0);
			int[] vFactorization = new int[ls.Blocks.Count];
			DmLogSigUtil.TrivialFactorize(ls, el, dg.Identity, 0, dg, vFactorization);
			dg.FactorizePrPow(ls, 3, el, new DmTraceInfo());

			FactorizeAllElements(dg, 3, ls, false);
		}

		public static void Main3(string[] args)
		{
			const int nPrime = 3;
			DmDihedralGroup dg = new DmDihedralGroup(2 * nPrime * nPrime * nPrime * nPrime);
			var ls = DmBlockSeq<DmDihedralGroupElement>.CreateEmpty(dg);

			var dictWritten = new SortedDictionary<string, bool>();
			while(true)
			{
				DmLogSigUtil.RandomizeCanonicalEnsureGenerating(ls, dg, Program.CryptoRandom);
				// dg.GenerateRandom2PowSig(ls);
				Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));

				string str = ls.ToString();
				if(dictWritten.ContainsKey(str)) continue;

				dictWritten.Add(str, true);
				Console.WriteLine(str);

				FactorizeAllElements(dg, nPrime, ls, false);
			}
		}

		public static void Main9(string[] args)
		{
			// Task t1 = new Task(() =>
			//	EnumDgLogSigs2Pow<DmDihedralGroup, DmDihedralGroupElement>(
			//		new DmDihedralGroup(1024)));
			// Task t2 = new Task(() =>
			//	EnumDgLogSigs2Pow<DmDihedralGroup, DmDihedralGroupElement>(
			//		new DmDihedralGroup(1024)));
			// t1.Start();
			// t2.RunSynchronously();

			// EnumDgLogSigs2Pow<DmDihedralGroup, DmDihedralGroupElement>(
			//		new DmDihedralGroup(512));

			DmDihedralGroup dg = new DmDihedralGroup(512);

			DmBlockSeq<DmDihedralGroupElement> ls = new DmBlockSeq<DmDihedralGroupElement>();
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(109, true));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 2 * 2 * 3 * 5, false));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 3 * 11, true));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 3 * 13, false));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 2 * 2 * 2 * 2, false));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 2 * 2 * 2, false));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 3 * 11, true));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 97, true));
			ls.AddBlock(dg.Identity, new DmDihedralGroupElement(2 * 2 * 2 * 3 * 3 * 3, false));
			// Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dg));
			// dg.TrivialFactorize(ls, new DmDihedralGroupElement(42, false),
			//	dg.Identity, 0);

			// Console.WriteLine(ls.ToString());
			// FactorizeAllElements(dg, 2, ls, true);

			// List<int> v = dg.Factorize2Pow(ls, new DmDihedralGroupElement(42, false));
			// StringBuilder sb = new StringBuilder();
			// for(int i = 0; i < v.Count; ++i)
			// {
			//	if(i > 0) sb.Append("*");
			//	sb.Append(v[i].ToString());
			// }
			// Console.WriteLine(sb.ToString());

			DmDihedralGroup dg2 = new DmDihedralGroup(256);
			EnumDgLogSigs2Pow(dg2);
		} */

		/* private static void EnumDgLogSigs2Pow(DmDihedralGroup dmGrp)
		{
			var blk = new DmLogSigBlock<DmDihedralGroupElement>();
			blk.Elements.Add(dmGrp.Identity);
			blk.Elements.Add(dmGrp.Identity);

			var ls = new DmBlockSeq<DmDihedralGroupElement>();
			for(int iCnt = 1; iCnt < dmGrp.Order; iCnt <<= 1)
				ls.Blocks.Add(blk.Clone());

			SortedDictionary<string, bool> dictWritten = new SortedDictionary<string, bool>();
			while(true)
			{
				DmLogSigUtil.RandomizeCanonicalEnsureGenerating(ls, dmGrp, Program.CryptoRandom);

				string strCanonical = ls.ToString();
				if(dictWritten.ContainsKey(strCanonical)) continue;

				Console.WriteLine(strCanonical);
				dictWritten.Add(strCanonical, true);

				Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, dmGrp));
				FactorizeAllElements(dmGrp, 2, ls, true);
			}
		} */

		/* private static void FactorizeAllElements(DmDihedralGroup dg,
			int nPrime, DmBlockSeq<DmDihedralGroupElement> ls, bool bExtend)
		{
			SortedDictionary<int, int> dictCalls = new SortedDictionary<int, int>();

			List<int> v;
			for(int r = 0; r < dg.Order / 2; ++r)
			{
				DmTraceInfo traceInfo = new DmTraceInfo();

				DmDihedralGroupElement d = new DmDihedralGroupElement(r, false);
				v = dg.FactorizePrPow(ls, nPrime, d, traceInfo);
				if(v == null)
				{
					// dg.TrivialFactorize(ls, d, dg.Identity, 0);
					int[] vFactorization = new int[ls.Blocks.Count];
					DmLogSigUtil.TrivialFactorize(ls, d, dg.Identity, 0, dg, vFactorization);

					dg.FactorizePrPow(ls, nPrime, d, new DmTraceInfo());
					return;
				}
				if(ls.MultiplyElements(v, dg).CompareTo(d) != 0)
					throw new InvalidOperationException();

				if(!dictCalls.ContainsKey(traceInfo.RecursiveCallCount))
					dictCalls.Add(traceInfo.RecursiveCallCount, 1);
				else
					dictCalls[traceInfo.RecursiveCallCount] =
						dictCalls[traceInfo.RecursiveCallCount] + 1;

				traceInfo = new DmTraceInfo();

				DmDihedralGroupElement s = new DmDihedralGroupElement(r, true);
				v = dg.FactorizePrPow(ls, nPrime, s, traceInfo);
				if(ls.MultiplyElements(v, dg).CompareTo(s) != 0)
					throw new InvalidOperationException();

				if(!dictCalls.ContainsKey(traceInfo.RecursiveCallCount))
					dictCalls.Add(traceInfo.RecursiveCallCount, 1);
				else
					dictCalls[traceInfo.RecursiveCallCount] =
						dictCalls[traceInfo.RecursiveCallCount] + 1;
			}

			long lTotalCalls = 0;
			Console.Write("Recursive call counts:");
			foreach(KeyValuePair<int, int> kvp in dictCalls)
			{
				lTotalCalls += kvp.Key * kvp.Value;
				Console.Write(" " + kvp.Key.ToString() + ": " +
					kvp.Value.ToString() + "x");
			}
			Console.WriteLine();
			Console.WriteLine("Total rec. calls: " + lTotalCalls.ToString());
			Console.WriteLine("Avg. rec. calls per element: " +
				(lTotalCalls / dg.Order).ToString());
			Console.WriteLine();

			if(bExtend)
			{
				// const int nNewBlocks = 5;
				// DmDihedralGroup dgExt = new DmDihedralGroup(dg.Order * (1 << nNewBlocks));
				// DmBlockSeq<DmDihedralGroupElement> lsExt = ls.Clone();
				// int nRot = dg.Order / 2;
				// for(int i = 0; i < nNewBlocks; ++i)
				// {
				//	lsExt.AddBlock(dgExt.Identity, new DmDihedralGroupElement(
				//		nRot, false));
				//	nRot *= 2;
				// }
				// Console.WriteLine(lsExt.ToString());
				// FactorizeAllElements(dgExt, nPrime, lsExt, false);

				// var lsExt = ls.Clone();
				// var ls2 = DmBlockSeq<DmDihedralGroupElement>.CreateEmpty(dg);
				// DmLogSigUtil.RandomizeCanonicalEnsureGenerating(ls2, dg);
				// Console.WriteLine("+ " + ls2.ToString());
				// DmDihedralGroup dg2 = new DmDihedralGroup(dg.Order * dg.Order);
				// DmDihedralGroupElement elTrans = new DmDihedralGroupElement(dg.Order / 2, false);
				// for(int x = 0; x < ls2.Blocks.Count; ++x)
				// {
				//	List<DmDihedralGroupElement> vElem = ls2.Blocks[x].Elements;
				//	for(int y = 1; y < vElem.Count; ++y)
				//		vElem[y] = new DmDihedralGroupElement(elTrans.Rotation *
				//			vElem[y].Rotation, vElem[y].Reflection);
				//	lsExt.AddBlock(vElem.ToArray());
				// }

				DmDihedralGroup dg2 = new DmDihedralGroup(dg.Order * dg.Order);
				var lsCan = dg2.CreateDefaultSig();
				var lsExt = ls.Clone();
				for(int i = ls.Blocks.Count; i < lsCan.Blocks.Count; ++i)
					lsExt.Blocks.Add(lsCan.Blocks[i]);

				Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(lsExt, dg2));
				FactorizeAllElements(dg2, nPrime, lsExt, false);
			}
		} */

		private static void Count2PowLogSigs(object sender, AppModEventArgs e)
		{
			BigInteger nOrder = e.Args.GetBigInt("n", BigInteger.Zero);
			if(nOrder <= 0) return;
			if(nOrder > new BigInteger(long.MaxValue)) return;

			DmDihedralGroup dg = new DmDihedralGroup(nOrder);
			var ls = dg.CreateDefaultSig();
			HashSet<string> hsWritten = new HashSet<string>();

			// double dblExpected = 1.0;
			// dblExpected *= Math.Pow(2.0, NumUtil.GetPow(nOrder, 2) - 2);
			// for(int i = 0; i < NumUtil.GetPow(nOrder, 2) - 1; ++i)
			//	dblExpected *= Math.Pow(2.0, i); // Classes
			// for(int f = 1; f < NumUtil.GetPow(nOrder, 2); ++f)
			//	dblExpected *= f; // Permutations
			// dblExpected *= (long)nOrder; // Right translation
			// Console.WriteLine("Expected count: " + dblExpected.ToString());

			int n = NumUtil.GetPow(nOrder, 2); // Number of blocks
			double dblExpected = 1.0;
			dblExpected *= Math.Pow(2.0, n - 1); // Translation of tau
			for(int f = 1; f < n; ++f)
				dblExpected *= f; // Permutations
			dblExpected *= Math.Pow(2.0, ((n - 2) * (n - 1)) / 2); // Classes
			dblExpected *= (Math.Pow(2.0, n) - 1.0);
			Console.WriteLine("Expected count: " + dblExpected.ToString());

			double dbl = Math.Pow(Math.Pow(2.0, n), n);
			Console.WriteLine("Pseudo-log. sig.: " + dbl.ToString());

			int iTicks = Environment.TickCount;
			while(true)
			{
				DmLogSigUtil.RandomizeCanonical(ls, dg, e.Random);
				if(DmLogSigUtil.GeneratesWholeGroup(ls, dg))
				{
					if((Environment.TickCount - iTicks) >= 5000)
					{
						Console.Write(hsWritten.Count);
						Console.Write(" ");
						iTicks = Environment.TickCount;
					}

					// ls = dg.Simplify2PowCanonical(ls);
					string str = ls.ToString();
					if(hsWritten.Contains(str)) continue;

					hsWritten.Add(str);
					// Console.WriteLine(str);
				}
			}
		}

		private static void CountLogSigs(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeCountCanonicalSigs), e);
		}

		internal static class AmeCountCanonicalSigs
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e)
				where T : IDmGroupElement<T>
			{
				var ls = grp.CreateDefaultSig();
				DmLogSigUtil.ShuffleList(ls.Blocks, e.Random);
				Console.WriteLine(ls.ToString());
				Console.WriteLine();

				BigInteger bi = BigInteger.One;
				for(int b = 0; b < ls.Blocks.Count; ++b)
					bi *= BigInteger.Pow(grp.Order, ls.Blocks[b].Elements.Count - 1);
				Console.WriteLine("Pseudo-log. sig. of above type: " + ((double)(bi)).ToString());
				Console.WriteLine();

				HashSet<string> hsWritten = new HashSet<string>();

				int iTicks = Environment.TickCount;
				FastRandomGen rand = new FastRandomGen();
				while(true)
				{
					DmLogSigUtil.RandomizeCanonical(ls, grp, rand);
					if(DmLogSigUtil.GeneratesWholeGroup(ls, grp))
					{
						if((Environment.TickCount - iTicks) >= 5000)
						{
							Console.Write(hsWritten.Count);
							Console.Write(" ");
							iTicks = Environment.TickCount;
						}

						string str = ls.ToString();
						if(hsWritten.Contains(str)) continue;

						hsWritten.Add(str);
					}
				}
			}
		}

		// private static object m_objLock = new object();
		// private static void WriteLine(string str)
		// {
		//	lock(m_objLock) { Console.WriteLine(str); }
		// }

		private static void GenRandom(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeGenRandom), e);
		}

		internal static class AmeGenRandom
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e)
				where T : IDmGroupElement<T>
			{
				BigInteger biRounds = BigInteger.Abs(e.Args.GetBigInt("r", 5000));
				if(biRounds > new BigInteger(ulong.MaxValue)) biRounds = ulong.MaxValue;
				ulong uRounds = (ulong)biRounds;

				HashSet<string> hsWritten = new HashSet<string>();
				while(true)
				{
					// T tLeft, tRight;
					// var ls = DmLogSigGen.Generate(grp, uRounds, e.Random,
					//	out tLeft, out tRight, null);

					string strRels = null;
					List<int> lFac;
					var ls = LsGen<T>.Generate(grp, uRounds, e.Random,
						default(T), ref strRels, out lFac);

					string str = ls.ToString();
					if(hsWritten.Contains(str)) continue;

					Console.WriteLine(str);
					hsWritten.Add(str);
				}
			}
		}

		private static void Test23Swaps(object sender, AppModEventArgs e)
		{
			DmDihedralGroup dg = new DmDihedralGroup(2 * 3 * 3 * 3);

			long n = 0;
			while(true)
			{
				// DmDihedralGroupElement tLeft, tRight;
				string strTrfIORel = string.Empty;
				List<int> lFac;

				// var ls = DmLogSigGen.Generate(dg, 9, e.Random,
				//	out tLeft, out tRight, null);
				var ls = LsGen<DmDihedralGroupElement>.Generate(dg, 9,
					e.Random, null, ref strTrfIORel, out lFac);

				if(ls.Blocks[3].Elements.Count != 2) continue;
				Console.WriteLine(n.ToString() + ": " + ls.ToString());

				byte[] pbKey = new byte[32];
				byte[] pbCtr = new byte[16];
				AesCryptoRandomGen rand = new AesCryptoRandomGen(pbKey, pbCtr);
				var lsSub = ls.GetSubSig(2, 2);

				while(true)
				{
					var lsNew = lsSub.GetRandomEquivalentCanonical(dg, rand);
					if((lsNew == null) || (lsNew == lsSub)) continue;
					if(lsNew.Blocks[0].Elements.Count != 2) continue;

					Console.WriteLine(lsNew.ToString());
					Console.WriteLine();
					break;
				}

				++n;
			}
		}

		private static void TestNewLsGen(object sender, AppModEventArgs e)
		{
			DmDihedralGroup dg = new DmDihedralGroup(2 * 3 * 3 * 3 * 3);

			for(int r = 0; r < 100; ++r)
			{
				ulong uRounds = (ulong)Program.CryptoRandom.GenerateInt32(300);
				byte[] pbRandKey = new byte[32];
				byte[] pbRandCtr = new byte[16];
				Program.CryptoRandom.Generate(pbRandKey, 0, pbRandKey.Length);

				CryptoRandomGen rand = new AesCryptoRandomGen(pbRandKey, pbRandCtr);
				string strRels = string.Empty;
				List<int> lFac;

				// Get transformation I/O rels
				var ls = LsGen<DmDihedralGroupElement>.Generate(dg, uRounds, rand,
					null, ref strRels, out lFac);

				rand = new AesCryptoRandomGen(pbRandKey, pbRandCtr); // Reset CSPRNG
				DmDihedralGroupElement el = new DmDihedralGroupElement(
					new BigInteger(r % ((int)dg.Order / 2)), true);
				ls = LsGen<DmDihedralGroupElement>.Generate(dg, uRounds, rand,
					el, ref strRels, out lFac);
				if(!DmLogSigUtil.GeneratesWholeGroup(ls, dg))
					throw new InvalidOperationException();

				if(!ls.MultiplyElements(lFac, dg).Equals(el))
				{
					Debug.Assert(false);
					throw new InvalidOperationException();
				}
				else Console.Write(".");
			}
		}

		private static void TestILP(object sender, AppModEventArgs e)
		{
			//DmMultiZnGroup grp = new DmMultiZnGroup(new BigInteger[] {
			//	65536, 65536, 65536, 65536 });
			DmMultiZnGroup grp = new DmMultiZnGroup(new BigInteger[] {
				2, 3, 3, 5 });

			// DmMultiZnGroupElement tLeft, tRight;
			string strTrfIORel = string.Empty;
			List<int> lFac;

			// var ls = DmLogSigGen.Generate(grp, 1000, e.Random, out tLeft,
			//	out tRight);
			var ls = LsGen<DmMultiZnGroupElement>.Generate(grp, 1000, e.Random,
				null, ref strTrfIORel, out lFac);

			Console.WriteLine(ls.ToString());
			Console.WriteLine();

			for(int r = 0; r < 10; ++r)
			{
				var el = grp.GenerateRandomElement(e.Random);
				int[] v = DmMultiZnGroup.FactorByILP(ls, grp, el);
				var c = ls.MultiplyElements(new List<int>(v), grp);
				Console.WriteLine(el.ToString() + ": " + (el.Equals(c) ?
					"Success" : "Failed") + ".");
			}
		}

		private static void BenchmarkILP(object sender, AppModEventArgs e)
		{
			Benchmark(e, 1, 12, false, BenchmarkType.LPSolve); // Warm-up

			int iOffset = e.Args.GetInt("i", 1);
			Benchmark(e, iOffset, int.MaxValue, true, BenchmarkType.LPSolve);
		}

		private static void CreateLPFiles(object sender, AppModEventArgs e)
		{
			int iOffset = e.Args.GetInt("i", 1);
			int iMax = e.Args.GetInt("m", 20);
			Benchmark(e, iOffset, iMax, true, BenchmarkType.CreateLPFiles);
		}

		private static void BenchmarkAbFac(object sender, AppModEventArgs e)
		{
			Benchmark(e, 1, 12, false, BenchmarkType.AbFac); // Warm-up

			int iOffset = e.Args.GetInt("i", 1);
			int iMax = e.Args.GetInt("m", 20);
			Benchmark(e, iOffset, iMax, true, BenchmarkType.AbFac);
		}

		private static void BenchmarkAbFacEx(object sender, AppModEventArgs e)
		{
			Benchmark(e, 1, 12, false, BenchmarkType.AbFacEx); // Warm-up

			int iOffset = e.Args.GetInt("i", 1);
			int iMax = e.Args.GetInt("m", 20);
			Benchmark(e, iOffset, iMax, true, BenchmarkType.AbFacEx);
		}

		private static void Benchmark(AppModEventArgs e, int iOffset,
			int iMaxExcl, bool bPrint, BenchmarkType t)
		{
			string strLogFile = e.Args["logfile"];
			UTF8Encoding utf8 = new UTF8Encoding(false);
			Stopwatch sw = new Stopwatch();

			for(int n = iOffset; n < iMaxExcl; ++n)
			{
				BigInteger[] vMods = new BigInteger[n];
				for(int i = 0; i < n; ++i) vMods[i] = 2;
				DmMultiZnGroup grp = new DmMultiZnGroup(vMods);

				// var ls = grp.CreateDefaultSig();
				// DmMultiZnGroupElement tLeft, tRight;
				// var ls = DmLogSigGen.Generate(grp, 1000, e.Random, out tLeft,
				//	out tRight);
				var ls = grp.CreateDefaultSig();
				
				// uint r = 2463534242U;
				RijndaelManaged aes = new RijndaelManaged();
				aes.BlockSize = 128;
				aes.KeySize = 256;
				byte[] pbKey = new byte[32];
				aes.Key = pbKey;
				aes.Mode = CipherMode.ECB;
				ICryptoTransform tAes = aes.CreateEncryptor();
				byte[] pbPT = new byte[16];
				byte[] pbCT = new byte[17]; // 16 + 1 (for sign byte)

				for(int i = 0; i < (n * n); ++i)
				{
					// r ^= (r << 13);
					// r ^= (r >> 17);
					// r ^= (r << 5);
					byte[] pbI = MemUtil.Int64ToBytes(i);
					Array.Clear(pbPT, 0, pbPT.Length);
					Array.Copy(pbI, pbPT, pbI.Length);
					if(tAes.TransformBlock(pbPT, 0, pbPT.Length, pbCT, 0) != 16)
						throw new InvalidOperationException("AES encryption failed!");

#if DEBUG
					if(i == 0xC0) // From official test vectors, ECBVarTxt256.rsp
					{
						byte[] pbExpc = new byte[16] {
							0x0A, 0x6B, 0xDC, 0x6D, 0x4C, 0x1E, 0x62, 0x80,
							0x30, 0x1F, 0xD8, 0xE9, 0x7D, 0xDB, 0xE6, 0x01 };
						for(int j = 0; j < 16; ++j)
						{
							Debug.Assert(pbCT[j] == pbExpc[j]);
						}
					}
#endif

					int x = (i % n);
					// int y = (int)(r % (uint)n);
					// int y = (int)(MemUtil.BytesToUInt64(pbCT, 0) % (ulong)n);
					Debug.Assert((pbCT.Length >= 17) && (pbCT[16] == 0));
					BigInteger biV = new BigInteger(pbCT);
					Debug.Assert(biV.Sign >= 0);
					int y = (int)BigInteger.Remainder(biV, new BigInteger(n));

					if(x != y)
						ls.Blocks[x].Elements[1] = grp.Multiply(
							ls.Blocks[x].Elements[1], ls.Blocks[y].Elements[1]);
				}

				tAes.Dispose();
				aes.Dispose();

				BigInteger[] vEl = new BigInteger[n];
				for(int i = 0; i < n; i += 2) vEl[i] = 1;
				DmMultiZnGroupElement el = new DmMultiZnGroupElement(vEl);

				StringBuilder sb = new StringBuilder();
				sb.Append((@"2^" + n.ToString()).PadLeft(5) + ": ");

				if(t == BenchmarkType.CreateLPFiles)
				{
					int nLsLen = ls.Blocks.Sum(blk => blk.Elements.Count);
					int nCmp = grp.Mods.Length;
					int nVars = nLsLen + nCmp;
					LpFileGenerator lp = new LpFileGenerator(nVars);

					string strBase = "ILP_" + n.ToString();

					sb.Clear();
					sb.AppendLine("Group: " + grp.ToString());
					sb.AppendLine("Order: 2^" + n.ToString());
					sb.AppendLine();
					sb.AppendLine(ls.ToString());
					sb.AppendLine();
					sb.AppendLine("Element to factor: " + el.ToString());
					File.WriteAllText(strBase + ".txt", sb.ToString(), utf8);

					DmMultiZnGroup.FactorByILP(ls, grp, el, lp);
					File.WriteAllText(strBase + ".lp", lp.BuildLP(), utf8);
					continue;
				}

				int[] v = new int[ls.Blocks.Count];

				sw.Reset();
				sw.Start();

				if(t == BenchmarkType.LPSolve)
					v = DmMultiZnGroup.FactorByILP(ls, grp, el);
				else if(t == BenchmarkType.AbFac)
				{
					if(!DmMultiZnGroup.Factorize(ls, grp, el, v)) { Debug.Assert(false); }
				}
				else if(t == BenchmarkType.AbFacEx)
				{
					List<int[]> lFacs = DmMultiZnGroup.FactorizeAbEx(ls, grp, el,
						new DmTraceInfo());
					if((lFacs != null) && (lFacs.Count == 1) && (lFacs[0] != null) &&
						(lFacs[0].Length == v.Length))
						Array.Copy(lFacs[0], v, v.Length);
					else { Debug.Assert(false); }
				}
				else { Debug.Assert(false); }

				sw.Stop();

				var c = ls.MultiplyElements(new List<int>(v), grp);
				if(!el.Equals(c))
					sb.AppendLine("Factorization failed!");
				else
				{
					double d = (double)sw.ElapsedMilliseconds;
					d /= 1000.0;
					sb.AppendLine(d.ToString() + " s.");
				}

				if(bPrint)
				{
					if(string.IsNullOrEmpty(strLogFile))
						Console.Write(sb.ToString());
					else File.AppendAllText(strLogFile, sb.ToString(), utf8);
				}
			}
		}

		private static void TestWr24(object sender, AppModEventArgs e)
		{
			var z2 = new DmZnGroup(2);
			var z4 = new DmZnGroup(4);
			var grp = new DmWreathProductGroup<DmZnGroupElement,
				DmZnGroupElement>(z2, z4);
			var z = new DmZnGroupElement(0);
			var o = new DmZnGroupElement(1);

			var gl = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, o, z, z }, o);
			var gr = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, z, o, o },
				new DmZnGroupElement(2));
			var g = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, o, z, z },
				new DmZnGroupElement(3));

			var gli = grp.Invert(gl);
			var gri = grp.Invert(gr);
			var p = grp.Multiply(grp.Multiply(gli, g), gri);

			var g1 = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, o, z, z },
				new DmZnGroupElement(1));
			var g2 = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { z, o, o, o },
				new DmZnGroupElement(0));
			var g3 = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, o, z, z },
				new DmZnGroupElement(0));
			var g4 = new DmWreathProductGroupElement<DmZnGroupElement,
				DmZnGroupElement>(new DmZnGroupElement[] { o, z, o, o },
				new DmZnGroupElement(2));
			var r = grp.Multiply(grp.Multiply(grp.Multiply(g1, g2), g3), g4);

			Console.WriteLine(p);
			Console.WriteLine(r);
		}

		private static void CountDiffModSets(object sender, AppModEventArgs e)
		{
			int n = e.Args.GetInt("n", 0);
			int m = e.Args.GetInt("m", 0);
			if(n == 0) throw new ArgumentOutOfRangeException("n");
			if(m == 0) throw new ArgumentOutOfRangeException("m");
			if(n < m) throw new ArgumentOutOfRangeException("n < m");

			KeyValuePair<int, int> kvp = NumUtil.CountDiffModSets(n, m);
			Console.WriteLine("d(" + n.ToString() + ", " + m.ToString() +
				") = " + kvp.Key.ToString() + ", d^(" + n.ToString() + ", " +
				m.ToString() + ") = " + kvp.Value.ToString() + ".");
		}

		private static void TestMultDecid(object sender, AppModEventArgs e)
		{
			const int nRotMod = 21;
			const int nRotFac = 21 / 7;
			DmDihedralGroup grp = new DmDihedralGroup(2 * nRotMod);

			var l = new List<DmDihedralGroupElement>();
			for(int i = 0; i < nRotMod; i += nRotFac)
				l.Add(new DmDihedralGroupElement(i, false));

			while(true)
			{
				DmDihedralGroupElement t = new DmDihedralGroupElement(
					e.Random.GenerateInt32(nRotMod), true);

				var lm = new List<DmDihedralGroupElement>(l);
				for(int y = 1; y < lm.Count; ++y)
				{
					if(e.Random.GenerateBool())
						lm[y] = grp.Multiply(t, lm[y]);
				}
				if(lm.Intersect(l).Count() == l.Count) continue;

				var lPossible = new List<DmDihedralGroupElement>();
				for(int r = 0; r < nRotMod; ++r)
				{
					var tr = new DmDihedralGroupElement(r, true);

					var lr = lm.Select(el => (el.Reflection ? grp.Multiply(tr, el) :
						el)).Distinct().ToList();
					if((lr.Count == l.Count) && (lr.Intersect(l).Count() == l.Count))
						lPossible.Add(tr);
				}

				Debug.Assert(lPossible.Count == 1);
			}
		}

		private static void TestSizePermMEven(object sender, AppModEventArgs e)
		{
			while(true)
			{
				int nRot = e.Random.GenerateInt32(28) + 4;
				DmDihedralGroup grp = new DmDihedralGroup(2 * nRot);
				int nA = e.Random.GenerateInt32(nRot - 3) + 3;

				List<DmDihedralGroupElement> lT = new List<DmDihedralGroupElement>();
				for(int i = 0; i < 2; ++i) lT.Add(grp.Identity);
				List<DmDihedralGroupElement> lA = new List<DmDihedralGroupElement>();
				for(int i = 0; i < nA; ++i) lA.Add(grp.Identity);
				List<DmDihedralGroupElement> lTN = new List<DmDihedralGroupElement>();
				for(int i = 0; i < 2; ++i) lTN.Add(grp.Identity);
				List<DmDihedralGroupElement> lAN = new List<DmDihedralGroupElement>();
				for(int i = 0; i < nA; ++i) lAN.Add(grp.Identity);

				lT[1] = new DmDihedralGroupElement(e.Random.GenerateBigInt(
					grp.Order / 2), true);

				for(int i = 1; i < nA; ++i)
					lA[i] = grp.GenerateRandomElement(e.Random);

				List<DmDihedralGroupElement> lProd =
					DmBlockSeq<DmDihedralGroupElement>.Multiply(grp, lT, lA);
				Debug.Assert(lProd.Count == (lT.Count * lA.Count));

				lTN[1] = lProd[e.Random.GenerateInt32(lProd.Count)];
				for(int i = 1; i < nA; ++i)
					lAN[i] = lProd[e.Random.GenerateInt32(lProd.Count)];

				List<DmDihedralGroupElement> lProdN =
					DmBlockSeq<DmDihedralGroupElement>.Multiply(grp, lAN, lTN);

				if(lProd.Intersect(lProdN).Count() != lProd.Count)
					continue;

				List<DmDihedralGroupElement> lR = new List<DmDihedralGroupElement>();
				for(int i = 0; i < nA; ++i)
				{
					if(!lA[i].Reflection) lR.Add(lA[i]);
					else lR.Add(grp.Multiply(lT[1], lA[i]));
				}

				List<BigInteger> lZ = grp.GetZRotOnly(lR);
				if(lZ.Count == 0)
					break;
			}
		}

		private static void TestZSet(object sender, AppModEventArgs e)
		{
			while(true)
			{
				int nRot = e.Random.GenerateInt32(16) + 8;
				DmDihedralGroup grp = new DmDihedralGroup(2 * nRot);

				int nR = (int)e.Random.GenerateBigInt(nRot / 2 - 3) + 3;

				List<DmDihedralGroupElement> l = new List<DmDihedralGroupElement>();
				while(l.Count < nR)
				{
					var elGen = grp.GenerateRandomElement(e.Random);
					if(!l.Contains(elGen)) l.Add(elGen);
				}

				var lZ = DmLogSigUtil.GetZ(l, grp);
				if(lZ.Count == 0) continue;

				var lZHat = DmLogSigUtil.GetZHat(l, grp);
				Debug.Assert(lZHat.Intersect(l).Count() == lZHat.Count);

				var lZHatInv = lZHat.Select(el => grp.Invert(el)).ToList();
				var lProd = DmBlockSeq<DmDihedralGroupElement>.Multiply(
					grp, lZHat, lZHatInv).Distinct().ToList();
				if(lProd.Intersect(lZHat).Count() != lZHat.Count)
					Console.WriteLine("No subgroup.");
			}
		}

		private static void GetZSet(object sender, AppModEventArgs e)
		{
			BigInteger n = e.Args.GetBigInt("n", 4);
			DmZnGroup grp = new DmZnGroup(n);
			var blk = DmLogSigBlock<DmZnGroupElement>.Parse(e.Args["blk"], grp);

			var lZ = DmLogSigUtil.GetZ(blk.Elements, grp);

			var blkZ = new DmLogSigBlock<DmZnGroupElement>();
			blkZ.Elements.AddRange(lZ);
			Console.WriteLine(blkZ.ToString());
		}

		private static void TestLogSigs(object sender, AppModEventArgs e)
		{
			string[] vD30 = new string[] {
				"((id, 4t), (id, 9t, 14t), (id, 8, 2t, 3t, 5t))",
				"((id, 5t), (id, 0t, 10t), (id, 8, 3t, 11t, 14t))",
				"((id, 9t), (id, 5, 10), (id, 3, 11, 2t, 10t))",
				"((id, 14t), (id, 5, 4t), (id, 8, 9, 12t, 13t))"
			};
			DmDihedralGroup dg30 = new DmDihedralGroup(30);
			foreach(string str in vD30)
			{
				var ls = DmBlockSeq<DmDihedralGroupElement>.Parse(str, dg30);
				if(!DmLogSigUtil.GeneratesWholeGroup(ls, dg30))
				{
					Debug.Assert(false);
					throw new InvalidOperationException();
				}
			}

			string strD54 = "((7t, 15, 13), (4t, 22t, 10), (2, 1t), (2, 14, 9t))";
			DmDihedralGroup dg54 = new DmDihedralGroup(54);
			var ls54 = DmBlockSeq<DmDihedralGroupElement>.Parse(strD54, dg54);
			if(!DmLogSigUtil.GeneratesWholeGroup(ls54, dg54))
			{
				Debug.Assert(false);
				throw new InvalidOperationException();
			}
		}
	}
}
