﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM;
using LogSig.DM.DihedralGroup;
using LogSig.DM.MultiZnGroup;
using LogSig.DM.LsGen;
using LogSig.DM.ZnGroup;
using LogSig.Util;

namespace LogSig.AppModules
{
	public sealed class AppModFac : AppMod
	{
		internal enum FacMethod
		{
			Periodic = 0,
			Generic = 1
		}

		public AppModFac()
		{
			AddCommand("abfac", AbFac);
			AddCommand("abfacex", AbFacEx);
		}

		private static void AbFac(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeAbFac), e,
				FacMethod.Periodic);
		}

		private static void AbFacEx(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeAbFac), e,
				FacMethod.Generic);
		}

		internal static class AmeAbFac
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e,
				FacMethod fm)
				where T : IDmGroupElement<T>
			{
				DmMultiZnGroup mg = (grp as DmMultiZnGroup);
				DmZnGroup zg = (grp as DmZnGroup);
				if(mg == null)
				{
					if(zg != null)
						mg = new DmMultiZnGroup(new BigInteger[] { zg.Order });
					else
					{
						Debug.Assert(false);
						return;
					}
				}

				string strLS = e.Args["ls"];
				if(strLS != null)
				{
					DmBlockSeq<DmMultiZnGroupElement> ls;
					if(zg != null)
					{
						var lsZ = DmBlockSeq<DmZnGroupElement>.Parse(strLS, zg);
						ls = DmBlockSeq<DmMultiZnGroupElement>.ConvertFrom(lsZ,
							elZ => new DmMultiZnGroupElement(new BigInteger[] {
								elZ.Value }));
					}
					else ls = DmBlockSeq<DmMultiZnGroupElement>.Parse(strLS, mg);

					// if(mg.Order <= 1000)
					// {
					//	if(!DmLogSigUtil.GeneratesWholeGroup(ls, mg))
					//	{
					//		Program.TraceLine("The input is not a logarithmic signature!");
					//		return;
					//	}
					// }

					Factorize(mg, e, ls, fm);
				}
				else Generate(mg, e);
			}

			private static void TraceIntro(DmBlockSeq<DmMultiZnGroupElement> ls,
				DmMultiZnGroupElement el, DmMultiZnGroup mg)
			{
				Program.TraceLine("% " + StrUtil.TrimEx(Environment.CommandLine));
				Program.TraceLine("Let $G = " + mg.Name + "$,");
				Program.TraceLine("$g = " + el.ToString() + "$, and");
				Program.TraceLine("\\begin{center} \\footnotesize");
				Program.TraceLine(ls, "$\\alpha = ", "$");
				Program.TraceLine("\\end{center}");
				Program.TraceLine("\\noindent");
				Program.TraceLine("a logarithmic signature for $G$.");
				Program.TraceLine("We are looking for a factorization of $g$ with respect to $\\alpha$.");
			}

			private static void Generate(DmMultiZnGroup mg, AppModEventArgs e)
			{
				var ls = mg.CreateDefaultSig();
				if(mg.Order <= 2048)
				{
					Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, mg));
				}

				BigInteger biRounds = BigInteger.Abs(e.Args.GetBigInt("r", 500));
				if(biRounds > new BigInteger(ulong.MaxValue)) biRounds = ulong.MaxValue;
				ulong uRounds = (ulong)biRounds;

				// DmLogSigGenOptions opt = new DmLogSigGenOptions();
				// opt.FuseInMiddle = (e.Args["fuse"] != null);

				string strType = e.Args["bstype"];
				if(!string.IsNullOrEmpty(strType))
				{
					ls = AppModGen.CreateEmptySig(strType, mg);

					if(ls.Blocks.Aggregate(BigInteger.One, (acc, blk) =>
						acc * blk.Elements.Count) != mg.Order)
						Program.TraceLine("Logarithmic signature size invalid!");
				}
				bool bCompletelyRandom = !string.IsNullOrEmpty(strType);

				while(true)
				{
					if(bCompletelyRandom)
					{
						do { DmLogSigUtil.RandomizeCanonical(ls, mg, e.Random); }
						while(!DmLogSigUtil.GeneratesWholeGroup(ls, mg));
						// DmLogSigUtil.RandomizeCanonicalEnsureGenerating(ls, mg, e.Random);
					}
					else
					{
						// DmMultiZnGroupElement tLeft, tRight;
						string strTrfIORels = string.Empty;
						List<int> lFac;

						// ls = DmLogSigGen.Generate(mg, uRounds, e.Random, out tLeft,
						//	out tRight, null, opt);
						ls = LsGen<DmMultiZnGroupElement>.Generate(mg, uRounds,
							e.Random, null, ref strTrfIORels, out lFac);
					}

					if(!Program.Verbose) Program.TraceLine(ls.ToString());

					for(int t = 0; t < 10; ++t)
					{
						DmMultiZnGroupElement elRand = mg.GenerateRandomElement(e.Random);
						int[] vFac = new int[ls.Blocks.Count];

						if(Program.Verbose) TraceIntro(ls, elRand, mg);

						if(!DmMultiZnGroup.Factorize(ls, mg, elRand, vFac))
							Program.TraceLine("Failed to find factorization!");
						else
						{
							Debug.Assert(ls.MultiplyElements(new List<int>(vFac),
								mg).Equals(elRand));
						}
					}

					Console.WriteLine();
					if(e.Args["once"] != null) break;
				}
			}

			private static void Factorize(DmMultiZnGroup mg, AppModEventArgs e,
				DmBlockSeq<DmMultiZnGroupElement> ls, FacMethod fm)
			{
				string strEl = e.Args["el"];
				if(string.IsNullOrEmpty(strEl)) return;

				DmMultiZnGroupElement el = DmMultiZnGroupElement.Parse(strEl, mg);

				if(Program.Verbose) TraceIntro(ls, el, mg);

				List<int[]> lFacs = new List<int[]>();
				if(fm == FacMethod.Periodic)
				{
					int[] vFac = new int[ls.Blocks.Count];
					if(!DmMultiZnGroup.Factorize(ls, mg, el, vFac))
					{
						Program.TraceLine("Failed to find factorization!");
						return;
					}
					lFacs.Add(vFac);
				}
				else if(fm == FacMethod.Generic)
				{
					lFacs = DmMultiZnGroup.FactorizeAbEx(ls, mg, el, new DmTraceInfo());
				}
				else
				{
					Program.TraceLine("Unknown factorization method.");
					return;
				}

				bool bTeX = (Program.OutputFormat == OutputFmt.TeX);

				foreach(int[] v in lFacs)
				{
					StringBuilder sbFac = new StringBuilder();
					for(int i = 0; i < v.Length; ++i)
					{
						if(i > 0) sbFac.Append(" + ");
						sbFac.Append(ls.Blocks[i].Elements[v[i]]);
					}

					if(!Program.Verbose)
						Console.WriteLine((bTeX ? "$" : string.Empty) + el.ToString() +
							" = " + sbFac.ToString() + (bTeX ? "$." : "."));
				}
			}
		}
	}
}
