﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Numerics;
using System.Text;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM;
using LogSig.DM.DihedralGroup;
using LogSig.DM.MultiZnGroup;
using LogSig.DM.LsGen;
using LogSig.Util;

namespace LogSig.AppModules
{
	public sealed class AppModGen : AppMod
	{
		public AppModGen()
		{
			AddCommand("genaperiodic", GenAperiodic);
			AddCommand("gendih", GenDih);
			AddCommand("enumdih", EnumDih);
			AddCommand("enumdihrand", EnumDihRandomized);
			AddCommand("crf", ComputeLogSigGenRelFreq);
		}

		private static void GenAperiodic(object sender, AppModEventArgs e)
		{
			DmMultiZnGroup grpS = new DmMultiZnGroup(new BigInteger[] { 2, 3, 5 });
			DmBlockSeq<DmMultiZnGroupElement> lsS = new DmBlockSeq<DmMultiZnGroupElement>();
			lsS.AddBlock(new DmMultiZnGroupElement(new BigInteger[] { 0, 0, 0 }),
				new DmMultiZnGroupElement(new BigInteger[] { 1, 0, 0 }));
			lsS.AddBlock(new DmMultiZnGroupElement(new BigInteger[] { 0, 0, 0 }),
				new DmMultiZnGroupElement(new BigInteger[] { 0, 1, 0 }),
				new DmMultiZnGroupElement(new BigInteger[] { 1, 2, 0 }));
			lsS.AddBlock(new DmMultiZnGroupElement(new BigInteger[] { 0, 0, 0 }),
				new DmMultiZnGroupElement(new BigInteger[] { 1, 1, 1 }),
				new DmMultiZnGroupElement(new BigInteger[] { 0, 2, 2 }),
				new DmMultiZnGroupElement(new BigInteger[] { 1, 2, 3 }),
				new DmMultiZnGroupElement(new BigInteger[] { 0, 1, 4 }));
			DmTraceInfo tiS = new DmTraceInfo();
			List<int[]> vS = DmMultiZnGroup.FactorizeAbEx(lsS, grpS, grpS.Identity,
				tiS);
			Debug.Assert((vS != null) && (vS.Count == 1));

			DmMultiZnGroup grp;
			DmBlockSeq<DmMultiZnGroupElement> ls;
			int p = e.Args.GetInt("p", 3);
			int n = e.Args.GetInt("n", 10);
			int nRounds = e.Args.GetInt("r", 100);

			int s = e.Args.GetInt("s", 1);
			for(int si = 0; si < s; ++si)
			{
				DmMultiZnGroup.GenAperiodic(p, n, out grp, out ls);

				Console.WriteLine("Group: " + grp.Name);
				Console.WriteLine(ls.ToString());

				int nMaxWidth = 0;
				if(e.Args["testfacs"] != null)
				{
					for(int r = 0; r < nRounds; ++r)
					{
						DmMultiZnGroupElement g = grp.GenerateRandomElement(e.Random);

						int iStartMS = Environment.TickCount;

						DmTraceInfo ti = new DmTraceInfo();
						List<int[]> lFacs = DmMultiZnGroup.FactorizeAbEx(ls, grp, g, ti);
						if((lFacs == null) || (lFacs.Count != 1))
							throw new InvalidOperationException();
						if(!ls.MultiplyElements(new List<int>(lFacs[0]),
							grp).Equals(g))
							throw new InvalidOperationException();

						if(ti.MaxWidth > nMaxWidth) nMaxWidth = ti.MaxWidth;

						if(r >= 1) Console.Write(", ");
						Console.Write(Environment.TickCount - iStartMS);
						Console.Write(" ms");
					}
					Console.WriteLine();
				}

				if(nMaxWidth > 0) Console.WriteLine("Max width: " + nMaxWidth.ToString());
				Console.WriteLine();
			}
		}

		private static void GenDih(object sender, AppModEventArgs e)
		{
			bool bSort = (e.Args["sort"] != null);
			bool bCountOnly = (e.Args["countonly"] != null);
			int iSigsToGen = e.Args.GetInt("c", -1);

			BigInteger biOrder = e.Args.GetBigInt("n", new BigInteger(2 * 3 * 5));
			DmDihedralGroup grp = new DmDihedralGroup(biOrder);

			HashSet<string> hsWritten = new HashSet<string>();
			var lsType = CreateLogSig(grp, e);

			DmLogSigDelegate<DmDihedralGroupElement> fReport = delegate(
				DmBlockSeq<DmDihedralGroupElement> ls)
			{
				if((iSigsToGen >= 0) && (hsWritten.Count >= iSigsToGen))
					return false;

				string str = ls.ToString();
				if(hsWritten.Add(str))
				{
					if(!bCountOnly) Console.WriteLine(str);
				}
				return true;
			};

			int tLast = Environment.TickCount;
			ulong uRounds = 0;
			while(true)
			{
				var ls = grp.CreateRandomSig(lsType, bSort, e.Random);
				if(biOrder <= 100000) { Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, grp)); }

				if(!fReport(ls)) break;

				++uRounds;
				if(bCountOnly && ((uRounds % 10000) == 0))
				{
					int t = Environment.TickCount;
					if((t - tLast) >= 5000)
					{
						Console.Write(hsWritten.Count.ToString() + " ");
						tLast = t;
					}
				}
			}
		}

		private static void ComputeLogSigGenRelFreq(object sender, AppModEventArgs e)
		{
			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmeComputeGenRelFreq), e);
		}

		internal static DmBlockSeq<T> CreateEmptySig<T>(string strType, IDmGroup<T> grp)
			where T : IDmGroupElement<T>
		{
			string[] vBlkSizes = strType.Split(',');

			var ls = new DmBlockSeq<T>();
			ls.Blocks.Clear();
			foreach(string strBlkSize in vBlkSizes)
			{
				ls.AddBlock();
				for(int iElem = 0; iElem < int.Parse(strBlkSize); ++iElem)
					ls.Blocks[ls.Blocks.Count - 1].Elements.Add(grp.Identity);
			}

			return ls;
		}

		internal static DmBlockSeq<T> CreateLogSig<T>(IDmGroup<T> grp, AppModEventArgs e)
			where T : IDmGroupElement<T>
		{
			DmBlockSeq<T> ls;

			string strType = e.Args["bstype"];
			if(!string.IsNullOrEmpty(strType)) ls = CreateEmptySig(strType, grp);
			else ls = grp.CreateDefaultSig();

			if(ls.Blocks.Aggregate(BigInteger.One, (acc, blk) =>
				acc * blk.Elements.Count) != grp.Order)
				throw new Exception("Product of block sizes not matching group order!");
			Console.Write(grp.Name + ", ");
			Console.WriteLine("type: " + ls.GetTypeString());

			return ls;
		}

		internal static class AmeComputeGenRelFreq
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e)
				where T : IDmGroupElement<T>
			{
				DmBlockSeq<T> ls = CreateLogSig(grp, e);
				Console.WriteLine("Random number generator type: " +
					e.Random.GetType().ToString());

				// bool bDih2Blk = ((grp is DmDihedralGroup) && ls.Blocks.TrueForAll(
				//	blk => (blk.Elements.Count == 2)));
				// DmDihedralGroup gDih = (grp as DmDihedralGroup);

				ulong ulGen = 0, ulLogSigs = 0;
				int nTicks = Environment.TickCount;
				while(true)
				{
					DmLogSigUtil.RandomizeCanonical(ls, grp, e.Random);

					++ulGen;
					if(ulGen == 0) throw new OverflowException("ulGen");

					/* bool bIsLogSig;
					if(bDih2Blk)
					{
						bIsLogSig = true;
						ls = gDih.Simplify2PowCanonical(ls);

						DmDihedralGroupElement elLast = ls.Blocks[
							ls.Blocks.Count - 1].Elements[1];
						if(!elLast.Reflection) continue;

						List<int> vPows = new List<int>();
						for(int x = 0; x < ls.Blocks.Count - 1; ++x)
						{
							DmDihedralGroupElement el = ls.Blocks[x].Elements[1];
							if(el.Rotation == 0) { bIsLogSig = false; break; }
							else
							{
								int nPow = NumUtil.GetPow(el.Rotation, 2);
								if(vPows.IndexOf(nPow) >= 0) { bIsLogSig = false; break; }
								else vPows.Add(nPow);
							}
						}
					} */
					bool bIsLogSig = DmLogSigUtil.GeneratesWholeGroup(ls, grp);

					if(bIsLogSig)
					{
						++ulLogSigs;

						if((Environment.TickCount - nTicks) >= 10000)
						{
							Console.Write((double)ulLogSigs / (double)ulGen);
							Console.Write(" = ");
							Console.Write(ulLogSigs);
							Console.Write(" / ");
							Console.Write(ulGen);
							Console.Write(" = ");
							Console.Write(((double)ulLogSigs).ToString("e"));
							Console.Write(" / ");
							Console.Write(((double)ulGen).ToString("e"));
							Console.WriteLine(".");
							nTicks = Environment.TickCount;
						}
					}
				}
			}
		}

		private static void EnumDih(object sender, AppModEventArgs e)
		{
			bool bSort = (e.Args["sort"] != null);
			bool bCountOnly = (e.Args["countonly"] != null);
			bool bTrySimplify = (e.Args["trysimplify"] != null);
			bool bBreakOnHighTau = (e.Args["breakonhightau"] != null);

			BigInteger biOrder = e.Args.GetBigInt("n", new BigInteger(2 * 3 * 5));
			DmDihedralGroup grp = new DmDihedralGroup(biOrder);

			// HashSet<string> hsWritten = new HashSet<string>();
			var ls = AppModGen.CreateLogSig(grp, e);

			int x2 = ls.Blocks.FindIndex(blk => (blk.Elements.Count == 2));
			if(bBreakOnHighTau && (x2 != 0))
			{
				Console.WriteLine("Block of size 2 at invalid position!");
				return;
			}

			ulong uCount = 0;
			DmLogSigDelegate<DmDihedralGroupElement> f = delegate(
				DmBlockSeq<DmDihedralGroupElement> bs)
			{
				if(bBreakOnHighTau)
				{
					Debug.Assert(bs.Blocks[x2].Elements.Count == 2);
					DmDihedralGroupElement dt = bs.Blocks[x2].Elements[1];
					Debug.Assert(dt.Reflection);
					Debug.Assert(dt.Rotation <= BigInteger.One);
					if(dt.Reflection && !dt.Rotation.IsZero)
						return false;
				}

				if(!bCountOnly)
				{
					string str = bs.ToString();
					// Debug.Assert(!hsWritten.Contains(str));
					Console.WriteLine(str);
				}

				if(bTrySimplify)
				{
					var lsSmp = grp.SimplifyUsingSwapsAndSelShifts(bs);
					if(!lsSmp.Blocks.TrueForAll(blk => ((blk.Elements.Count == 2) ||
						(blk.Elements.FindIndex(el => el.Reflection) < 0))))
						Console.WriteLine(bs.ToString() + " => Simplification failed!");
				}

				++uCount;
				return true;
			};

			DmLogSigUtil.EnumCanBlockSeq(ls, grp, f, true);

			if(bBreakOnHighTau)
				uCount *= (ulong)(grp.Order / 2);

			if(!bSort) Console.WriteLine("Count: " + uCount.ToString());
			else
			{
				BigInteger biUns = new BigInteger(uCount);
				foreach(var blk in ls.Blocks)
					biUns *= NumUtil.Factorial(blk.Elements.Count - 1);

				Console.WriteLine("Sorted count:   " + uCount.ToString() + ".");
				Console.WriteLine("Unsorted count: " + biUns.ToString() + ".");
			}
		}

		private static void EnumDihRandomized(object sender, AppModEventArgs e)
		{
			bool bSort = (e.Args["sort"] != null);
			BigInteger biOrder = e.Args.GetBigInt("n", new BigInteger(2 * 3 * 5));
			DmDihedralGroup grp = new DmDihedralGroup(biOrder);

			HashSet<string> hsWritten = new HashSet<string>();
			var ls = AppModGen.CreateLogSig(grp, e);

			int iEnsure2Blk = -1;
			if(!BigInteger.Divide(biOrder, 2).IsEven)
				iEnsure2Blk = ls.Blocks.FindIndex(blk => (blk.Elements.Count == 2));

			bool bl = (iEnsure2Blk > 0);
			bool br = ((iEnsure2Blk >= 0) && (iEnsure2Blk < (ls.Blocks.Count - 1)));
			BigInteger biSels = 2 * BigInteger.Pow(2,
				(bl ? (ls.Blocks[iEnsure2Blk - 1].Elements.Count - 1) : 0) +
				(br ? (ls.Blocks[iEnsure2Blk + 1].Elements.Count - 1) : 0));

			int tLastPrint = Environment.TickCount;
			long lRound = 0;
			while(true)
			{
				++lRound;
				if((lRound >= 500000) && ((Environment.TickCount - tLastPrint) > 1000))
				{
					Console.Write(hsWritten.Count.ToString() + " ");

					lRound = 0;
					tLastPrint = Environment.TickCount;
				}

				DmLogSigUtil.RandomizeCanonical(ls, grp, e.Random);

				if(iEnsure2Blk >= 0)
				{
					var vBlk = ls.Blocks[iEnsure2Blk].Elements;
					if(!vBlk[1].Reflection)
						vBlk[1] = new DmDihedralGroupElement(vBlk[1].Rotation, true);
				}

				if(DmLogSigUtil.GeneratesWholeGroup(ls, grp))
				{
					if(iEnsure2Blk < 0)
					{
						if(bSort) ls.Blocks.ForEach(blk => blk.Elements.Sort());

						string strSig = ls.ToString();
						if(hsWritten.Contains(strSig)) continue;

						// Console.WriteLine(strSig);
						hsWritten.Add(strSig);
					}
					else
					{
						DmDihedralGroupElement d = ls.Blocks[iEnsure2Blk].Elements[1];

						for(BigInteger r = 0; r < biSels; ++r)
						{
							List<DmDihedralGroupElement> lBlk;
							if(bl && br)
							{
								bool bOpL = e.Random.GenerateBool();
								lBlk = (bOpL ? ls.Blocks[iEnsure2Blk - 1].Elements :
									ls.Blocks[iEnsure2Blk + 1].Elements);
								int y = e.Random.GenerateInt32(lBlk.Count - 1) + 1;
								if(bOpL) lBlk[y] = grp.Multiply(lBlk[y], d);
								else lBlk[y] = grp.Multiply(d, lBlk[y]);
							}
							else if(bl)
							{
								lBlk = ls.Blocks[iEnsure2Blk - 1].Elements;
								int y = e.Random.GenerateInt32(lBlk.Count - 1) + 1;
								lBlk[y] = grp.Multiply(lBlk[y], d);
							}
							else if(br)
							{
								lBlk = ls.Blocks[iEnsure2Blk + 1].Elements;
								int y = e.Random.GenerateInt32(lBlk.Count - 1) + 1;
								lBlk[y] = grp.Multiply(d, lBlk[y]);
							}
							else { Debug.Assert(false); return; }

							Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(ls, grp));

							if(bSort) ls.Blocks.ForEach(blk => blk.Elements.Sort());

							string strSig = ls.ToString();
							if(hsWritten.Contains(strSig)) continue;

							// Console.WriteLine(strSig);
							hsWritten.Add(strSig);
						}
					}
				}
			}
		}
	}
}
