﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Numerics;
using System.Security.Cryptography;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM;
using LogSig.DM.LsGen;
using LogSig.Pkc;
using LogSig.Util;

namespace LogSig.AppModules
{
	public sealed class AppModMst1 : AppMod
	{
		private const string StrPreReq = "mst1";

		private const string HdrFileType = "PKC-MST1";
		private const string HdrFileDesc = "Encrypted";
		private const string HdrVersionDecl = "Version";
		private const string HdrVersionInfo = "1.00";
		private const string HdrGroupElement = "GroupElement";
		private const string HdrIV = "IV";

		public AppModMst1()
		{
			AddCommand("c", CreateKeyPair);
			AddCommand("e", EncryptFile);
			AddCommand("d", DecryptFile);
		}

		private static void CreateKeyPair(object sender, AppModEventArgs e)
		{
			if(e.Args[StrPreReq] == null) return;

			Program.GroupFactories.InvokeGroupMethod(e, typeof(AmmCreateKeyPair), e);
		}

		internal static class AmmCreateKeyPair
		{
			public static void Invoke<T>(IDmGroup<T> grp, AppModEventArgs e)
				where T : IDmGroupElement<T>
			{
				string strCmdBaseName = e.Args["base"];
				string strBaseName = (strCmdBaseName ?? "Generated");

				ulong uRounds;
				if(!ulong.TryParse(e.Args["r"], out uRounds)) uRounds = 5000;

				string strPassword = e.Args["pw"];
				if(string.IsNullOrEmpty(strPassword))
				{
					Console.WriteLine("A password must be passed using the '-pw' option.");
					return;
				}

				byte[] pbKey32 = new byte[32];
				Program.CryptoRandom.Generate(pbKey32, 0, pbKey32.Length);

				byte[] pbCtr16 = new byte[16];
				Program.CryptoRandom.Generate(pbCtr16, 0, pbCtr16.Length);

				CryptoRandomGen rand = new AesCryptoRandomGen(pbKey32, pbCtr16);

				string strTrfIORels = string.Empty;

				List<int> lFac;
				var ls = LsGen<T>.Generate(grp, uRounds, rand, default(T),
					ref strTrfIORels, out lFac);

				PkcPublicKey pubKey = PkcPublicKey.CreatePublic(grp, ls);
				PkcPrivateKey prvKey = PkcPrivateKey.CreatePrivate(grp, ls, pbKey32,
					pbCtr16, uRounds, strTrfIORels, strPassword);

				using(FileStream fsPub = new FileStream(strBaseName + ".pub",
					FileMode.Create, FileAccess.Write, FileShare.None))
				{
					PkcKeyFile.Save(fsPub, pubKey);
				}

				using(FileStream fsPriv = new FileStream(strBaseName + ".prv",
					FileMode.Create, FileAccess.Write, FileShare.None))
				{
					PkcKeyFile.Save(fsPriv, prvKey);
				}
			}
		}

		private static void EncryptFile(object sender, AppModEventArgs e)
		{
			if(e.Args[StrPreReq] == null) return;

			string strPubKey = e.Args["key"];
			PkcPublicKey pk = PkcKeyFile.Load<PkcPublicKey>(strPubKey);
			if(pk == null) throw new FileNotFoundException();

			foreach(string strFile in e.Args.FileNames)
			{
				using(FileStream fsIn = new FileStream(strFile, FileMode.Open,
					FileAccess.Read, FileShare.Read))
				{
					string strOutFile = (e.Args["outfile"] ?? (strFile + ".enc"));
					using(FileStream fsOut = new FileStream(strOutFile,
						FileMode.Create, FileAccess.Write, FileShare.None))
					{
						Program.GroupFactories.InvokeGroupMethod(pk.Group,
							typeof(AmmEncryptFile), fsIn, fsOut, pk);
					}
				}
			}
		}

		internal static class AmmEncryptFile
		{
			public static void Invoke<T>(IDmGroup<T> grp, FileStream fsIn,
				FileStream fsOut, PkcPublicKey pk)
				where T : IDmGroupElement<T>
			{
				byte[] pbIV16 = new byte[16];
				Program.CryptoRandom.Generate(pbIV16, 0, pbIV16.Length);

				DmBlockSeq<T> ls = DmBlockSeq<T>.Parse(pk.LogSig, grp);

				BigInteger biSecFac = Program.CryptoRandom.GenerateBigInt(grp.Order);
				SHA256Managed sha256 = new SHA256Managed();
				byte[] pbKey32 = sha256.ComputeHash(biSecFac.ToByteArray());

				T tPubVal = DmLogSigUtil.BigIntToElement(ls, grp, biSecFac);

				BinaryWriter bw = new BinaryWriter(fsOut);

				WriteString(bw, HdrFileType, HdrFileDesc);
				WriteString(bw, HdrVersionDecl, HdrVersionInfo);
				WriteString(bw, HdrGroupElement, tPubVal.ToString());
				WriteString(bw, HdrIV, Convert.ToBase64String(pbIV16, Base64FormattingOptions.None));
				WriteString(bw, null, null); // Terminate header

				Stream sEnc = MemUtil.EncryptCbc(fsOut, pbKey32, pbIV16, true);
				MemUtil.CopyStream(fsIn, sEnc);

				fsIn.Close();
				sEnc.Close();
				bw.Close();
				fsOut.Close();
			}
		}

		private static KeyValuePair<string, string> ReadString(BinaryReader br)
		{
			int nFieldLength = br.ReadInt32();
			if(nFieldLength < 0) return new KeyValuePair<string, string>(null, null);

			byte[] pb = br.ReadBytes(nFieldLength);
			string strKey = Encoding.UTF8.GetString(pb);

			nFieldLength = br.ReadInt32();
			pb = br.ReadBytes(nFieldLength);
			string strValue = Encoding.UTF8.GetString(pb);

			return new KeyValuePair<string, string>(strKey, strValue);
		}

		private static void WriteString(BinaryWriter bw, string strKey, string strValue)
		{
			if(strKey == null) { bw.Write((int)-1); return; }

			UTF8Encoding utf8 = new UTF8Encoding(false);

			bw.Write(strKey.Length);
			if(strKey.Length != 0) bw.Write(utf8.GetBytes(strKey));

			if(strValue == null) strValue = string.Empty;
			bw.Write(strValue.Length);
			if(strValue.Length != 0) bw.Write(utf8.GetBytes(strValue));
		}

		private static void DecryptFile(object sender, AppModEventArgs e)
		{
			if(e.Args[StrPreReq] == null) return;

			string strPassword = e.Args["pw"];
			if(string.IsNullOrEmpty(strPassword))
			{
				Console.WriteLine("A password must be passed using the '-pw' option.");
				return;
			}

			string strPubKey = e.Args["key"];
			PkcPrivateKey pk = PkcKeyFile.Load<PkcPrivateKey>(strPubKey);
			if(pk == null) throw new FileNotFoundException();

			foreach(string strFile in e.Args.FileNames)
			{
				using(FileStream fsIn = new FileStream(strFile, FileMode.Open,
					FileAccess.Read, FileShare.Read))
				{
					string strOutFile = (e.Args["outfile"] ?? (strFile + ".dec"));
					using(FileStream fsOut = new FileStream(strOutFile,
						FileMode.Create, FileAccess.Write, FileShare.None))
					{
						Program.GroupFactories.InvokeGroupMethod(pk.Group,
							typeof(AmmDecryptFile), fsIn, fsOut, pk, strPassword);
					}
				}
			}
		}

		internal static class AmmDecryptFile
		{
			public static void Invoke<T>(IDmGroup<T> grp, FileStream fsIn,
				FileStream fsOut, PkcPrivateKey pk, string strPassword)
				where T : class, IDmGroupElement<T>
			{
				BinaryReader br = new BinaryReader(fsIn);

				T tPubVal = null;
				byte[] pbIV16 = null;
				while(true)
				{
					KeyValuePair<string, string> kvp = ReadString(br);
					if(kvp.Key == null) break;

					if(kvp.Key.Equals(HdrFileType))
					{
						if(!kvp.Value.Equals(HdrFileDesc))
							throw new FormatException("Unknown file format!");
					}
					else if(kvp.Key.Equals(HdrVersionDecl))
					{
						if(!kvp.Value.Equals(HdrVersionInfo))
							throw new FormatException("Unknown file version!");
					}
					else if(kvp.Key.Equals(HdrGroupElement))
						tPubVal = grp.ParseElement(kvp.Value);
					else if(kvp.Key.Equals(HdrIV))
						pbIV16 = Convert.FromBase64String(kvp.Value);
					else { Debug.Assert(false); }
				}

				if((tPubVal == null) || (pbIV16 == null))
					throw new FormatException("File loading error!");

				T tToFactor = tPubVal;

				string strTrfIORels = pk.GeneratorTrfIORels;

				CryptoRandomGen rand = new AesCryptoRandomGen(PkcPrivateKey.DecryptKey(
					pk.GeneratorKey, strPassword), PkcPrivateKey.DecryptKey(
					pk.GeneratorCtr, strPassword));

				List<int> lFactorization;
				var ls = LsGen<T>.Generate(grp, pk.GeneratorRounds, rand, tToFactor,
					ref strTrfIORels, out lFactorization);

				Debug.Assert(ls.ToString().Equals(pk.LogSig));

				BigInteger biSecKey = DmLogSigUtil.FactorizationToBigInt(ls,
					lFactorization.ToArray());
				SHA256Managed sha256 = new SHA256Managed();
				byte[] pbKey32 = sha256.ComputeHash(biSecKey.ToByteArray());

				Stream sDec = MemUtil.EncryptCbc(fsIn, pbKey32, pbIV16, false);
				MemUtil.CopyStream(sDec, fsOut);

				sDec.Close();
				br.Close();
				fsIn.Close();
				fsOut.Close();
			}
		}
	}
}
