﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;
using System.Diagnostics;

using LogSig.Interfaces;

namespace LogSig.Crypto
{
	/// <summary>
	/// AES in CTR mode as a cryptographically secure pseudo-random
	/// number generator.
	/// </summary>
	public sealed class AesCryptoRandomGen : CryptoRandomGen
	{
		private byte[] m_pbBuffer = new byte[16];
		private int m_iBufIndex = 16;

		private byte[] m_pbCtr = new byte[16];
		private ICryptoTransform m_trfAes;

		public AesCryptoRandomGen(byte[] pbKey32, byte[] pbCtr16)
		{
			if(pbKey32 == null) throw new ArgumentNullException("pbKey32");
			if(pbKey32.Length != 32) throw new ArgumentException("pbKey32");
			if(pbCtr16 == null) throw new ArgumentNullException("pbCtr16");
			if(pbCtr16.Length != 16) throw new ArgumentException("pbCtr16");

			RijndaelManaged r = new RijndaelManaged();
			if(r.BlockSize != 128) throw new CryptographicUnexpectedOperationException();

			r.KeySize = 256;
			r.Key = pbKey32;
			r.Mode = CipherMode.ECB;

			m_trfAes = r.CreateEncryptor();
			Debug.Assert(m_trfAes.CanReuseTransform);

			Array.Copy(pbCtr16, m_pbCtr, 16);
		}

		private void GenerateBlock()
		{
			m_trfAes.TransformBlock(m_pbCtr, 0, m_pbCtr.Length, m_pbBuffer, 0);
			m_iBufIndex = 0;

			int p = m_pbCtr.Length - 1;
			while(p >= 0) // Increment CTR, big-endian
			{
				++m_pbCtr[p];
				if(m_pbCtr[p] == 0) --p;
				else break;
			}
		}

		public override void Generate(byte[] pb, int iOffset, int iCount)
		{
			for(int pOut = iOffset; pOut < (iOffset + iCount); ++pOut)
			{
				if(m_iBufIndex >= m_pbBuffer.Length) GenerateBlock();

				pb[pOut] = m_pbBuffer[m_iBufIndex];
				++m_iBufIndex;
			}
		}
	}
}
