﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;

using LogSig.Util;

namespace LogSig.Crypto
{
	public abstract class CryptoRandomGen
	{
		public abstract void Generate(byte[] pb, int iOffset, int iCount);

		public virtual bool GenerateBool()
		{
			byte[] pb = new byte[1];
			Generate(pb, 0, 1);
			return ((pb[0] & 1) != 0);
		}

		public virtual int GenerateInt32(int nExclMax)
		{
			if(nExclMax <= 0) throw new ArgumentOutOfRangeException("nExclMax");

			byte[] pb = new byte[4];

			int nValue, nRem, nTest;
			do
			{
				Generate(pb, 0, pb.Length);
				nValue = MemUtil.BytesToInt32(pb);
				nValue &= 0x7FFFFFFF;
				nRem = nValue % nExclMax;
				nTest = nValue - nRem + (nExclMax - 1);
			}
			while(nTest < 0);

			return nRem;
		}

		public virtual long GenerateInt64(long nExclMax)
		{
			if(nExclMax <= 0) throw new ArgumentOutOfRangeException("nExclMax");

			byte[] pb = new byte[8];

			long nValue, nRem, nTest;
			do
			{
				Generate(pb, 0, pb.Length);
				nValue = MemUtil.BytesToInt64(pb);
				nValue &= 0x7FFFFFFFFFFFFFFFL;
				nRem = nValue % nExclMax;
				nTest = nValue - nRem + (nExclMax - 1);
			}
			while(nTest < 0);

			return nRem;
		}

		public virtual BigInteger GenerateBigInt(BigInteger nExclMax)
		{
			if(nExclMax <= 0) throw new ArgumentOutOfRangeException("nExclMax");

			byte[] pbMod = nExclMax.ToByteArray();

			byte btTop = pbMod[pbMod.Length - 1]; // Little-endian
			byte btMask = 0;
			while(btTop != 0)
			{
				btMask <<= 1;
				btMask |= 1;
				btTop >>= 1;
			}

			byte[] pbRand = new byte[pbMod.Length + 1];
			pbRand[pbRand.Length - 1] = 0; // Disambiguate sign

			BigInteger bi;
			while(true)
			{
				Generate(pbRand, 0, pbRand.Length - 1);
				pbRand[pbRand.Length - 2] &= btMask;

				bi = new BigInteger(pbRand);
				if(bi < nExclMax) break;
			}

			return bi;
		}
	}
}
