﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM.SnGroup;

namespace LogSig.DM.AnGroup
{
	public sealed class DmAnGroup : IDmGroup<DmAnGroupElement>
	{
		public string Name
		{
			get { return ("A" + m_nDegree.ToString()); }
		}
		// public string ShortName { get { return "Alt"; } }
		// public string ShortNameAlt { get { return "An"; } }

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private readonly int m_nDegree;
		/// <summary>
		/// Degree of the group (size of the set that the permutation
		/// group acts on).
		/// </summary>
		public int Degree
		{
			get { return m_nDegree; }
		}

		private readonly DmAnGroupElement m_id;
		public DmAnGroupElement Identity { get { return m_id; } }

		public DmAnGroup(BigInteger biOrder)
		{
			if(biOrder <= 0) throw new ArgumentOutOfRangeException("biOrder");

			m_biOrder = biOrder;
			m_nDegree = 1;

			BigInteger p = BigInteger.One;
			while(true)
			{
				if(p > (biOrder * 2)) throw new ArgumentOutOfRangeException("biOrder");
				if(p == (biOrder * 2)) break;

				++m_nDegree;
				p *= m_nDegree;
			}

			m_id = new DmAnGroupElement(m_nDegree);
		}

		public DmAnGroupElement ParseElement(string str)
		{
			DmSnGroupElement el = DmSnGroupElement.Parse(str, m_nDegree);
			return new DmAnGroupElement(el.Permutation); // Checks for sign == 1
		}

		public DmAnGroupElement Multiply(DmAnGroupElement a, DmAnGroupElement b)
		{
			int[] vA = a.Permutation;
			int[] vB = b.Permutation;
			int[] v = new int[m_nDegree];
			for(int i = 0; i < m_nDegree; ++i) v[i] = vA[vB[i]];

			return new DmAnGroupElement(v);
		}

		public DmAnGroupElement Invert(DmAnGroupElement t)
		{
			int[] vT = t.Permutation;
			int[] v = new int[m_nDegree];
			for(int i = 0; i < m_nDegree; ++i) v[vT[i]] = i;

			return new DmAnGroupElement(v);
		}

		public DmAnGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			int[] v = new int[m_nDegree];
			for(int a = 0; a < m_nDegree; ++a) v[a] = a;

			for(int i = (m_nDegree - 1); i >= 1; --i)
			{
				int p = rand.GenerateInt32(i + 1);

				int t = v[p];
				v[p] = v[i];
				v[i] = t;
			}

			if(DmSnGroupElement.ComputeSign(v) == -1)
			{
				int px = rand.GenerateInt32(m_nDegree);
				int py = px;
				while(py == px) { py = rand.GenerateInt32(m_nDegree); }

				int t = v[px];
				v[px] = v[py];
				v[py] = t;
			}

			Debug.Assert(DmSnGroupElement.ComputeSign(v) == 1);
			return new DmAnGroupElement(v);
		}

		public DmAnGroupElement NextElement(DmAnGroupElement t)
		{
			int[] v = new int[m_nDegree];
			Array.Copy(t.Permutation, v, m_nDegree);

			do { NumUtil.NextPermutation(v); }
			while(DmSnGroupElement.ComputeSign(v) == -1);

			return new DmAnGroupElement(v);
		}

		public static BigInteger ElementToBigInt(int[] vPerm)
		{
			BigInteger bi = BigInteger.Zero;

			List<int> l = new List<int>(vPerm);
			for(int i = 0; ; ++i)
			{
				if(l.Count <= 1) break;

				int p = l.IndexOf(i);
				Debug.Assert(p >= 0);

				bi *= l.Count;
				bi += p;

				l.RemoveAt(p);
			}

			return bi;
		}

		public BigInteger ElementToBigInt(DmAnGroupElement t)
		{
			return ElementToBigInt(t.Permutation);
		}

		public DmAnGroupElement BigIntToElement(BigInteger biElem)
		{
			List<int> vIndices = new List<int>();
			for(int i = 0; i < m_nDegree; ++i)
			{
				vIndices.Add((int)(biElem % (i + 1)));
				biElem /= (i + 1);
			}

			int[] v = new int[m_nDegree];
			for(int a = 0; a < m_nDegree; ++a) v[a] = -1;
			for(int j = 0; j < m_nDegree; ++j)
			{
				int vp = 0;
				for(int s = 0; s < m_nDegree; ++s)
				{
					if(v[s] < 0)
					{
						if(vp == vIndices[m_nDegree - j - 1])
						{
							v[s] = j;
							break;
						}

						++vp;
					}
				}
			}

			return new DmAnGroupElement(v);
		}

		public DmBlockSeq<DmAnGroupElement> CreateDefaultSig()
		{
			/* DmBlockSeq<DmAnGroupElement> ls = new DmBlockSeq<DmAnGroupElement>();

			for(int i = 1; i < m_nDegree; ++i)
			{
				var blk = ls.AddBlock(m_id);

				for(int j = 0; j < i; ++j)
				{
					int[] v = new int[m_nDegree];
					for(int a = 0; a < m_nDegree; ++a) v[a] = a;
					int t = v[i];
					v[i] = v[j];
					v[j] = t;

					blk.Elements.Add(new DmAnGroupElement(v));
				}
			}

			return ls; */

			throw new NotImplementedException();
		}

		public int[] FactorInDefaultSig(DmAnGroupElement t)
		{
			/* DmBlockSeq<DmAnGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			DmAnGroupElement elToFactor = t;
			for(int i = (vFactorization.Length - 1); i >= 0; --i)
			{
				int[] v = elToFactor.Permutation;
				int p = Array.IndexOf<int>(v, i + 1);
				Debug.Assert(p >= 0);

				if(p == (i + 1)) vFactorization[i] = 0;
				else
				{
					vFactorization[i] = p + 1;
					elToFactor = Multiply(elToFactor,
						Invert(ls.Blocks[i].Elements[p + 1]));
				}
			}

			return vFactorization; */

			throw new NotImplementedException();
		}
	}
}
