﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.DM.DirectProductGroup;

namespace LogSig.DM.DihedralGroup
{
	/// <summary>
	/// Dihedral group (group of symmetries of a regular polygon).
	/// </summary>
	public sealed partial class DmDihedralGroup : IDmGroup<DmDihedralGroupElement>
	{
		public int[] FactorInDefaultSig(DmDihedralGroupElement t)
		{
			DmBlockSeq<DmDihedralGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			BigInteger biRot = t.Rotation;
			for(int i = 0; i < vFactorization.Length - 1; ++i)
			{
				int p = ls.Blocks[i].Elements.Count;
				vFactorization[i] = (int)(biRot % p);
				biRot /= p;
			}

			vFactorization[vFactorization.Length - 1] = (t.Reflection ? 1 : 0);
			return vFactorization;
		}

		public List<int> FactorizePrPow(DmBlockSeq<DmDihedralGroupElement> ls,
			int nPrime, DmDihedralGroupElement elToFactor, DmTraceInfo traceInfo)
		{
			List<int> vSelected = new List<int>();
			for(int i = 0; i < ls.Blocks.Count; ++i) vSelected.Add(-1); // None selected

			return FactorizePrPowRec(ls, nPrime, elToFactor.Rotation, -1,
				vSelected, false, elToFactor, traceInfo);
		}

		private List<int> FactorizePrPowRec(DmBlockSeq<DmDihedralGroupElement> ls,
			int nPrime, BigInteger nRotToFactor, int nCurPow, List<int> vSelected,
			bool bTailTau, DmDihedralGroupElement elGoal, DmTraceInfo traceInfo)
		{
			traceInfo.RecursiveCallCount = traceInfo.RecursiveCallCount + 1;

			// DmDihedralGroupElement gl = ls.Multiply(vSelected, this);
			// Debug.Assert(!gl.Reflection); // Reflection in tail block only
			// if(bTailTau) gl = new DmDihedralGroupElement(gl.Rotation, true);
			// if(gl.CompareTo(elGoal) == 0) return vSelected;
			// if(vSelected.IndexOf(-1) < 0) return null;

			if(vSelected.IndexOf(-1) < 0)
			{
				DmDihedralGroupElement gl = ls.MultiplyElements(vSelected, this);
				Debug.Assert(!gl.Reflection); // Reflection in tail block only
				if(bTailTau) gl = new DmDihedralGroupElement(gl.Rotation, true);
				return ((gl.CompareTo(elGoal) == 0) ? vSelected : null);
			}

			// if(nCurPow == NumUtil.GetPow(m_nRotMod, nPrime)) return null;

			if(!TauTest(ls, vSelected, bTailTau, elGoal)) return null;

			int nPowToFactor = NumUtil.GetPow(nRotToFactor, nPrime);
			if((nPowToFactor < nCurPow) && (nPowToFactor != -1)) return null;

			bool bEvenSel = (nPowToFactor != nCurPow);

			List<List<int>> vCand = new List<List<int>>();
			List<int> vCurCand = new List<int>();
			for(int icf = 0; icf < ls.Blocks.Count; ++icf) vCurCand.Add(-1);
			FindCandidatesRec(ls, nPrime, nCurPow, vSelected, vCand,
				vCurCand, bEvenSel, 0);

			foreach(List<int> lCand in vCand)
			{
				DmBlockSeq<DmDihedralGroupElement> lsNew = ls.Clone();
				List<int> vSelectedNew = new List<int>(vSelected);
				bool bTailTauNew = bTailTau;

				for(int iCandSel = 0; iCandSel < lCand.Count; ++iCandSel)
				{
					if(lCand[iCandSel] >= 0)
						SelectElement(lsNew, iCandSel, lCand[iCandSel],
							vSelectedNew, ref bTailTauNew);
				}

				AutoSelectMinor(lsNew, nPrime, nCurPow, vSelectedNew);

				SortedDictionary<DmDihedralGroupElement, bool> dictCur =
					new SortedDictionary<DmDihedralGroupElement, bool>();
				GetPossiblePartialProducts(lsNew, 0, vSelectedNew, 0, 1, dictCur);

				foreach(KeyValuePair<DmDihedralGroupElement, bool> kvpCur in dictCur)
				{
					BigInteger nRotToFactorNew = ((elGoal.Rotation - kvpCur.Key.Rotation +
						m_nRotMod) % m_nRotMod);

					List<int> vSub = FactorizePrPowRec(lsNew, nPrime, nRotToFactorNew,
						nCurPow + 1, vSelectedNew, bTailTauNew, elGoal, traceInfo);
					if(vSub != null) return vSub;
				}
			}

			return null;
		}

		private void SelectElement(DmBlockSeq<DmDihedralGroupElement> ls,
			int iBlock, int iElem, List<int> vSelected, ref bool bTailTau)
		{
			Debug.Assert(vSelected[iBlock] < 0);
			vSelected[iBlock] = iElem;

			DmDihedralGroupElement g = ls.Blocks[iBlock].Elements[iElem];
			if(g.Reflection)
			{
				ls.Blocks[iBlock].Elements[iElem] = new DmDihedralGroupElement(
					g.Rotation, false);

				for(int i = iBlock + 1; i < ls.Blocks.Count; ++i)
				{
					for(int y = 1; y < ls.Blocks[i].Elements.Count; ++y)
					{
						DmDihedralGroupElement h = ls.Blocks[i].Elements[y];
						ls.Blocks[i].Elements[y] = new DmDihedralGroupElement(
							((m_nRotMod - h.Rotation) % m_nRotMod), h.Reflection);
					}
				}

				bTailTau = !bTailTau;
			}
		}

		private static void AutoSelectMinor(DmBlockSeq<DmDihedralGroupElement> ls,
			int nPrime, int nCurPow, List<int> vSelected)
		{
			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				if(vSelected[x] >= 0) continue;

				List<DmDihedralGroupElement> vElem = ls.Blocks[x].Elements;
				bool bMinorOnly = true;
				for(int y = 1; y < vElem.Count; ++y)
				{
					if(NumUtil.GetPow(vElem[y].Rotation, nPrime) > nCurPow)
					{
						bMinorOnly = false;
						break;
					}
				}

				if(bMinorOnly) vSelected[x] = 0; // Select identity
			}
		}

		private void GetPossiblePartialProducts(DmBlockSeq<DmDihedralGroupElement> ls,
			int iBlock, List<int> vSelected, BigInteger nCurrentRot, int iFactor,
			SortedDictionary<DmDihedralGroupElement, bool> dictGen)
		{
			if(iBlock == ls.Blocks.Count)
			{
				dictGen[new DmDihedralGroupElement(nCurrentRot, false)] = true;
				return;
			}

			if(vSelected[iBlock] == 0)
			{
				GetPossiblePartialProducts(ls, iBlock + 1, vSelected, nCurrentRot,
					iFactor, dictGen);
			}
			else if(vSelected[iBlock] > 0)
			{
				DmDihedralGroupElement g = ls.Blocks[iBlock].Elements[vSelected[iBlock]];
				nCurrentRot = (((nCurrentRot + (iFactor * g.Rotation)) +
					m_nRotMod) % m_nRotMod);
				if(g.Reflection) iFactor *= -1;

				GetPossiblePartialProducts(ls, iBlock + 1, vSelected, nCurrentRot,
					iFactor, dictGen);
			}
			else // (vSelected[iBlock] < 0)
			{
				if(ls.Blocks[iBlock].Elements.Count(el => el.Reflection) > 0)
				{
					GetPossiblePartialProducts(ls, iBlock + 1, vSelected, nCurrentRot,
						-1, dictGen);
					GetPossiblePartialProducts(ls, iBlock + 1, vSelected, nCurrentRot,
						1, dictGen);
				}
				else
					GetPossiblePartialProducts(ls, iBlock + 1, vSelected, nCurrentRot,
						iFactor, dictGen);
			}
		}

		private void FindCandidatesRec(DmBlockSeq<DmDihedralGroupElement> ls,
			int nPrime, BigInteger nCurPow, List<int> vSelected, List<List<int>> vStorage,
			List<int> vCurCand, bool bReq0EquivSel, int iBlock)
		{
			if(iBlock == ls.Blocks.Count)
			{
				if(nCurPow != -1)
				{
					int nSelCount = vCurCand.Count(iCand => (iCand > 0));

					if(nPrime == 2)
					{
						if(bReq0EquivSel && ((nSelCount & 1) != 0)) return;
						if(!bReq0EquivSel && ((nSelCount & 1) == 0)) return;
					}
				}

				vStorage.Add(vCurCand);
				return;
			}

			FindCandidatesRec(ls, nPrime, nCurPow, vSelected, vStorage,
				vCurCand, bReq0EquivSel, iBlock + 1);
			if(vSelected[iBlock] != -1) return;

			List<DmDihedralGroupElement> vElem = ls.Blocks[iBlock].Elements;
			Debug.Assert(vElem[0].CompareTo(m_id) == 0);

			for(int i = 1; i < vElem.Count; ++i)
			{
				DmDihedralGroupElement g = vElem[i];
				if(NumUtil.GetPow(g.Rotation, nPrime) == nCurPow)
				{
					List<int> vCurCandNew = new List<int>(vCurCand);
					vCurCandNew[iBlock] = i;
					FindCandidatesRec(ls, nPrime, nCurPow, vSelected, vStorage,
						vCurCandNew, bReq0EquivSel, iBlock + 1);
				}
			}
		}

		private static bool TauTest(DmBlockSeq<DmDihedralGroupElement> ls,
			List<int> vSelected, bool bTailTau, DmDihedralGroupElement elGoal)
		{
			bool bMulTau = bTailTau;
			for(int i = 0; i < ls.Blocks.Count; ++i)
			{
				if(vSelected[i] < 0)
				{
					foreach(DmDihedralGroupElement el in ls.Blocks[i].Elements)
					{
						if(el.Reflection) return true;
					}
				}
				else
				{
					if(ls.Blocks[i].Elements[vSelected[i]].Reflection)
						bMulTau = !bMulTau;
				}
			}

			return (bMulTau == elGoal.Reflection);
		}

		/* public DmBlockSeq<DmDihedralGroupElement> Simplify2PowCanonical(
			DmBlockSeq<DmDihedralGroupElement> lsSig)
		{
			var ls = lsSig.Clone();

			for(int r = 0; r < 2000; ++r)
			{
				for(int x = 0; x < ls.Blocks.Count - 2; ++x)
				{
					var lsSubSig = ls.GetSubSig(x, 3);
					var lsNewSubSig = lsSubSig.GetRandomEquivalent2PowCanonical(this);
					if(lsSubSig != lsNewSubSig)
					{
						int nOrgTaus = lsSubSig.Blocks.Count(blk => blk.Elements[1].Reflection);
						int nNewTaus = lsNewSubSig.Blocks.Count(blk => blk.Elements[1].Reflection);

						if(nOrgTaus >= nNewTaus)
						{
							for(int xc = 0; xc < 3; ++xc)
								ls.Blocks[x + xc].Elements[1] = lsNewSubSig.Blocks[xc].Elements[1];
						}
					}
				}

				for(int x = 0; x < ls.Blocks.Count - 3; ++x)
				{
					var lsSubSig = ls.GetSubSig(x, 4);
					var lsNewSubSig = lsSubSig.GetRandomEquivalent2PowCanonical(this);
					if(lsSubSig != lsNewSubSig)
					{
						int nOrgTaus = lsSubSig.Blocks.Count(blk => blk.Elements[1].Reflection);
						int nNewTaus = lsNewSubSig.Blocks.Count(blk => blk.Elements[1].Reflection);

						if(nOrgTaus >= nNewTaus)
						{
							for(int xc = 0; xc < 4; ++xc)
								ls.Blocks[x + xc].Elements[1] = lsNewSubSig.Blocks[xc].Elements[1];
						}
					}
				}
			}

			return ls;
		} */

		public DmBlockSeq<DmDihedralGroupElement> Simplify2PowCanonical(
			DmBlockSeq<DmDihedralGroupElement> lsSig)
		{
			if(!lsSig.Blocks.TrueForAll(blk => (blk.Elements.Count == 2)) ||
				!lsSig.Blocks.TrueForAll(blk => (blk.Elements[0].CompareTo(m_id) == 0)))
			{
				Debug.Assert(false);
				return lsSig;
			}

			var ls = lsSig.Clone();

			for(int x = 0; x < (ls.Blocks.Count - 1); ++x)
			{
				DmLogSigBlock<DmDihedralGroupElement> blkL = ls.Blocks[x];
				DmLogSigBlock<DmDihedralGroupElement> blkR = ls.Blocks[x + 1];
				DmDihedralGroupElement elL = blkL.Elements[1], elR = blkR.Elements[1];

				if(elL.Reflection)
				{
					if(elR.Reflection)
					{
						blkL.Elements[1] = new DmDihedralGroupElement(((elL.Rotation -
							elR.Rotation) + m_nRotMod) % m_nRotMod, false);
					}
					else // !blkR.Elements[1].Reflection
					{
						blkL.Elements[1] = elR;
						blkR.Elements[1] = new DmDihedralGroupElement(((elL.Rotation -
							elR.Rotation) + m_nRotMod) % m_nRotMod, true);
					}
				}
			}

			return ls;
		}

		public static DmBlockSeq<T> Simplify2PowCanonicalEx<T>(DmBlockSeq<T> lsSig,
			IDmGroup<T> grp)
			where T : IDmGroupElement<T>
		{
			if(!lsSig.Blocks.TrueForAll(blk => (blk.Elements.Count == 2)) ||
				!lsSig.Blocks.TrueForAll(blk => (blk.Elements[0].CompareTo(grp.Identity) == 0)))
			{
				Debug.Assert(false);
				return lsSig;
			}

			Type tElem = typeof(T);
			if(!tElem.Name.StartsWith(@"DmDirectProductGroupElement`")) return lsSig;

			Type[] vSubTypes = tElem.GetGenericArguments();
			if((vSubTypes == null) || (vSubTypes.Length != 2)) { Debug.Assert(false); return lsSig; }

			if(vSubTypes[0].Name != "DmDihedralGroupElement") return lsSig;

			var ls = lsSig.Clone();

			for(int x = 0; x < (ls.Blocks.Count - 1); ++x)
			{
				DmLogSigBlock<T> blkL = ls.Blocks[x], blkR = ls.Blocks[x + 1];
				T genL = blkL.Elements[1], genR = blkR.Elements[1];
				DmDihedralGroupElement elL = ((dynamic)genL).Value0;
				DmDihedralGroupElement elR = ((dynamic)genR).Value0;

				if(elL.Reflection)
				{
					if(elR.Reflection)
						blkL.Elements[1] = grp.Multiply(genL, genR);
					else // !blkR.Elements[1].Reflection
					{
						blkL.Elements[1] = genR;
						blkR.Elements[1] = grp.Multiply(genL, genR);
					}
				}
			}

			ls.Blocks.Sort(SortSimplifiedBy2Pow<T>);

#if DEBUG
			Dictionary<int, bool> dictRots = new Dictionary<int, bool>();
			for(int i = 0; i < (ls.Blocks.Count - 1); ++i)
			{
				T t = ls.Blocks[i].Elements[1];
				BigInteger biRot = ((dynamic)t).Value0.Rotation;
				dictRots[NumUtil.GetPow(biRot, 2)] = true;
			}
			dictRots[-1] = true;
			BigInteger biOrd = ((dynamic)grp).GroupU.Order;
			int nExp = NumUtil.GetPow(biOrd, 2);
			if(dictRots.Count != nExp)
				throw new InvalidOperationException("dictRots.Count != nExp.");
#endif

			return ls;
		}

		private static int SortSimplifiedBy2Pow<T>(DmLogSigBlock<T> x,
			DmLogSigBlock<T> y)
			where T : IDmGroupElement<T>
		{
			T genL = x.Elements[1], genR = y.Elements[1];
			DmDihedralGroupElement l = ((dynamic)genL).Value0;
			DmDihedralGroupElement r = ((dynamic)genR).Value0;

			if(l.Reflection) return 1;
			if(r.Reflection) return -1;

			int powL = NumUtil.GetPow(l.Rotation, 2), powR = NumUtil.GetPow(r.Rotation, 2);
			if(powL == powR) return l.Rotation.CompareTo(r.Rotation);
			return ((powL < powR) ? -1 : 1);
		}

		public DmBlockSeq<DmDihedralGroupElement> SimplifyUsingSwapsAndSelShifts(
			DmBlockSeq<DmDihedralGroupElement> lsBlockSeq)
		{
			if(lsBlockSeq == null) { Debug.Assert(false); return null; }

			var ls = lsBlockSeq.Clone();
			if(m_nRotMod.IsEven)
			{
				Debug.Assert(false); // Only odd supported
				return ls;
			}

			int x2 = ls.Blocks.FindIndex(blk => (blk.Elements.Count == 2));
			if(x2 < 0) { Debug.Assert(false); return ls; }

			UndoSelShiftAndMoveRec(ls, x2, ls.Blocks[x2].Elements[1], false);
			UndoSelShiftAndMoveRec(ls, x2, ls.Blocks[x2].Elements[1], true);

			return ls;
		}

		private int UndoSelShiftAndMoveRec(DmBlockSeq<DmDihedralGroupElement> ls,
			int x, DmDihedralGroupElement dt, bool bRight)
		{
			if(!bRight && (x <= 0)) return 0;
			if(bRight && (x >= (ls.Blocks.Count - 1))) return (ls.Blocks.Count - 1);

			int px = (bRight ? (x + 1) : (x - 1));
			List<DmDihedralGroupElement> l = ls.Blocks[px].Elements;

			for(int i = 0; i < l.Count; ++i)
			{
				if(l[i].Reflection)
					l[i] = (bRight ? Multiply(dt, l[i]) : Multiply(l[i], dt));
			}

			List<BigInteger> lZ = GetZRotOnly(l);
			if(lZ.Count == 0) return px;

			BigInteger z = lZ[0];
			if(lZ.Count > 1)
			{
				int iBest = x;

				foreach(BigInteger zCand in lZ)
				{
					var lsTest = ls.Clone();

					DmDihedralGroupElement dtTest = new DmDihedralGroupElement((dt.Rotation -
						zCand + m_nRotMod) % m_nRotMod, true);
					int iCurBest = UndoSelShiftAndMoveRec(lsTest, px, dtTest, bRight);

					if((bRight && (iCurBest > iBest)) || (!bRight && (iCurBest < iBest)))
					{
						z = zCand;
						iBest = iCurBest;
					}
				}
			}
			DmDihedralGroupElement dtNew = new DmDihedralGroupElement((dt.Rotation -
				z + m_nRotMod) % m_nRotMod, true);
			return UndoSelShiftAndMoveRec(ls, px, dtNew, bRight);
		}

		/* private int UndoSelShiftAndMoveRec(DmBlockSeq<DmDihedralGroupElement> ls,
			int x, DmDihedralGroupElement dt, bool bRight)
		{
			if(!bRight && (x <= 0)) return 0;
			if(bRight && (x >= (ls.Blocks.Count - 1))) return (ls.Blocks.Count - 1);

			for(int cntSpan = 1; ; ++cntSpan)
			{
				int px = (bRight ? (x + cntSpan) : (x - cntSpan));
				List<DmDihedralGroupElement> l;

				if(cntSpan > 1)
				{
					var vBlkSpan = ls.Blocks.Skip(bRight ? (x + 1) : px).Take(
						cntSpan).Select(blk => blk.Elements).ToArray();
					l = DmBlockSeq<DmDihedralGroupElement>.Multiply(this, vBlkSpan);
				}
				else l = ls.Blocks[px].Elements;

				for(int i = 0; i < l.Count; ++i)
				{
					if(l[i].Reflection)
						l[i] = (bRight ? Multiply(dt, l[i]) : Multiply(l[i], dt));
				}

				List<BigInteger> lZ = GetZ(l);
				if(lZ.Count == 0) return px;

				BigInteger z = lZ[0];
				if(lZ.Count > 1)
				{
					int iBest = x;

					foreach(BigInteger zCand in lZ)
					{
						var lsTest = ls.Clone();

						DmDihedralGroupElement dtTest = new DmDihedralGroupElement((dt.Rotation -
							zCand + m_nRotMod) % m_nRotMod, true);
						int iCurBest = UndoSelShiftAndMoveRec(lsTest, px, dtTest, bRight);

						if((bRight && (iCurBest > iBest)) || (!bRight && (iCurBest < iBest)))
						{
							z = zCand;
							iBest = iCurBest;
						}
					}
				}
				DmDihedralGroupElement dtNew = new DmDihedralGroupElement((dt.Rotation -
					z + m_nRotMod) % m_nRotMod, true);
				return UndoSelShiftAndMoveRec(ls, px, dtNew, bRight);
			}
		} */
	}
}
