﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;

namespace LogSig.DM.DihedralGroup
{
	/// <summary>
	/// Dihedral group (group of symmetries of a regular polygon).
	/// </summary>
	public sealed partial class DmDihedralGroup : IDmGroup<DmDihedralGroupElement>
	{
		public DmBlockSeq<DmDihedralGroupElement> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmDihedralGroupElement>();

			BigInteger n = m_nRotMod, p = 2, pp = 1;
			while(n != 1)
			{
				if((n % p).IsZero)
				{
					var blk = ls.AddBlock();
					for(int i = 0; i < p; ++i)
						blk.Elements.Add(new DmDihedralGroupElement(pp * i, false));

					pp *= p;
					n /= p;
					continue;
				}

				++p;
			}

			ls.AddBlock(m_id, m_tau);
			return ls;
		}

		/* public void GenerateRandom2PowSig(DmBlockSeq<DmDihedralGroupElement> ls)
		{
			GrsSetCanonicalSig(ls);

			int r = GrsTranslateTau(ls);
			GrsTranslate2Pows(ls);

			// GrsTauify(ls, r);

			ls.ShuffleBlocks();
			GrsApplyRandomTranslation(ls);
		}

		private static void GrsSetCanonicalSig(DmBlockSeq<DmDihedralGroupElement> ls)
		{
			ls.Blocks[0].Elements[0] = m_id;
			ls.Blocks[0].Elements[1] = m_tau;

			for(int i = 1; i < ls.Blocks.Count; ++i)
			{
				ls.Blocks[i].Elements[0] = m_id;
				ls.Blocks[i].Elements[1] = new DmDihedralGroupElement(
					1 << (i - 1), false);
			}
		}

		private void GrsTranslate2Pows(DmBlockSeq<DmDihedralGroupElement> ls)
		{
			for(int i = 1; i < ls.Blocks.Count; ++i)
			{
				int r = 2;
				while(NumUtil.GetPow(r, 2) >= 1)
				{
					r = Program.Random.Next(0, m_nRotMod);
					if((r & 1) == 0) ++r;
				}

				List<DmDihedralGroupElement> l = ls.Blocks[i].Elements;
				l[1] = new DmDihedralGroupElement((l[1].Rotation * r) % m_nRotMod, false);
			}
		}

		private int GrsTranslateTau(DmBlockSeq<DmDihedralGroupElement> ls)
		{
			int r = Program.Random.Next(0, m_nRotMod);
			ls.Blocks[0].Elements[1] = new DmDihedralGroupElement(r, true);
			return r;
		}

		private void GrsTauify(DmBlockSeq<DmDihedralGroupElement> ls, int nInvalid2Pow)
		{
			if(nInvalid2Pow < 1) return;

			for(int i = 1; i < ls.Blocks.Count; ++i)
			{
				if(Program.Random.Next(0, 2) == 0) continue;

				List<DmDihedralGroupElement> l = ls.Blocks[i].Elements;
				int p2 = NumUtil.GetPow(l[1].Rotation, 2);
				if(p2 == nInvalid2Pow) continue;

				l[1] = Multiply(l[1], m_tau);
			}
		}

		private void GrsApplyRandomTranslation(DmBlockSeq<DmDihedralGroupElement> ls)
		{
			DmDihedralGroupElement elTrans = GenerateRandomElement();
			List<DmDihedralGroupElement> lLast = ls.Blocks[ls.Blocks.Count - 1].Elements;
			lLast[0] = Multiply(lLast[0], elTrans);
			lLast[1] = Multiply(lLast[1], elTrans);
		} */

		public DmBlockSeq<DmDihedralGroupElement> CreateRandomSig(
			DmBlockSeq<DmDihedralGroupElement> lsType = null, bool bSort = false,
			CryptoRandomGen rng = null)
		{
			CryptoRandomGen r = (rng ?? Program.CryptoRandom);

			var ls = ((lsType != null) ? lsType.Clone() : CreateDefaultSig());
			if(ls.Blocks.Select(blk => new BigInteger(blk.Elements.Count)).Aggregate(
				(biProd, biCur) => (biProd * biCur)) != m_nOrder)
				throw new InvalidOperationException();
			if((m_nRotMod % 2) != 1)
				throw new InvalidOperationException(@"Only D_{2n} with n odd are supported by this method.");

			var lx2 = ls.Blocks.Where(blk => (blk.Elements.Count == 2)).ToList();
			if(lx2.Count != 1) throw new InvalidOperationException();
			var blkX2 = lx2[0];
			int x2 = ls.Blocks.IndexOf(blkX2);

			blkX2.Elements[0] = m_id;
			blkX2.Elements[1] = new DmDihedralGroupElement(r.GenerateBigInt(m_nRotMod), true);

			int xLastTauModInclL = 0;
			while(xLastTauModInclL < x2)
			{
				BigInteger biP = BigInteger.Pow(2, ls.Blocks[xLastTauModInclL].Elements.Count - 1);
				if(!r.GenerateBigInt(biP).IsZero) break;
				++xLastTauModInclL;
			}
			int xLastTauModInclR = ls.Blocks.Count - 1;
			while(xLastTauModInclR > x2)
			{
				BigInteger biP = BigInteger.Pow(2, ls.Blocks[xLastTauModInclR].Elements.Count - 1);
				if(!r.GenerateBigInt(biP).IsZero) break;
				--xLastTauModInclR;
			}

			var lRemBlkIdc = new List<int>(ls.Blocks.Count);
			for(int i = 0; i < ls.Blocks.Count; ++i)
			{
				if(i != x2) lRemBlkIdc.Add(i);
			}

			BigInteger biGrp = m_nRotMod;
			BigInteger biRotModDivGrp = m_nRotMod / biGrp;
			while(lRemBlkIdc.Count > 0)
			{
				int x = lRemBlkIdc[r.GenerateInt32(lRemBlkIdc.Count)];
				var blk = ls.Blocks[x].Elements;

				int cEl = blk.Count;
				BigInteger biSubGen = biGrp / cEl;

				blk[0] = m_id;

				if((x <= xLastTauModInclL) || (x >= xLastTauModInclR))
				{
					for(int i = 1; i < cEl; ++i)
					{
						BigInteger rElOffset = biGrp * r.GenerateBigInt(biRotModDivGrp);
						BigInteger rEl = ((biSubGen * i) + rElOffset) % m_nRotMod;

						blk[i] = new DmDihedralGroupElement(rEl, false);
					}
				}
				else
				{
					List<BigInteger> lSubEl = new List<BigInteger>(cEl - 1);
					BigInteger biSubElEnum = BigInteger.Zero;
					for(int i = 1; i < cEl; ++i)
					{
						biSubElEnum += biSubGen;
						lSubEl.Add(biSubElEnum);
					}

					int y = 1;

					BigInteger z = biSubGen * r.GenerateInt32(cEl);
					if(!z.IsZero)
					{
						Debug.Assert(z == (z % biGrp));
						if(!lSubEl.Remove(z))
						{
							Debug.Assert(false);
							continue;
						}

						BigInteger zOffset = biGrp * r.GenerateBigInt(biRotModDivGrp);
						z = (z + zOffset) % m_nRotMod;

						blk[1] = new DmDihedralGroupElement(z, false);
						// Correct element removed from lSubEl already, see above

						BigInteger zFixedPoint;
						if((z & 1).IsZero) zFixedPoint = z / 2;
						else zFixedPoint = (z * ((m_nRotMod + 1) / 2)) % m_nRotMod;
						Debug.Assert(zFixedPoint == ((z - zFixedPoint + m_nRotMod) % m_nRotMod));

						blk[2] = new DmDihedralGroupElement(zFixedPoint, false);
						if(!lSubEl.Remove(zFixedPoint % biGrp))
						{
							Debug.Assert(false);
							continue;
						}

						y += 2;
					}

					while(y < cEl)
					{
						BigInteger rEl = lSubEl[0];
						BigInteger rElOffset = biGrp * r.GenerateBigInt(biRotModDivGrp);
						rEl = (rEl + rElOffset) % m_nRotMod;

						blk[y] = new DmDihedralGroupElement(rEl, false);
						blk[y + 1] = new DmDihedralGroupElement((z - rEl + m_nRotMod) % m_nRotMod, false);

						BigInteger biOthClass = blk[y + 1].Rotation % biGrp;
						if(lSubEl[0] == biOthClass) { Debug.Assert(false); continue; }
						if(!lSubEl.Remove(biOthClass))
						{
							Debug.Assert(false);
							continue;
						}
						lSubEl.RemoveAt(0);

						y += 2;
					}
				}

				if(!lRemBlkIdc.Remove(x)) { Debug.Assert(false); }
				biGrp /= cEl;
				biRotModDivGrp = m_nRotMod / biGrp;
			}
			Debug.Assert(biGrp == BigInteger.One);

			MoveAndSelShiftRec(ls, x2, blkX2.Elements[1], r, false, xLastTauModInclL);
			MoveAndSelShiftRec(ls, x2, blkX2.Elements[1], r, true, xLastTauModInclR);

			SortOrShuffle(ref ls, r, bSort);
			return ls;
		}

		private static void SortOrShuffle(ref DmBlockSeq<DmDihedralGroupElement> ls,
			CryptoRandomGen r, bool bSort)
		{
			if(bSort) ls.Blocks.ForEach(blk => blk.Elements.Sort());
			else
			{
				var lsSh = new DmBlockSeq<DmDihedralGroupElement>();
				ls.Blocks.ForEach(blk =>
				{
					var blkSh = blk.Clone();
					Debug.Assert(blkSh.Elements[0].Equals(m_id));
					blkSh.Elements.RemoveAt(0);
					DmLogSigUtil.ShuffleList(blkSh.Elements, r);
					blkSh.Elements.Insert(0, m_id);
					lsSh.Blocks.Add(blkSh);
				});
				ls = lsSh;
			}
		}

		internal List<BigInteger> GetZRotOnly(List<DmDihedralGroupElement> lBlk)
		{
			if(!lBlk.TrueForAll(el => !el.Reflection))
			{
				Debug.Assert(false);
				return new List<BigInteger>();
			}

			List<DmDihedralGroupElement> lZ = DmLogSigUtil.GetZ<DmDihedralGroupElement>(
				lBlk, this);
			return lZ.Select(el => el.Rotation).ToList();
		}

		private void MoveAndSelShiftRec(DmBlockSeq<DmDihedralGroupElement> ls,
			int x, DmDihedralGroupElement dt, CryptoRandomGen r, bool bRight,
			int xLastTauModIncl)
		{
			if(!bRight && (x <= xLastTauModIncl)) return;
			if(bRight && (x >= xLastTauModIncl)) return;

			int px = (bRight ? (x + 1) : (x - 1));
			List<DmDihedralGroupElement> l = ls.Blocks[px].Elements;

			if((!bRight && (px > xLastTauModIncl)) || (bRight && (px < xLastTauModIncl)))
			{
				List<BigInteger> lZ = GetZRotOnly(l);
				Debug.Assert(lZ.Count >= 1); // By construction

				BigInteger z = lZ[r.GenerateInt32(lZ.Count)];
				BigInteger dNew = (((bRight ? (dt.Rotation - z) : (dt.Rotation + z)) +
					m_nRotMod) % m_nRotMod);
				DmDihedralGroupElement dtNew = new DmDihedralGroupElement(dNew, true);

#if DEBUG
				var blkX2 = new List<DmDihedralGroupElement>();
				blkX2.Add(this.Identity);
				blkX2.Add(dt);

				var blkPreL = (bRight ? blkX2 : l);
				var blkPreR = (bRight ? l : blkX2);
				var lGenPre = DmBlockSeq<DmDihedralGroupElement>.Multiply(this,
					blkPreL, blkPreR);

				var blkX2New = new List<DmDihedralGroupElement>();
				blkX2New.Add(this.Identity);
				blkX2New.Add(dtNew);

				var blkPostL = (bRight ? l : blkX2New);
				var blkPostR = (bRight ? blkX2New : l);
				var lGenPost = DmBlockSeq<DmDihedralGroupElement>.Multiply(this,
					blkPostL, blkPostR);

				Debug.Assert(lGenPre.Intersect(lGenPost).Count() == (l.Count * 2));
#endif

				MoveAndSelShiftRec(ls, px, dtNew, r, bRight, xLastTauModIncl);
			}

			for(int i = 1; i < l.Count; ++i)
			{
				if(r.GenerateBool())
					l[i] = (bRight ? Multiply(dt, l[i]) : Multiply(l[i], dt));
			}
		}
	}
}
