﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;

namespace LogSig.DM.DihedralGroup
{
	/// <summary>
	/// Dihedral group (group of symmetries of a regular polygon).
	/// </summary>
	public sealed partial class DmDihedralGroup : IDmGroup<DmDihedralGroupElement>
	{
		public string Name
		{
			get { return ("D2*" + NumUtil.FactorsToString(m_nRotMod, true)); }
		}
		// public string ShortName { get { return "Dih"; } }
		// public string ShortNameAlt { get { return "Dn"; } }

		private readonly BigInteger m_nOrder;
		private readonly BigInteger m_nRotMod; // = m_nOrder / 2
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_nOrder; }
		}

		private static readonly DmDihedralGroupElement m_id =
			new DmDihedralGroupElement(BigInteger.Zero, false);
		public DmDihedralGroupElement Identity
		{
			get { return m_id; }
		}

		private static readonly DmDihedralGroupElement m_tau =
			new DmDihedralGroupElement(BigInteger.Zero, true);

		public DmDihedralGroup(BigInteger nOrder)
		{
			if((nOrder < 2) || ((nOrder % 2) != 0))
				throw new ArgumentOutOfRangeException("nOrder");

			m_nOrder = nOrder;
			m_nRotMod = (nOrder / 2);
		}

		public DmDihedralGroupElement ParseElement(string str)
		{
			return DmDihedralGroupElement.Parse(str);
		}

		public DmDihedralGroupElement Multiply(DmDihedralGroupElement a,
			DmDihedralGroupElement b)
		{
			bool aRefl = a.Reflection, bRefl = b.Reflection;
			BigInteger aRot = a.Rotation, bRot = b.Rotation;

			BigInteger nRot = ((aRot + (aRefl ? (m_nRotMod - bRot) : bRot)) % m_nRotMod);
			bool nRefl = (aRefl ^ bRefl);

			return new DmDihedralGroupElement(nRot, nRefl);
		}

		public DmDihedralGroupElement Invert(DmDihedralGroupElement t)
		{
			if(t.Reflection) return t;

			if(t.Rotation.IsZero) return t;
			return new DmDihedralGroupElement(m_nRotMod - t.Rotation, false);
		}

		public DmDihedralGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			BigInteger nRot = rand.GenerateBigInt(m_nRotMod);
			bool bRefl = rand.GenerateBool();

			return new DmDihedralGroupElement(nRot, bRefl);
		}

		public DmDihedralGroupElement NextElement(DmDihedralGroupElement t)
		{
			BigInteger nNewRot = ((t.Rotation + 1) % m_nRotMod);
			if(nNewRot != 0) return new DmDihedralGroupElement(nNewRot, t.Reflection);

			return new DmDihedralGroupElement(0, !t.Reflection);
		}

		public BigInteger ElementToBigInt(DmDihedralGroupElement t)
		{
			return ((t.Rotation << 1) + (t.Reflection ? 1 : 0));
		}

		public DmDihedralGroupElement BigIntToElement(BigInteger biElem)
		{
			return new DmDihedralGroupElement(biElem >> 1, ((biElem & 1) != 0));
		}
	}
}
