﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;

using LogSig.Interfaces;

namespace LogSig.DM.DihedralGroup
{
	/// <summary>
	/// Immutable element of a dihedral group.
	/// </summary>
	public sealed class DmDihedralGroupElement : IDmGroupElement<DmDihedralGroupElement>
	{
		private int? m_hash = null;

		private readonly BigInteger m_biRot;
		public BigInteger Rotation
		{
			get { return m_biRot; }
			// set { m_biRot = value; } // Immutable
		}

		private readonly bool m_bRefl;
		public bool Reflection
		{
			get { return m_bRefl; }
			// set { m_bRefl = value; } // Immutable
		}

		public DmDihedralGroupElement()
		{
			m_biRot = 0;
			m_bRefl = false;
		}

		public DmDihedralGroupElement(BigInteger biRotation, bool bReflection)
		{
			m_biRot = biRotation;
			m_bRefl = bReflection;
		}

		public DmDihedralGroupElement Clone()
		{
			return (DmDihedralGroupElement)MemberwiseClone();
		}

		public override string ToString()
		{
			if(m_biRot == 0) return (m_bRefl ? "0t" : "id");

			string str = NumUtil.FactorsToString(m_biRot, true);
			return (str + (m_bRefl ? "t" : string.Empty));
		}

		public static DmDihedralGroupElement Parse(string str)
		{
			if(str == null) throw new ArgumentNullException("str");

			str = str.Trim().ToLower();
			if(str.Equals("id")) return new DmDihedralGroupElement(BigInteger.Zero, false);

			bool bRefl = false;
			if(str.EndsWith("t"))
			{
				bRefl = true;
				str = str.Substring(0, str.Length - 1);
			}

			return new DmDihedralGroupElement(NumUtil.Parse(str), bRefl);
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			int h = m_biRot.GetHashCode();
			if(m_bRefl) h ^= 0x5A70E69F;

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			DmDihedralGroupElement other = (obj as DmDihedralGroupElement);
			return Equals(other);
		}

		public bool Equals(DmDihedralGroupElement other)
		{
			if(other == null) return false;
			return ((m_bRefl == other.m_bRefl) && (m_biRot == other.m_biRot));
		}

		public int CompareTo(DmDihedralGroupElement other)
		{
			if(other == null) return 1;

			if(m_bRefl && !other.m_bRefl) return 1;
			if(!m_bRefl && other.m_bRefl) return -1;
			if(m_biRot == other.m_biRot) return 0;

			return ((m_biRot < other.m_biRot) ? -1 : 1);
		}
	}
}
