﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;

using LogSig.Crypto;

namespace LogSig.DM.DirectProductGroup
{
	public sealed class DmDirectProductGroup<U, V> :
		IDmGroup<DmDirectProductGroupElement<U, V>>
		where U : IDmGroupElement<U>
		where V : IDmGroupElement<V>
	{
		public string Name
		{
			get { return (m_u.Name + "x" + m_v.Name); }
		}
		// public string ShortName { get { return "DP"; } }
		// public string ShortNameAlt { get { return "x"; } }

		private readonly IDmGroup<U> m_u;
		private readonly IDmGroup<V> m_v;

		public IDmGroup<U> GroupU { get { return m_u; } }
		public IDmGroup<V> GroupV { get { return m_v; } }

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private readonly DmDirectProductGroupElement<U, V> m_id;
		public DmDirectProductGroupElement<U, V> Identity { get { return m_id; } }

		public DmDirectProductGroup(IDmGroup<U> u, IDmGroup<V> v)
		{
			if(u == null) throw new ArgumentNullException("u");
			if(v == null) throw new ArgumentNullException("v");

			m_u = u;
			m_v = v;
			m_biOrder = m_u.Order * m_v.Order;

			m_id = new DmDirectProductGroupElement<U, V>(m_u.Identity, m_v.Identity);
		}

		public DmDirectProductGroupElement<U, V> ParseElement(string str)
		{
			return DmDirectProductGroupElement<U, V>.Parse(str, m_u, m_v);
		}

		public DmDirectProductGroupElement<U, V> Multiply(DmDirectProductGroupElement<U, V> a,
			DmDirectProductGroupElement<U, V> b)
		{
			U u = m_u.Multiply(a.Value0, b.Value0);
			V v = m_v.Multiply(a.Value1, b.Value1);

			return new DmDirectProductGroupElement<U, V>(u, v);
		}

		public DmDirectProductGroupElement<U, V> Invert(DmDirectProductGroupElement<U, V> t)
		{
			U u = m_u.Invert(t.Value0);
			V v = m_v.Invert(t.Value1);

			return new DmDirectProductGroupElement<U, V>(u, v);
		}

		public DmDirectProductGroupElement<U, V> GenerateRandomElement(CryptoRandomGen rand)
		{
			U u = m_u.GenerateRandomElement(rand);
			V v = m_v.GenerateRandomElement(rand);

			return new DmDirectProductGroupElement<U, V>(u, v);
		}

		public DmDirectProductGroupElement<U, V> NextElement(DmDirectProductGroupElement<U, V> t)
		{
			V v = m_v.NextElement(t.Value1);
			U u = (v.Equals(m_v.Identity) ? m_u.NextElement(t.Value0) : t.Value0);

			return new DmDirectProductGroupElement<U, V>(u, v);
		}

		public BigInteger ElementToBigInt(DmDirectProductGroupElement<U, V> t)
		{
			BigInteger bi = m_u.ElementToBigInt(t.Value0);
			bi *= m_v.Order;
			bi += m_v.ElementToBigInt(t.Value1);

			return bi;
		}

		public DmDirectProductGroupElement<U, V> BigIntToElement(BigInteger biElem)
		{
			V v = m_v.BigIntToElement(biElem % m_v.Order);
			U u = m_u.BigIntToElement(biElem / m_v.Order);

			return new DmDirectProductGroupElement<U, V>(u, v);
		}

		public DmBlockSeq<DmDirectProductGroupElement<U, V>> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmDirectProductGroupElement<U, V>>();

			U idU = m_u.Identity;
			V idV = m_v.Identity;

			DmBlockSeq<U> lsU = m_u.CreateDefaultSig();
			foreach(DmLogSigBlock<U> blkU in lsU.Blocks)
			{
				var blk = ls.AddBlock();
				foreach(U u in blkU.Elements)
					blk.Elements.Add(new DmDirectProductGroupElement<U, V>(u, idV));
			}

			DmBlockSeq<V> lsV = m_v.CreateDefaultSig();
			foreach(DmLogSigBlock<V> blkV in lsV.Blocks)
			{
				var blk = ls.AddBlock();
				foreach(V v in blkV.Elements)
					blk.Elements.Add(new DmDirectProductGroupElement<U, V>(idU, v));
			}

			return ls;
		}

		public int[] FactorInDefaultSig(DmDirectProductGroupElement<U, V> t)
		{
			int[] vU = m_u.FactorInDefaultSig(t.Value0);
			int[] vV = m_v.FactorInDefaultSig(t.Value1);

			int[] v = new int[vU.Length + vV.Length];
			Array.Copy(vU, 0, v, 0, vU.Length);
			Array.Copy(vV, 0, v, vU.Length, vV.Length);
			return v;
		}
	}
}
