﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.Interfaces;

namespace LogSig.DM
{
	public sealed class DmBlockSeq<T> : IDeepCloneable<DmBlockSeq<T>>
		where T : IDmGroupElement<T>
	{
		private List<DmLogSigBlock<T>> m_vBlocks = new List<DmLogSigBlock<T>>();
		public List<DmLogSigBlock<T>> Blocks
		{
			get { return m_vBlocks; }
		}

		public DmBlockSeq<T> Clone()
		{
			DmBlockSeq<T> lsNew = new DmBlockSeq<T>();
			foreach(DmLogSigBlock<T> blk in m_vBlocks)
				lsNew.m_vBlocks.Add(blk.Clone());
			return lsNew;
		}

		public override string ToString()
		{
			return ToString(null);
		}

		public string ToString(int[] vFactorization)
		{
			string strInit = "(", strTerm = ")", strSep = ", ";
			if(Program.OutputFormat == OutputFmt.TeX)
			{
				strInit = string.Empty;
				strTerm = string.Empty;
				strSep = string.Empty;

				if(m_vBlocks.Count == 0) return "\\emptyset";
			}

			StringBuilder sb = new StringBuilder();
			sb.Append(strInit);

			for(int i = 0; i < m_vBlocks.Count; ++i)
			{
				if(i > 0) sb.Append(strSep);

				string strBlock;
				if((vFactorization == null) || (Program.OutputFormat != OutputFmt.TeX))
					strBlock = m_vBlocks[i].ToString();
				else strBlock = m_vBlocks[i].ToString(vFactorization[i]);

				sb.Append(strBlock);
			}

			sb.Append(strTerm);
			return sb.ToString();
		}

		public string GetTypeString()
		{
			StringBuilder sb = new StringBuilder();
			sb.Append('(');

			for(int i = 0; i < m_vBlocks.Count; ++i)
			{
				if(sb.Length > 1) sb.Append(", ");
				sb.Append(m_vBlocks[i].Elements.Count.ToString());
			}

			sb.Append(')');
			return sb.ToString();
		}

		public static DmBlockSeq<T> Parse(string str, IDmGroup<T> grp)
		{
			if(str == null) throw new ArgumentNullException("str");

			str = str.Trim();
			if(!str.StartsWith("(") || !str.EndsWith(")"))
				throw new FormatException("Can't parse block!");

			str = str.Substring(1, str.Length - 2);

			DmBlockSeq<T> ls = new DmBlockSeq<T>();

			StringBuilder sbElem = new StringBuilder();
			int nLevel = 0;
			for(int i = 0; i < str.Length; ++i)
			{
				char ch = str[i];

				if((ch == ',') && (nLevel == 0))
				{
					ls.m_vBlocks.Add(DmLogSigBlock<T>.Parse(sbElem.ToString().Trim(), grp));
					sbElem.Clear();

					if(((i + 1) < str.Length) && (str[i + 1] == ' ')) ++i;
				}
				else sbElem.Append(ch);

				if(ch == '(') ++nLevel;
				else if(ch == ')') --nLevel;
			}

			if(sbElem.Length > 0)
				ls.m_vBlocks.Add(DmLogSigBlock<T>.Parse(sbElem.ToString().Trim(), grp));

			return ls;
		}

		public DmLogSigBlock<T> AddBlock(params T[] v)
		{
			DmLogSigBlock<T> blk = new DmLogSigBlock<T>();

			foreach(T t in v) blk.Elements.Add(t);

			m_vBlocks.Add(blk);
			return blk;
		}

		public void SwapBlocks(int i, int j)
		{
			if((i < 0) || (i >= m_vBlocks.Count))
				throw new ArgumentOutOfRangeException("i");
			if((j < 0) || (j >= m_vBlocks.Count))
				throw new ArgumentOutOfRangeException("j");

			DmLogSigBlock<T> blkI = m_vBlocks[i];
			m_vBlocks[i] = m_vBlocks[j];
			m_vBlocks[j] = blkI;
		}

		public T MultiplyElements(List<int> vIndices, IDmGroup<T> dg)
		{
			if(vIndices == null) throw new ArgumentNullException("vIndices");
			if(vIndices.Count != m_vBlocks.Count)
				throw new ArgumentOutOfRangeException("vIndices");

			T tCur = dg.Identity;
			for(int i = 0; i < vIndices.Count; ++i)
			{
				int nIndex = vIndices[i];
				if(nIndex >= 0)
					tCur = dg.Multiply(tCur, m_vBlocks[i].Elements[nIndex]);
			}

			return tCur;
		}

		public static List<T> Multiply(IDmGroup<T> dg, params List<T>[] vBlocks)
		{
			List<T> v = new List<T>();
			MultiplyRec(vBlocks, 0, dg.Identity, dg, v);
			return v;
		}

		private static void MultiplyRec(List<T>[] vBlocks, int x, T tCur,
			IDmGroup<T> dg, List<T> vStorage)
		{
			if(x >= vBlocks.Length) { vStorage.Add(tCur); return; }

			for(int y = 0; y < vBlocks[x].Count; ++y)
			{
				T tNew = dg.Multiply(tCur, vBlocks[x][y]);
				MultiplyRec(vBlocks, x + 1, tNew, dg, vStorage);
			}
		}

		public DmBlockSeq<T> GetSubSig(int x, int nBlockCount)
		{
			if(x < 0) throw new ArgumentOutOfRangeException("x");
			if(nBlockCount < 0) throw new ArgumentOutOfRangeException("nBlockCount");
			if((x + nBlockCount) > m_vBlocks.Count)
				throw new ArgumentOutOfRangeException("x + nBlockCount > BlockCount");

			DmBlockSeq<T> ls = new DmBlockSeq<T>();
			for(int i = 0; i < nBlockCount; ++i)
				ls.Blocks.Add(m_vBlocks[x + i]);

			return ls;
		}

		/* public DmBlockSeq<T> GetRandomEquivalent2PowCanonical(IDmGroup<T> dg)
		{
			if(!m_vBlocks.TrueForAll(blk => (blk.Elements.Count == 2)))
			{
				Debug.Assert(false);
				return this;
			}

			List<List<T>> vBlocks = new List<List<T>>(m_vBlocks.Count);
			foreach(var blk in m_vBlocks) vBlocks.Add(blk.Elements);

			List<T> vOrgProd = Multiply(dg, vBlocks.ToArray());
			Debug.Assert(vOrgProd.Count == (1 << vBlocks.Count));

			List<T> vReorder = new List<T>(vOrgProd);
			DmLogSigUtil.ShuffleList(vReorder);

			DmBlockSeq<T> lsNew = new DmBlockSeq<T>();
			foreach(T tElem in vReorder.Take(m_vBlocks.Count))
				lsNew.AddBlock(dg.Identity, tElem);
			List<List<T>> vNewBlocks = new List<List<T>>(lsNew.Blocks.Count);
			foreach(var blk in lsNew.Blocks) vNewBlocks.Add(blk.Elements);

			var vNewProd = DmBlockSeq<T>.Multiply(dg, vNewBlocks.ToArray());

			if(vOrgProd.Intersect(vNewProd, dg).Count() == vOrgProd.Count)
				return lsNew;

			return this;
		} */

		/* public DmBlockSeq<T> GetRandomEquivalentCanonical(IDmGroup<T> dg,
			CryptoRandomGen rand)
		{
			List<List<T>> vBlocks = new List<List<T>>(m_vBlocks.Count);
			foreach(var blk in m_vBlocks) vBlocks.Add(blk.Elements);

			List<T> vOrgProd = Multiply(dg, vBlocks.ToArray());

			List<T> vReorder = new List<T>(vOrgProd);
			DmLogSigUtil.ShuffleList(vReorder, rand);

			DmBlockSeq<T> lsNew = new DmBlockSeq<T>();
			int p = 0;
			for(int x = 0; x < vBlocks.Count; ++x)
			{
				List<T> blkNew = new List<T>();
				blkNew.Add(dg.Identity);

				for(int y = 1; y < vBlocks[x].Count; ++y)
				{
					blkNew.Add(vReorder[p]);
					++p;
				}

				lsNew.AddBlock(blkNew.ToArray());
			}
			DmLogSigUtil.ShuffleList(lsNew.Blocks, rand);

			List<List<T>> vNewBlocks = new List<List<T>>(lsNew.Blocks.Count);
			foreach(var blk in lsNew.Blocks) vNewBlocks.Add(blk.Elements);

			var vNewProd = DmBlockSeq<T>.Multiply(dg, vNewBlocks.ToArray());

			if(vOrgProd.Intersect(vNewProd).Count() == vOrgProd.Count)
				return lsNew;

			return this;
		} */

		public DmBlockSeq<T> GetRandomEquivalentCanonical(IDmGroup<T> grp,
			CryptoRandomGen rand)
		{
			List<List<T>> lOrgBlocks = new List<List<T>>(m_vBlocks.Count);
			foreach(var blk in m_vBlocks) lOrgBlocks.Add(blk.Elements);

			DmBlockSeq<T> lsNew = new DmBlockSeq<T>();
			for(int x = 0; x < lOrgBlocks.Count; ++x)
			{
				List<T> blkNew = new List<T>();
				for(int y = 0; y < lOrgBlocks[x].Count; ++y)
					blkNew.Add(grp.Identity);
				lsNew.AddBlock(blkNew.ToArray());
			}

			List<T> lGenLeft = new List<T>();
			lGenLeft.Add(grp.Identity);

			List<T> lOrgProd = Multiply(grp, lOrgBlocks.ToArray());
			List<T> lFree = new List<T>(lOrgProd);
			if(!lFree.Remove(grp.Identity)) throw new InvalidOperationException();
			DmLogSigUtil.ShuffleList(lFree, rand);

			while(true)
			{
				DmLogSigUtil.ShuffleList(lsNew.Blocks, rand); // Permute blocks

				if(GetRandomEquivalentCanonical(lsNew, 0, 0, lGenLeft,
					lFree, grp, rand))
					break;
			}

#if DEBUG
			List<List<T>> lNewBlocks = new List<List<T>>(lsNew.Blocks.Count);
			foreach(var blk in lsNew.Blocks) lNewBlocks.Add(blk.Elements);
			List<T> lNewProd = Multiply(grp, lNewBlocks.ToArray());
			Debug.Assert((lNewProd.Count == lOrgProd.Count) &&
				(lNewProd.Intersect(lOrgProd).Count() == lOrgProd.Count));
#endif
			return lsNew;
		}

		private static bool GetRandomEquivalentCanonical(DmBlockSeq<T> ls, int x,
			int y, List<T> lGenLeft, List<T> lFree, IDmGroup<T> grp,
			CryptoRandomGen rand)
		{
			if(x == ls.Blocks.Count)
			{
				Debug.Assert(y == 0);
				Debug.Assert(lGenLeft.Count == ls.Blocks.Aggregate(1,
					(acc, blk) => acc * blk.Elements.Count));
				Debug.Assert(lFree.Count == 0);
				return true; // Blocks generate the set
			}

			if(y == 0)
			{
				Debug.Assert(ls.Blocks[x].Elements[0].Equals(grp.Identity));
				return GetRandomEquivalentCanonical(ls, x, y + 1, lGenLeft,
					lFree, grp, rand);
			}

			foreach(T tCand in lFree)
			{
				// Enforce blocks being ordered, to avoid unnecessary
				// checks, e.g. if (0, 1, 3) is not a valid block then
				// neither (0, 3, 1) is
				if(tCand.CompareTo(ls.Blocks[x].Elements[y - 1]) < 0) continue;

				ls.Blocks[x].Elements[y] = tCand;

				List<T> lNewFree = new List<T>(lFree);
				bool bCandValid = true;
				foreach(T tGenLeft in lGenLeft)
				{
					Debug.Assert(lNewFree.Count > 0);
					T tNewGen = grp.Multiply(tGenLeft, tCand);
					int iFreeIdx = lNewFree.IndexOf(tNewGen);
					if(iFreeIdx < 0) { bCandValid = false; break; }
					lNewFree.RemoveAt(iFreeIdx);
				}
				if(!bCandValid) continue;

				if(y == (ls.Blocks[x].Elements.Count - 1)) // Last el. in block
				{
					List<T> lNewGenLeft = new List<T>();
					foreach(T tBlkEl in ls.Blocks[x].Elements)
					{
						foreach(T tLeft in lGenLeft)
						{
							lNewGenLeft.Add(grp.Multiply(tLeft, tBlkEl));
						}
					}
					Debug.Assert(lNewGenLeft.Distinct().Count() ==
						(lGenLeft.Count * ls.Blocks[x].Elements.Count));
					Debug.Assert(lNewGenLeft.Intersect(lGenLeft).Count() ==
						lGenLeft.Count);
					Debug.Assert(lNewGenLeft.Intersect(lNewFree).Count() == 0);

					if(GetRandomEquivalentCanonical(ls, x + 1, 0, lNewGenLeft,
						lNewFree, grp, rand))
						return true;
				}
				else
				{
					if(GetRandomEquivalentCanonical(ls, x, y + 1, lGenLeft,
						lNewFree, grp, rand))
						return true;
				}
			}

			return false;
		}

		public static DmBlockSeq<T> CreateEmpty(IDmGroup<T> dg)
		{
			DmBlockSeq<T> ls = new DmBlockSeq<T>();
			BigInteger n = dg.Order;

			BigInteger p = 2;
			while(n != 1)
			{
				if((n % p).IsZero)
				{
					DmLogSigBlock<T> blk = new DmLogSigBlock<T>();
					for(BigInteger i = 0; i < p; ++i)
						blk.Elements.Add(dg.Identity);
					ls.Blocks.Add(blk);

					n /= p;
				}
				else ++p;
			}

			return ls;
		}

		public static DmBlockSeq<T> ConvertFrom<S>(DmBlockSeq<S> ls,
			Func<S, T> fConvert)
			where S : IDmGroupElement<S>
		{
			if(ls == null) { Debug.Assert(false); return null; }
			if(fConvert == null) throw new ArgumentNullException("fConvert");

			var lsNew = new DmBlockSeq<T>();
			foreach(var blk in ls.Blocks)
			{
				var blkNew = new DmLogSigBlock<T>();
				foreach(var elS in blk.Elements)
				{
					T t = fConvert(elS);
					if(t == null) throw new InvalidCastException();

					blkNew.Elements.Add(t);
				}

				lsNew.Blocks.Add(blkNew);
			}

			return lsNew;
		}
	}
}
