﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Numerics;

using LogSig.AppModules;
using LogSig.DM.DihedralGroup;
using LogSig.DM.SnGroup;
using LogSig.DM.AnGroup;
using LogSig.DM.ZnGroup;
using LogSig.DM.QuaternionGroup;
using LogSig.DM.MultiZnGroup;
using LogSig.DM.DirectProductGroup;
using LogSig.DM.WreathProductGroup;

namespace LogSig.DM
{
	public sealed class DmGroupFactoryPool
	{
		private List<IDmGroupFactory> m_vProv = new List<IDmGroupFactory>();

		public DmGroupFactoryPool()
		{
			m_vProv.Add(new DmDihedralGroupFactory());
			m_vProv.Add(new DmSnGroupFactory());
			m_vProv.Add(new DmAnGroupFactory());
			m_vProv.Add(new DmZnGroupFactory());
			m_vProv.Add(new DmQuaternionGroupFactory());

			m_vProv.Add(new DmMultiZnGroupFactory());
			m_vProv.Add(new DmDirectProductGroupFactory());
			m_vProv.Add(new DmWreathProductGroupFactory());
		}

		public void Add(IDmGroupFactory prov)
		{
			if(prov == null) throw new ArgumentNullException("prov");

			m_vProv.Add(prov);
		}

		public object ConstructGroup(string strGroup)
		{
			strGroup = strGroup.Trim();
			foreach(IDmGroupFactory prov in m_vProv)
			{
				object o = prov.CreateInstance(strGroup);
				if(o != null) return o;
			}

			return null;
		}

		public void InvokeGroupMethod(AppModEventArgs e, Type tStaticInvokable,
			params object[] vParams)
		{
			InvokeGroupMethod(e.Args["g"], tStaticInvokable, vParams);
		}

		public void InvokeGroupMethod(string strGroup, Type tStaticInvokable,
			params object[] vParams)
		{
			if(strGroup == null) throw new ArgumentNullException("strGroup");
			if(tStaticInvokable == null) throw new ArgumentNullException("tStaticInvokable");

			object grp = ConstructGroup(strGroup);
			if(grp == null)
				throw new InvalidOperationException("No group specified or group unknown.");

			MethodInfo miGen = tStaticInvokable.GetMethod("Invoke",
				BindingFlags.Static | BindingFlags.Public | BindingFlags.NonPublic |
				BindingFlags.InvokeMethod);
			if(miGen == null) throw new InvalidCastException();

			Type tGrp = grp.GetType();
			Type tElem = GetGroupElementType(tGrp);
			if(tElem == null) throw new InvalidCastException();

			MethodInfo miFn = miGen.MakeGenericMethod(tElem);

			int nCustomParams = ((vParams != null) ? vParams.Length : 0);
			object[] vAllParams = new object[nCustomParams + 1];
			vAllParams[0] = grp;
			Array.Copy(vParams, 0, vAllParams, 1, vParams.Length);

			miFn.Invoke(null, vAllParams);
		}

		internal static Type GetGroupElementType(Type tGrp)
		{
			Type[] vInterfaces = tGrp.GetInterfaces();

			foreach(Type tInterface in vInterfaces)
			{
				if(!tInterface.IsGenericType) continue;
				if(!tInterface.Name.StartsWith(@"IDmGroup`")) continue;

				Type[] vGen = tInterface.GetGenericArguments();
				if((vGen == null) || (vGen.Length != 1)) continue;

				Type[] vSub = vGen[0].GetInterfaces();
				foreach(Type tSubInterface in vSub)
				{
					if(!tSubInterface.IsGenericType) continue;
					if(!tSubInterface.Name.StartsWith(@"IDmGroupElement`")) continue;

					return vGen[0];
				}
			}

			return null;
		}
	}
}
