﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using LogSig.Interfaces;

namespace LogSig.DM
{
	public sealed class DmLogSigBlock<T> : IDeepCloneable<DmLogSigBlock<T>>
		where T : IDmGroupElement<T>
	{
		private List<T> m_vElements = new List<T>();
		public List<T> Elements
		{
			get { return m_vElements; }
		}

		public DmLogSigBlock<T> Clone()
		{
			DmLogSigBlock<T> lsNew = new DmLogSigBlock<T>();
			foreach(T t in m_vElements) lsNew.m_vElements.Add(t.Clone());
			return lsNew;
		}

		public override string ToString()
		{
			return ToString(-1);
		}

		public string ToString(int iSelectIndex)
		{
			string strInit = "(", strTerm = ")", strSep = ", ";
			if(Program.OutputFormat == OutputFmt.TeX)
			{
				strInit = "\\begin{pmatrix} ";
				strTerm = " \\end{pmatrix}" + Environment.NewLine;
				strSep = " \\\\ ";
			}

			StringBuilder sb = new StringBuilder();
			sb.Append(strInit);

			for(int i = 0; i < m_vElements.Count; ++i)
			{
				if(i > 0) sb.Append(strSep);

				if((Program.OutputFormat == OutputFmt.TeX) && (i == iSelectIndex))
					sb.Append("\\lsbox{");

				sb.Append(m_vElements[i].ToString());

				if((Program.OutputFormat == OutputFmt.TeX) && (i == iSelectIndex))
					sb.Append("}");
			}

			sb.Append(strTerm);
			return sb.ToString();
		}

		public static DmLogSigBlock<T> Parse(string str, IDmGroup<T> grp)
		{
			if(str == null) throw new ArgumentNullException("str");

			if(!str.StartsWith("(") || !str.EndsWith(")"))
				throw new FormatException("Cannot parse block!");

			str = str.Substring(1, str.Length - 2);

			DmLogSigBlock<T> blk = new DmLogSigBlock<T>();

			StringBuilder sbElem = new StringBuilder();
			int nLevel = 0;
			for(int i = 0; i < str.Length; ++i)
			{
				char ch = str[i];

				if((ch == ',') && (nLevel == 0))
				{
					blk.m_vElements.Add(grp.ParseElement(sbElem.ToString().Trim()));
					sbElem.Clear();

					if(((i + 1) < str.Length) && (str[i + 1] == ' ')) ++i;
				}
				else sbElem.Append(ch);

				if(ch == '(') ++nLevel;
				else if(ch == ')') --nLevel;
			}

			if(sbElem.Length > 0)
				blk.m_vElements.Add(grp.ParseElement(sbElem.ToString().Trim()));

			return blk;
		}

		public static DmLogSigBlock<T> Fusion(DmLogSigBlock<T> l, DmLogSigBlock<T> r,
			IDmGroup<T> dmGrp)
		{
			DmLogSigBlock<T> blk = new DmLogSigBlock<T>();
			List<T> vElements = blk.m_vElements;

			foreach(T tL in l.m_vElements)
			{
				foreach(T tR in r.m_vElements)
				{
					vElements.Add(dmGrp.Multiply(tL, tR));
				}
			}

			return blk;
		}
	}
}
