﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Diagnostics;
using System.Numerics;

using LogSig.Crypto;
using LogSig.Interfaces;

namespace LogSig.DM
{
	public delegate bool DmLogSigDelegate<T>(DmBlockSeq<T> ls)
		where T : IDmGroupElement<T>;

	public static class DmLogSigUtil
	{
		public static DmBlockSeq<T> GetLStarCanonical<T>(DmBlockSeq<T> dmSig,
			IDmGroup<T> dmGrp)
			where T : IDmGroupElement<T>
		{
			return GetStarCanonical<T>(dmSig, dmGrp, 1);
		}

		public static DmBlockSeq<T> GetRStarCanonical<T>(DmBlockSeq<T> dmSig,
			IDmGroup<T> dmGrp)
			where T : IDmGroupElement<T>
		{
			return GetStarCanonical<T>(dmSig, dmGrp, -1);
		}

		private static DmBlockSeq<T> GetStarCanonical<T>(DmBlockSeq<T> dmSig,
			IDmGroup<T> dmGrp, int iStep)
			where T : IDmGroupElement<T>
		{
			if(dmSig == null) throw new ArgumentNullException("dmSig");
			if(dmGrp == null) throw new ArgumentNullException("dmGrp");

			if(dmSig.Blocks.Count == 0) return new DmBlockSeq<T>();

			DmBlockSeq<T> ls = dmSig.Clone();
			int n = ls.Blocks.Count;

			DmLogSigBlock<T> blkNew = new DmLogSigBlock<T>();
			blkNew.Elements.Add(dmGrp.Identity);
			if(iStep > 0) ls.Blocks.Add(blkNew);
			else ls.Blocks.Insert(0, blkNew);

			int iStart = ((iStep > 0) ? 0 : n); // New R block has index n
			int iExcl = ((iStep > 0) ? n : 0);

			for(int iBlock = iStart; iBlock != iExcl; iBlock += iStep)
			{
				List<T> vElements = ls.Blocks[iBlock].Elements;
				if(vElements.Count == 0)
					throw new InvalidOperationException();

				T t = vElements[0];
				T tInv = dmGrp.Invert(t);

				for(int i = 0; i < vElements.Count; ++i)
				{
					if(iStep > 0)
						vElements[i] = dmGrp.Multiply(vElements[i], tInv);
					else
						vElements[i] = dmGrp.Multiply(tInv, vElements[i]);
				}

				List<T> vElementsNext = ls.Blocks[iBlock + iStep].Elements;
				for(int j = 0; j < vElementsNext.Count; ++j)
				{
					if(iStep > 0)
						vElementsNext[j] = dmGrp.Multiply(t, vElementsNext[j]);
					else
						vElementsNext[j] = dmGrp.Multiply(vElementsNext[j], t);
				}
			}

			return ls;
		}

		public static bool GeneratesWholeGroup<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp)
			where T : IDmGroupElement<T>
		{
			if(ls == null) throw new ArgumentNullException("ls");
			if(ls.Blocks.Count == 0) return false;
			if(dmGrp == null) throw new ArgumentNullException("dmGrp");

			if(ls.Blocks.Aggregate(1UL, (uProd, blkCur) => (uProd *
				(ulong)blkCur.Elements.Count)) != dmGrp.Order) return false;

			HashSet<T> hsGen = new HashSet<T>();
			return GenerateRec(ls, dmGrp, 0, dmGrp.Identity, hsGen);
		}

		private static bool GenerateRec<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp,
			int iBlock, T elCur, HashSet<T> hsGen)
			where T : IDmGroupElement<T>
		{
			bool bIsLastBlock = (iBlock == (ls.Blocks.Count - 1));

			foreach(T el in ls.Blocks[iBlock].Elements)
			{
				T elGen = dmGrp.Multiply(elCur, el);
				if(bIsLastBlock)
				{
					if(hsGen.Contains(elGen)) return false;

					hsGen.Add(elGen);
				}
				else
				{
					if(!GenerateRec(ls, dmGrp, iBlock + 1, elGen, hsGen))
						return false;
				}
			}

			return true;
		}

		public static void Randomize<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp,
			CryptoRandomGen rand)
			where T : IDmGroupElement<T>
		{
			for(int iBlock = 0; iBlock < ls.Blocks.Count; ++iBlock)
			{
				List<T> vElements = ls.Blocks[iBlock].Elements;
				for(int iElem = 0; iElem < vElements.Count; ++iElem)
					vElements[iElem] = dmGrp.GenerateRandomElement(rand);
			}
		}

		public static void RandomizeCanonical<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp,
			CryptoRandomGen rand)
			where T : IDmGroupElement<T>
		{
			for(int iBlock = 0; iBlock < ls.Blocks.Count; ++iBlock)
			{
				List<T> vElements = ls.Blocks[iBlock].Elements;

				vElements[0] = dmGrp.Identity;
				for(int iElem = 1; iElem < vElements.Count; ++iElem)
					vElements[iElem] = dmGrp.GenerateRandomElement(rand);
			}
		}

		public static bool GeneratesDuplicatesCanonical<T>(DmBlockSeq<T> ls,
			IDmGroup<T> dmGrp, int nBlocks)
			where T : IDmGroupElement<T>
		{
			SortedDictionary<T, bool> dictGen = new SortedDictionary<T, bool>();
			dictGen.Add(dmGrp.Identity, true);

			return GeneratesDupCncRec(ls, dmGrp, 0, nBlocks, dictGen);
		}

		private static bool GeneratesDupCncRec<T>(DmBlockSeq<T> ls,
			IDmGroup<T> dmGrp, int iOffset, int nBlocks,
			SortedDictionary<T, bool> dictGen)
			where T : IDmGroupElement<T>
		{
			if(iOffset == nBlocks)
			{
				if(nBlocks == ls.Blocks.Count)
				{
					Debug.Assert(dictGen.Count == dmGrp.Order);
				}
				return false;
			}

			List<T> vElements = ls.Blocks[iOffset].Elements;
			Debug.Assert(vElements.Count >= 1);
			Debug.Assert(vElements[0].CompareTo(dmGrp.Identity) == 0);

			SortedDictionary<T, bool> dictIn = new SortedDictionary<T, bool>(dictGen);

			for(int i = 1; i < vElements.Count; ++i)
			{
				T tRight = vElements[i];

				List<T> vNew = new List<T>(dictIn.Count + 1);
				Debug.Assert(vNew.Count == 0);
				foreach(KeyValuePair<T, bool> kvp in dictIn)
					vNew.Add(dmGrp.Multiply(kvp.Key, tRight));

				foreach(T tAdd in vNew)
				{
					if(dictGen.ContainsKey(tAdd)) return true;
					dictGen.Add(tAdd, true);
				}
			}

			return GeneratesDupCncRec(ls, dmGrp, iOffset + 1, nBlocks, dictGen);
		}

		public static void RandomizeCanonicalEnsureGenerating<T>(DmBlockSeq<T> ls,
			IDmGroup<T> dmGrp, CryptoRandomGen rand)
			where T : IDmGroupElement<T>
		{
			RandomizeEnsureGeneratingRec(ls, dmGrp, 0, rand);
		}

		private static bool RandomizeEnsureGeneratingRec<T>(DmBlockSeq<T> ls,
			IDmGroup<T> dmGrp, int iBlock, CryptoRandomGen rand)
			where T : IDmGroupElement<T>
		{
			if(iBlock == ls.Blocks.Count) return true;

			List<T> vElements = ls.Blocks[iBlock].Elements;
			Debug.Assert(vElements.Count >= 1);
			Debug.Assert(vElements[0].CompareTo(dmGrp.Identity) == 0);

			List<T> vInitRnd = new List<T>();
			vInitRnd.Add(dmGrp.Identity);
			for(int iRnd = 1; iRnd < vElements.Count; ++iRnd)
			{
				vInitRnd.Add(dmGrp.GenerateRandomElement(rand));
				vElements[iRnd] = vInitRnd[iRnd];
			}

			bool bTriedAll = false;
			while(!bTriedAll)
			{
				if(!GeneratesDuplicatesCanonical(ls, dmGrp, iBlock + 1))
				{
					if(RandomizeEnsureGeneratingRec(ls, dmGrp, iBlock + 1, rand))
						return true;
				}

				for(int ix = (vElements.Count - 1); ix >= 0; --ix)
				{
					if(ix == 0) { bTriedAll = true; break; }

					vElements[ix] = dmGrp.NextElement(vElements[ix]);
					if(vElements[ix].CompareTo(vInitRnd[ix]) != 0) break;
				}
			}

			return false;
		}

		public static bool TrivialFactorize<T>(DmBlockSeq<T> ls, T elToFactor,
			T elCur, int iBlock, IDmGroup<T> grp, int[] vFactorization)
			where T : IDmGroupElement<T>
		{
			if(iBlock == ls.Blocks.Count) return (elCur.CompareTo(elToFactor) == 0);

			for(int iElem = 0; iElem < ls.Blocks[iBlock].Elements.Count; ++iElem)
			{
				vFactorization[iBlock] = iElem;

				if(TrivialFactorize(ls, elToFactor, grp.Multiply(elCur,
					ls.Blocks[iBlock].Elements[iElem]), iBlock + 1, grp, vFactorization))
					return true;
			}

			return false;
		}

		public static void ShuffleList<T>(List<T> v, CryptoRandomGen rand,
			int nItems = -1)
		{
			if(v == null) throw new ArgumentNullException("v");

			int nStart = ((nItems >= 0) ? nItems : v.Count);
			for(int x = (nStart - 1); x > 0; --x)
			{
				int r = rand.GenerateInt32(x + 1);

				T t = v[x];
				v[x] = v[r];
				v[r] = t;
			}
		}

		public static HashSet<T> GetAllElements<T>(IDmGroup<T> dmGrp)
			where T : IDmGroupElement<T>
		{
			var s = new HashSet<T>();

			T tId = dmGrp.Identity, t = dmGrp.Identity;
			do
			{
				s.Add(t);
				t = dmGrp.NextElement(t);
			}
			while(t.CompareTo(tId) != 0);

			return s;
		}

		/* private enum DmElementState : byte
		{
			ToGen,
			GenL,
			GenR,
			Excl
		}

		public static bool RandomizeCanonicalEx<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp)
			where T : IDmGroupElement<T>
		{
			DmElementState[] v = new DmElementState[(int)dmGrp.Order];
			for(int i = 0; i < v.Length; ++i) v[i] = DmElementState.ToGen;

			v[(int)dmGrp.ElementToBigInt(dmGrp.Identity)] = DmElementState.GenL;

			return RandomizeCanonicalExRec(ls, dmGrp, 0, 0, v);
		}

		private static bool RandomizeCanonicalExRec<T>(DmBlockSeq<T> ls, IDmGroup<T> dmGrp,
			int x, int y, DmElementState[] vStates)
			where T : IDmGroupElement<T>
		{
			if(x == ls.Blocks.Count) return GeneratesWholeGroup(ls, dmGrp);
			if(y == ls.Blocks[x].Elements.Count)
			{
				int gId = (int)dmGrp.ElementToBigInt(dmGrp.Identity);

				DmElementState[] vNewStates = new DmElementState[vStates.Length];
				Array.Copy(vStates, vNewStates, vStates.Length);

				Debug.Assert(vStates.Count(st => (st == DmElementState.GenL)) ==
					ls.Blocks.Take(x).Aggregate(1, (acc, blk) =>
						(acc * blk.Elements.Count)));

				// Update left generated set of elements
				for(int g1 = 0; g1 < vNewStates.Length; ++g1)
				{
					if(vStates[g1] != DmElementState.GenL) continue;

					for(int j = 1; j < ls.Blocks[x].Elements.Count; ++j)
					{
						int gp = dmGrp.Multiply(g1, dmGrp.ElementToBigInt(
							ls.Blocks[x].Elements[j]));

						if(vNewStates[gp] == DmElementState.GenL) return false;
						vNewStates[gp] = DmElementState.GenL;
					}
				}

				Debug.Assert(vNewStates.Count(st => (st == DmElementState.GenL)) ==
					ls.Blocks.Take(x + 1).Aggregate(1, (acc, blk) =>
						(acc * blk.Elements.Count)));

				bool bModified = true;
				while(bModified)
				{
					bModified = false;

					for(int gp = 0; gp < vNewStates.Length; ++gp)
					{
						if(vNewStates[gp] != DmElementState.ToGen) continue;

						bool bRightOnly = true;
						for(int gl = 0; gl < vNewStates.Length; ++gl)
						{
							if(vNewStates[gl] != DmElementState.GenL) continue;
							if(gl == gId) continue;

							int rx = dmGrp.Multiply(dmGrp.Invert(gl), gp);
							if((vNewStates[rx] != DmElementState.GenL) &&
								(vNewStates[rx] != DmElementState.Excl))
							{
								bRightOnly = false;
								break;
							}
						}

						if(bRightOnly)
						{
							vNewStates[gp] = DmElementState.GenR;
							break;
						}
					}

					for(int gl = 0; gl < vNewStates.Length; ++gl)
					{
						if(vNewStates[gl] != DmElementState.GenL) continue;

						for(int gr = 0; gr < vNewStates.Length; ++gr)
						{
							if(vNewStates[gr] != DmElementState.GenR) continue;

							int gs = dmGrp.Multiply(gl, gr);
							if(vNewStates[gs] == DmElementState.GenL) return false;
							if((vNewStates[gs] == DmElementState.GenR) && (gl != gId))
								return false;
							if(vNewStates[gs] == DmElementState.ToGen)
							{
								vNewStates[gs] = DmElementState.Excl;
								bModified = true;
							}
						}
					}
				}

				for(int gs = 0; gs < vNewStates.Length; ++gs)
				{
					if((vNewStates[gs] == DmElementState.GenL) ||
						(vNewStates[gs] == DmElementState.Excl)) continue;

					for(int gl = 0; gl < vNewStates.Length; ++gl)
					{
						if(vNewStates[gl] != DmElementState.GenL) continue;
						if(gl == gId) continue;

						int gp = dmGrp.Multiply(gl, gs);
						if(vNewStates[gp] == DmElementState.GenL)
							vNewStates[gs] = DmElementState.Excl;
					}
				}

				for(int i = 0; i < vNewStates.Length; ++i)
				{
					if(vNewStates[i] == DmElementState.GenR)
						vNewStates[i] = DmElementState.ToGen;
				}
				return RandomizeCanonicalExRec(ls, dmGrp, x + 1, 0, vNewStates);
			}
			if(y == 0)
			{
				ls.Blocks[x].Elements[0] = dmGrp.Identity;
				return RandomizeCanonicalExRec(ls, dmGrp, x, y + 1, vStates);
			}

			T tStop = dmGrp.GenerateRandomElement();
			T t = tStop;
			while(true)
			{
				t = dmGrp.NextElement(t);
				if(t.CompareTo(tStop) == 0) break;

				int it = dmGrp.ElementToInt(t);
				if((vStates[it] != DmElementState.GenL) && (vStates[it] != DmElementState.Excl))
				{
					ls.Blocks[x].Elements[y] = t;

					if(RandomizeCanonicalExRec(ls, dmGrp, x, y + 1, vStates))
						return true;
				}
			}

			return false;
		} */

		public static BigInteger FactorizationToBigInt<T>(DmBlockSeq<T> ls,
			int[] vFactorization)
			where T : IDmGroupElement<T>
		{
			BigInteger bi = BigInteger.Zero;

			for(int i = 0; i < ls.Blocks.Count; ++i)
			{
				bi *= ls.Blocks[i].Elements.Count;
				bi += vFactorization[i];
			}

			return bi;
		}

		public static T BigIntToElement<T>(DmBlockSeq<T> ls, IDmGroup<T> grp,
			BigInteger biFactorization)
			where T : IDmGroupElement<T>
		{
			T tProd = grp.Identity;

			BigInteger m = BigInteger.One;
			for(int i = 0; i < (ls.Blocks.Count - 1); ++i)
				m *= ls.Blocks[i].Elements.Count;

			BigInteger bi = biFactorization;
			for(int i = (ls.Blocks.Count - 1); i >= 0; --i)
			{
				int d = ls.Blocks[i].Elements.Count;
				int r = (int)(bi % d);

				tProd = grp.Multiply(ls.Blocks[i].Elements[r], tProd);

				bi /= d;
			}
			Debug.Assert(bi.IsZero);

			return tProd;
		}

		public static bool EnumCanBlockSeq<T>(DmBlockSeq<T> ls, IDmGroup<T> grp,
			DmLogSigDelegate<T> f, bool bReqSorted)
			where T : IDmGroupElement<T>
		{
			if(f == null) { Debug.Assert(false); return false; }
			if(ls == null) { Debug.Assert(false); return false; }

			int cx = ls.Blocks.Count;
			if(cx == 0) { Debug.Assert(false); return false; }
			for(int x = 0; x < cx; ++x)
			{
				int cy = ls.Blocks[x].Elements.Count;
				if(cy <= 1) { Debug.Assert(false); return false; }
				for(int y = 0; y < cy; ++y)
					ls.Blocks[x].Elements[y] = grp.Identity;
			}

			HashSet<T> hsId = new HashSet<T>(), hsEmpty = new HashSet<T>();
			hsId.Add(grp.Identity);

			return EnumCanBlockSeqRec(ls, grp, 0, 1, hsId, hsEmpty, f, bReqSorted);
		}

		private static bool EnumCanBlockSeqRec<T>(DmBlockSeq<T> ls, IDmGroup<T> grp,
			int x, int y, HashSet<T> hsLeft, HashSet<T> hsTop, DmLogSigDelegate<T> f,
			bool bReqSorted)
			where T : IDmGroupElement<T>
		{
			List<T> l = ls.Blocks[x].Elements;
			bool bLast = ((x == (ls.Blocks.Count - 1)) && (y == (l.Count - 1)));

			T t = (bReqSorted ? l[y - 1] : grp.Identity);
			while(true)
			{
				t = grp.NextElement(t);
				if(t.Equals(grp.Identity)) break;

				if(hsLeft.Contains(t) || hsTop.Contains(t)) continue;

				HashSet<T> hsNew = GeneratedElements(hsLeft, t, grp, hsTop);
				if(hsNew == null) continue;

				l[y] = t;

				if(bLast)
				{
					Debug.Assert(GeneratesWholeGroup(ls, grp));
					if(!f(ls)) return false;
				}
				else
				{
					bool bSub;
					if(y == (l.Count - 1))
					{
						HashSet<T> hsNewLeft = new HashSet<T>(hsLeft);
						foreach(T tTop in hsTop) hsNewLeft.Add(tTop);
						foreach(T tNew in hsNew) hsNewLeft.Add(tNew);
						Debug.Assert(hsNewLeft.Count == (hsLeft.Count +
							hsTop.Count + hsNew.Count));

						bSub = EnumCanBlockSeqRec(ls, grp, x + 1, 1, hsNewLeft,
							new HashSet<T>(), f, bReqSorted);
					}
					else
					{
						HashSet<T> hsNewTop = new HashSet<T>(hsTop);
						foreach(T tNew in hsNew) hsNewTop.Add(tNew);
						Debug.Assert(hsNewTop.Count == (hsTop.Count + hsNew.Count));

						bSub = EnumCanBlockSeqRec(ls, grp, x, y + 1, hsLeft,
							hsNewTop, f, bReqSorted);
					}

					if(!bSub) return false;
				}
			}

			return true;
		}

		private static HashSet<T> GeneratedElements<T>(HashSet<T> hsLeft, T t,
			IDmGroup<T> grp, HashSet<T> hsTop)
			where T : IDmGroupElement<T>
		{
			HashSet<T> hs = new HashSet<T>();

			foreach(T tLeft in hsLeft)
			{
				T tProd = grp.Multiply(tLeft, t);

				if(hsLeft.Contains(tProd) || hsTop.Contains(tProd))
					return null;

				hs.Add(tProd);
			}

			return hs;
		}

		public static List<T> GetZ<T>(List<T> l, IDmGroup<T> grp)
			where T : IDmGroupElement<T>
		{
			List<T> lZ = new List<T>();

			foreach(T z in l)
			{
				if(l.TrueForAll(r => (l.IndexOf(grp.Multiply(z,
					grp.Invert(r))) >= 0)))
					lZ.Add(z);
			}

			return lZ;
		}

		public static List<T> GetZHat<T>(List<T> l, IDmGroup<T> grp)
			where T : IDmGroupElement<T>
		{
			List<T> lZ = GetZ<T>(l, grp);
			List<T> lZInv = lZ.Select(z => grp.Invert(z)).ToList();

			return DmBlockSeq<T>.Multiply(grp, lZ, lZInv).Distinct().ToList();
		}
	}
}
