﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM.LsGen.Transformations;

namespace LogSig.DM.LsGen
{
	public static class LsGen<T>
		where T : IDmGroupElement<T>
	{
		private static List<LsGenTransformation<T>> GetTransformationsForGroup(
			IDmGroup<T> grp)
		{
			var l = new List<LsGenTransformation<T>>();

			AddTrfIfApplicable(l, grp, new LsgtBlockSubst<T>());
			AddTrfIfApplicable(l, grp, new LsgtTsn<T>());
			AddTrfIfApplicable(l, grp, new LsgtBlockShuffle<T>());
			AddTrfIfApplicable(l, grp, new LsgtNormalSelShifts<T>());

			return l;
		}

		private static void AddTrfIfApplicable(List<LsGenTransformation<T>> l,
			IDmGroup<T> grp, LsGenTransformation<T> t)
		{
			if(!t.SupportedForGroup(grp)) return;
			l.Add(t);
		}

		public static DmBlockSeq<T> Generate(IDmGroup<T> grp, ulong uRounds,
			CryptoRandomGen rand, T tOptToFactor, ref string strRels,
			out List<int> vFactorization)
		{
			if(grp == null) throw new ArgumentNullException("grp");
			if(rand == null) throw new ArgumentNullException("rand");

			DmBlockSeq<T> lsInit = grp.CreateDefaultSig();
			vFactorization = null;
			if(tOptToFactor != null)
			{
				T tTF = tOptToFactor;
				if(!string.IsNullOrEmpty(strRels))
				{
					var lRels = TransformationIORel<T>.ParseList(strRels, grp);
					tTF = TransformationIORel<T>.UndoList(tTF, grp, lRels);
				}
				vFactorization = new List<int>(grp.FactorInDefaultSig(tTF));
			}

			List<LsGenTransformation<T>> lTrf = GetTransformationsForGroup(grp);
			TransformationParams<T> p = new TransformationParams<T>(grp,
				rand, lsInit, vFactorization);

			for(ulong r = 0; r < uRounds; ++r)
			{
#if DEBUG
				DmBlockSeq<T> lsInput = p.LogSig.Clone();
#endif

				int t = rand.GenerateInt32(lTrf.Count);
				lTrf[t].Transform(p);

#if DEBUG
				if(grp.Order <= 256)
				{
					Debug.Assert(DmLogSigUtil.GeneratesWholeGroup(p.LogSig,
						p.Group), "Invalid transformation from " +
						lsInput.ToString() + " to " + p.LogSig.ToString());
				}
#endif
			}

			TransformationIORel<T>.FuseList(p.TransformationIORels, grp);
			strRels = TransformationIORel<T>.ListToString(p.TransformationIORels);

			if(tOptToFactor != null) vFactorization = p.Factorization;
			return p.LogSig;
		}
	}
}
