﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using LogSig.DM.LsGen.GroupImplCaps;

namespace LogSig.DM.LsGen.Transformations
{
	public sealed class LsgtNormalSelShifts<T> : LsGenTransformation<T>
		where T : IDmGroupElement<T>
	{
		public override bool SupportedForGroup(IDmGroup<T> grp)
		{
			return ((grp is INormalChainTest<T>) && (grp is INormalChainFac<T>));
		}

		public override void Transform(TransformationParams<T> p)
		{
			IDmGroup<T> grp = p.Group;
			INormalChainTest<T> t = (grp as INormalChainTest<T>);
			if(t == null) { Debug.Assert(false); return; }
			INormalChainFac<T> f = (grp as INormalChainFac<T>);
			if(f == null) { Debug.Assert(false); return; }

			var lBlocks = p.LogSig.Blocks;

			List<int> lPerm = new List<int>();
			for(int i = 0; i < lBlocks.Count; ++i) lPerm.Add(i);
			DmLogSigUtil.ShuffleList(lPerm, p.Random);

			var lChain = new List<DmLogSigBlock<T>>();
			var lChainIndices = new List<int>();

			while(lChain.Count < lBlocks.Count)
			{
				bool bBlockAdded = false;
				foreach(int i in lPerm)
				{
					if(lChainIndices.IndexOf(i) >= 0) continue;

					lChain.Add(lBlocks[i]);
					if(t.IsNormalChain(lChain))
					{
						lChainIndices.Add(i);
						bBlockAdded = true;
						break;
					}
					else lChain.RemoveAt(lChain.Count - 1);
				}

				if(!bBlockAdded) break;
			}

			if(lChain.Count == 0) return; // No normal subgroup block found

			int d = p.Random.GenerateInt32(lChain.Count);
			for(int di = 0; di < d; ++di)
			{
				lChain.RemoveAt(lChain.Count - 1);
				lChainIndices.RemoveAt(lChainIndices.Count - 1);
			}

			for(int x = 0; x < lBlocks.Count; ++x)
			{
				// Input logarithmic signature is canonical
				Debug.Assert(lBlocks[x].Elements[0].Equals(grp.Identity));

				if(lChainIndices.IndexOf(x) >= 0) continue;

				var blk = lBlocks[x].Elements;
				for(int y = 1; y < blk.Count; ++y)
				{
					T tAdd = grp.Identity, tChainProd = grp.Identity;
					for(int n = 0; n < lChain.Count; ++n)
					{
						int r = p.Random.GenerateInt32(lChain[n].Elements.Count);
						tAdd = grp.Multiply(tAdd, lChain[n].Elements[r]);

						if(p.Factorization != null)
							tChainProd = grp.Multiply(tChainProd, lChain[n].Elements[
								p.Factorization[lChainIndices[n]]]);
					}

					lBlocks[x].Elements[y] = grp.Multiply(lBlocks[x].Elements[y], tAdd);

					if((p.Factorization != null) && (p.Factorization[x] == y))
					{
						var lFac = f.FactorInNormalChain(lChain, grp.Multiply(
							tChainProd, grp.Invert(tAdd)));

						for(int n = 0; n < lChain.Count; ++n)
							p.Factorization[lChainIndices[n]] = lFac[n];
					}
				}
			}
		}
	}
}
