﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using LogSig.DM.LsGen.TransformationIORels;

namespace LogSig.DM.LsGen.Transformations
{
	/// <summary>
	/// TSN transformation (translation, element shuffle and
	/// normalization).
	/// </summary>
	public sealed class LsgtTsn<T> : LsGenTransformation<T>
		where T : IDmGroupElement<T>
	{
		public override void Transform(TransformationParams<T> p)
		{
			int iStep = (p.Random.GenerateBool() ? 1 : -1);

			Translate(p, iStep);
			ShuffleElements(p);
			Normalize(p, iStep);
		}

		private static void Translate(TransformationParams<T> p, int iStep)
		{
			T elTrans = p.Group.GenerateRandomElement(p.Random);
			int n = p.LogSig.Blocks.Count;
			List<T> v = p.LogSig.Blocks[(iStep > 0) ? 0 : (n - 1)].Elements;

			for(int y = 0; y < v.Count; ++y)
			{
				if(iStep > 0) v[y] = p.Group.Multiply(elTrans, v[y]);
				else v[y] = p.Group.Multiply(v[y], elTrans);
			}

			p.TransformationIORels.Add(new LeftRightTranslationRel<T>(
				((iStep > 0) ? elTrans : p.Group.Identity),
				((iStep > 0) ? p.Group.Identity : elTrans)));
		}

		private static void ShuffleElements(TransformationParams<T> p)
		{
			List<DmLogSigBlock<T>> lBlocks = p.LogSig.Blocks;
			List<int> lFac = p.Factorization;

			for(int x = 0; x < lBlocks.Count; ++x)
			{
				List<T> l = lBlocks[x].Elements;
				if(l.Count == 0) { Debug.Assert(false); continue; }

				T tSelected = ((lFac != null) ? l[lFac[x]] : l[0]);

				DmLogSigUtil.ShuffleList(l, p.Random);

				if(lFac != null) lFac[x] = l.IndexOf(tSelected);
			}
		}

		private static void Normalize(TransformationParams<T> p, int iStep)
		{
			int n = p.LogSig.Blocks.Count;
			var ls = ((iStep > 0) ? DmLogSigUtil.GetLStarCanonical(p.LogSig,
				p.Group) : DmLogSigUtil.GetRStarCanonical(p.LogSig, p.Group));
			p.LogSig = ls;

			int xNew = ((iStep > 0) ? (ls.Blocks.Count - 1) : 0);

			Debug.Assert(ls.Blocks[xNew].Elements.Count == 1);
			T elTrans = ls.Blocks[xNew].Elements[0];

			T tInv = p.Group.Invert(elTrans);
			p.TransformationIORels.Add(new LeftRightTranslationRel<T>(
				((iStep > 0) ? p.Group.Identity : tInv),
				((iStep > 0) ? tInv : p.Group.Identity)));

			if(ls.Blocks.Count != (n + 1)) { Debug.Assert(false); }
			else ls.Blocks.RemoveAt(xNew);
		}
	}
}
