// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.Util;

namespace LogSig.DM.MultiZnGroup
{
	public sealed partial class DmMultiZnGroup : IDmGroup<DmMultiZnGroupElement>
	{
		public static bool Factorize(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el, int[] vFactorization)
		{
			return Factorize(ls, grp, el, new DmBlockSeq<DmMultiZnGroupElement>(),
				vFactorization, 0);
		}

		private static bool Factorize(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el,
			DmBlockSeq<DmMultiZnGroupElement> lsDivChain, int[] vFactorization,
			uint uLevel)
		{
			NormalChainEq eq = new NormalChainEq(grp, lsDivChain);

			if(Program.Verbose) TraceParams(ls, lsDivChain, eq.NormalChain, uLevel);

			if(ls.Blocks.Count == 1)
			{
				for(int y = 0; y < ls.Blocks[0].Elements.Count; ++y)
				{
					if(eq.Equals(el, ls.Blocks[0].Elements[y]))
					{
						vFactorization[0] = y;

						if(Program.Verbose) TraceFac(ls, vFactorization, uLevel);
						return true;
					}
				}

				return false; // Element not expressible
			}

			int iSubBlock = FindSubGroupBlock(ls, grp, eq);
			if(iSubBlock < 0)
				return SplitPeriodicRec(ls, grp, el, lsDivChain,
					vFactorization, uLevel);

			if(Program.Verbose) TraceSubGroup(iSubBlock, uLevel);

			DmBlockSeq<DmMultiZnGroupElement> lsNewDivChain = lsDivChain.Clone();
			lsNewDivChain.AddBlock(ls.Blocks[iSubBlock].Elements.ToArray());

			DmBlockSeq<DmMultiZnGroupElement> lsSub = ls.Clone();
			lsSub.Blocks.RemoveAt(iSubBlock);

			int[] vSubFac = new int[lsSub.Blocks.Count];
			if(!Factorize(lsSub, grp, el, lsNewDivChain, vSubFac, uLevel + 1))
				return false;

			Array.Copy(vSubFac, 0, vFactorization, 0, iSubBlock);
			Array.Copy(vSubFac, iSubBlock, vFactorization, iSubBlock + 1,
				vSubFac.Length - iSubBlock);

			return CompleteFac(ls, grp, el, eq, iSubBlock, vFactorization, uLevel);
		}

		private static bool CompleteFac(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el,
			NormalChainEq eq, int iOpenBlock, int[] vFactorization,
			uint uLevel)
		{
			for(int i = 0; i < ls.Blocks[iOpenBlock].Elements.Count; ++i)
			{
				vFactorization[iOpenBlock] = i;
				DmMultiZnGroupElement d = ls.MultiplyElements(new List<int>(
					vFactorization), grp);
				if(eq.Equals(d, el))
				{
					if(Program.Verbose) TraceFac(ls, vFactorization, uLevel);
					return true;
				}
			}

			return false;
		}

		private static int FindSubGroupBlock(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, NormalChainEq eq)
		{
			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				if(IsSubGroupBlock(ls.Blocks[x].Elements, grp, eq))
					return x;
			}

			return -1; // No subgroup block found
		}

		// A set M is a subgroup, if a*b^-1 is in M for all a, b in M
		private static bool IsSubGroupBlock(List<DmMultiZnGroupElement> blk,
			DmMultiZnGroup grp, NormalChainEq eq)
		{
			if(blk.Count == 0) { Debug.Assert(false); return false; }

			for(int i = 0; i < blk.Count; ++i)
			{
				for(int j = 0; j < blk.Count; ++j)
				{
					DmMultiZnGroupElement el = grp.Multiply(blk[i],
						grp.Invert(blk[j]));

					bool bFound = false;
					for(int k = 0; k < blk.Count; ++k)
					{
						if(eq.Equals(el, blk[k])) { bFound = true; break; }
					}

					if(!bFound) return false;
					Debug.Assert(blk.Count(t => eq.Equals(t, el)) == 1);
				}
			}

			return true;
		}

		private static List<DmMultiZnGroupElement> GetBlockPeriods(
			List<DmMultiZnGroupElement> blk, DmMultiZnGroup grp,
			NormalChainEq eq)
		{
			List<DmMultiZnGroupElement> l = new List<DmMultiZnGroupElement>();

			// As blk is canonical, the only candidates for periods
			// are the elements in blk
			foreach(DmMultiZnGroupElement g in blk)
			{
				if(eq.Equals(g, grp.Identity)) continue;

				if(blk.TrueForAll(el => blk.Contains(grp.Multiply(el, g), eq)))
					l.Add(g);
			}

			return l;
		}

		private static bool SplitIfPeriodicBlock(List<DmMultiZnGroupElement> blk,
			DmMultiZnGroup grp, NormalChainEq eq, List<DmMultiZnGroupElement> blkOutB,
			List<DmMultiZnGroupElement> blkOutS)
		{
			List<DmMultiZnGroupElement> lPeriods = GetBlockPeriods(blk, grp, eq);
			if(lPeriods.Count == 0) return false;

			blkOutS.Add(grp.Identity);
			blkOutS.AddRange(lPeriods);

			List<DmMultiZnGroupElement> lRem = new List<DmMultiZnGroupElement>(blk);
			while(lRem.Count > 0)
			{
				DmMultiZnGroupElement r = lRem[0];
				blkOutB.Add(r);

				List<DmMultiZnGroupElement> lCoset = new List<DmMultiZnGroupElement>(
					blkOutS.Select(el => grp.Multiply(r, el)));

				lRem.RemoveAll(el => lCoset.Contains(el, eq));
			}

			return true;
		}

		private static bool SplitPeriodicRec(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el,
			DmBlockSeq<DmMultiZnGroupElement> lsDivChain, int[] vFactorization,
			uint uLevel)
		{
			NormalChainEq eq = new NormalChainEq(grp, lsDivChain);
			List<DmMultiZnGroupElement> lS = new List<DmMultiZnGroupElement>(),
				lB = new List<DmMultiZnGroupElement>();

			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				if(SplitIfPeriodicBlock(ls.Blocks[x].Elements, grp, eq, lB, lS))
				{
					if(Program.Verbose) TraceSplitPeriodic(x, uLevel);

					var blkB = new DmLogSigBlock<DmMultiZnGroupElement>();
					blkB.Elements.AddRange(lB);

					var blkS = new DmLogSigBlock<DmMultiZnGroupElement>();
					blkS.Elements.AddRange(lS);

					DmBlockSeq<DmMultiZnGroupElement> lsNew = ls.Clone();
					lsNew.Blocks.RemoveAt(x);
					lsNew.Blocks.Insert(x, blkS);
					lsNew.Blocks.Insert(x, blkB);

					int[] vSubFac = new int[lsNew.Blocks.Count];
					if(!Factorize(lsNew, grp, el, lsDivChain, vSubFac, uLevel + 1))
						return false;

					Array.Copy(vSubFac, 0, vFactorization, 0, x);
					Array.Copy(vSubFac, x + 2, vFactorization, x + 1,
						vSubFac.Length - (x + 2));

					return CompleteFac(ls, grp, el, eq, x, vFactorization, uLevel);
				}
			}

			return false;
		}
	}
}
