// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

namespace LogSig.DM.MultiZnGroup
{
	public sealed partial class DmMultiZnGroup : IDmGroup<DmMultiZnGroupElement>
	{
		/// <summary>
		/// Compute all factorizations of <paramref name="el" /> with
		/// respect to <paramref name="bs" />.
		/// </summary>
		/// <param name="bs">An arbitrary block sequence of elements
		/// in <paramref name="grp" /> (not necessarily a logarithmic signature
		/// or multiple factorization of <paramref name="grp" />).</param>
		/// <param name="grp">Group.</param>
		/// <param name="el">Element to factor.</param>
		/// <returns>All factorizations of <paramref name="el" />.</returns>
		public static List<int[]> FactorizeAbEx(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			DmMultiZnGroupElement el, DmTraceInfo ti)
		{
			return FactorizeAbEx(bs, grp, el,
				new DmBlockSeq<DmMultiZnGroupElement>(), ti);
		}

		private static List<int[]> FactorizeAbEx(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			DmMultiZnGroupElement el, DmBlockSeq<DmMultiZnGroupElement> lsDivChain,
			DmTraceInfo ti)
		{
			List<int[]> lRes = new List<int[]>();
			NormalChainEq eq = new NormalChainEq(grp, lsDivChain);

			if(bs.Blocks.Count == 1)
			{
				for(int i = 0; i < bs.Blocks[0].Elements.Count; ++i)
				{
					if(eq.Equals(bs.Blocks[0].Elements[i], el))
						lRes.Add(new int[] { i });
				}
				return UpdTrace(lRes, ti);
			}

			for(int x = 0; x < bs.Blocks.Count; ++x)
			{
				DmLogSigBlock<DmMultiZnGroupElement> blk = bs.Blocks[x];
				List<DmMultiZnGroupElement> lUni = blk.Elements.Distinct(
					eq).ToList();
				if(blk.Elements.Count > lUni.Count)
				{
					DmBlockSeq<DmMultiZnGroupElement> bsNew = bs.Clone();
					bsNew.Blocks[x].Elements.Clear();
					bsNew.Blocks[x].Elements.AddRange(lUni);

					List<int[]> lSubFac = FactorizeAbEx(bsNew, grp, el,
						lsDivChain, ti);
					return UpdTrace(EnumAndMatch(bs, grp, el, eq, lSubFac,
						x, false), ti);
				}
			}

			for(int x = 0; x < bs.Blocks.Count; ++x)
			{
				DmLogSigBlock<DmMultiZnGroupElement> blk = bs.Blocks[x];
				if(blk.Elements.Count == 1)
				{
					DmBlockSeq<DmMultiZnGroupElement> bsNew = bs.Clone();
					bsNew.Blocks.RemoveAt(x);

					List<int[]> lSubFac = FactorizeAbEx(bsNew, grp, el,
						lsDivChain, ti);
					return UpdTrace(EnumAndMatch(bs, grp, el, eq, lSubFac,
						x, true), ti);
				}
			}

			List<DmMultiZnGroupElement> lN = ChooseNormalSubgroup(bs, grp, eq);

			DmBlockSeq<DmMultiZnGroupElement> lsDivChainN = lsDivChain.Clone();
			lsDivChainN.AddBlock(lN.ToArray());

			List<int[]> lSubFacN = FactorizeAbEx(bs, grp, el, lsDivChainN, ti);
			foreach(int[] vCand in lSubFacN)
			{
				if(eq.Equals(bs.MultiplyElements(new List<int>(vCand), grp), el))
					lRes.Add(vCand);
			}

			return UpdTrace(lRes, ti);
		}

		private static List<int[]> UniqueFacs(List<int[]> lFacs)
		{
			return lFacs.Distinct(new ArraysEqualComparer<int>()).ToList();
		}

		private static List<int[]> UpdTrace(List<int[]> l, DmTraceInfo ti)
		{
			if(l.Count > ti.MaxWidth) ti.MaxWidth = l.Count;
			return l;
		}

		private static List<int[]> EnumAndMatch(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			DmMultiZnGroupElement el, NormalChainEq eq, List<int[]> lFacs,
			int x, bool bInsert)
		{
			List<int[]> lRes = new List<int[]>();

			foreach(int[] vCand in lFacs)
			{
				for(int j = 0; j < bs.Blocks[x].Elements.Count; ++j)
				{
					List<int> lFac = new List<int>(vCand);

					if(bInsert) lFac.Insert(x, j);
					else lFac[x] = j;

					if(eq.Equals(bs.MultiplyElements(lFac, grp), el))
						lRes.Add(lFac.ToArray());
				}
			}

			return UniqueFacs(lRes);
		}

		private sealed class ArraysEqualComparer<T> : IEqualityComparer<T[]>
			where T : IEquatable<T>
		{
			public bool Equals(T[] x, T[] y)
			{
				if((x == null) || (y == null))
				{
					Debug.Assert(false);
					return false;
				}

				if(x.Length != y.Length) return false;

				for(int i = 0; i < x.Length; ++i)
				{
					if(!x[i].Equals(y[i])) return false;
				}

				return true;
			}

			public int GetHashCode(T[] obj)
			{
				if(obj == null) { Debug.Assert(false); return 0; }

				return obj.Sum(o => o.GetHashCode());
			}
		}

		private sealed class ElemWithRatings
		{
			public DmMultiZnGroupElement Element;
			public List<BigInteger> Ratings;

			public ElemWithRatings(DmMultiZnGroupElement e,
				List<BigInteger> lRatings)
			{
				this.Element = e;
				this.Ratings = lRatings;
			}
		}

		private static List<DmMultiZnGroupElement> ChooseNormalSubgroup(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			NormalChainEq eq)
		{
			IEnumerable<DmMultiZnGroupElement> eC = GetLowOrderElements(bs, grp,
				eq);

			List<ElemWithRatings> lR = new List<ElemWithRatings>();
			foreach(DmMultiZnGroupElement e in eC)
			{
				NormalChainEq eqSub = AppendSubgroup(grp, eq, e);
				List<BigInteger> lRating = EstSeqComplexity(bs, grp, eqSub);
				if(lRating.TrueForAll(bi => bi.IsOne))
					return GetCyclicSpan(grp, eq, e).Elements;
				lRating[0] *= ElementOrder(e, grp, eq);
				lR.Add(new ElemWithRatings(e, lRating));
			}

			BigInteger biEta = lR.Min(r => r.Ratings[0]);
			lR = lR.Where(r => (r.Ratings[0] == biEta)).ToList();

			BigInteger biTheta = lR.Min(r => r.Ratings[1]);
			lR = lR.Where(r => (r.Ratings[1] == biTheta)).ToList();

			if(lR.Count == 0) throw new Exception("lR.Count == 0");
			int ri = Program.CryptoRandom.GenerateInt32(lR.Count);
			return GetCyclicSpan(grp, eq, lR[ri].Element).Elements;
		}

		private static IEnumerable<DmMultiZnGroupElement> GetLowOrderElements(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			NormalChainEq eq)
		{
			List<DmMultiZnGroupElement> lM = new List<DmMultiZnGroupElement>();
			for(int x = 0; x < bs.Blocks.Count; ++x)
			{
				var blk = bs.Blocks[x].Elements;
				for(int j = 0; j < (blk.Count - 1); ++j)
				{
					for(int k = j + 1; k < blk.Count; ++k)
					{
						var m = grp.Multiply(grp.Invert(blk[j]), blk[k]);
						if(lM.TrueForAll(el => !eq.Equals(el, m)))
							lM.Add(m);
					}
				}
			}

			BigInteger q = new BigInteger(bs.Blocks.Max(
				blk => blk.Elements.Count));

			int d = 1;
			List<DmMultiZnGroupElement> lC = new List<DmMultiZnGroupElement>();
			while(lC.Count == 0)
			{
				lC = lM.Where(e => ((BigInteger.Pow(q, d - 1) < ElementOrder(e,
					grp, eq)) && (ElementOrder(e, grp, eq) <=
					BigInteger.Pow(q, d)))).ToList();
				++d;
			}

			for(int x = lC.Count - 1; x > 0; --x)
			{
				BigInteger ordX = ElementOrder(lC[x], grp, eq);
				var blkSpanX = GetCyclicSpan(grp, eq, lC[x]);

				bool bValid = true;
				for(int y = 0; y < x; ++y)
				{
					if((ElementOrder(lC[y], grp, eq) == ordX) &&
						(blkSpanX.Elements.FindIndex(el => eq.Equals(el,
						lC[y])) >= 0))
					{
						bValid = false;
						break;
					}
				}

				if(bValid) yield return lC[x];
			}

			yield return lC[0];
		}

		private static BigInteger ElementOrder(DmMultiZnGroupElement el,
			DmMultiZnGroup grp, NormalChainEq eq)
		{
			if(eq.Equals(el, grp.Identity)) return BigInteger.One;

			List<BigInteger> lFactors = new List<BigInteger>(grp.OrderFactors);
			int p = lFactors.Count - 1;

			while(p >= 0)
			{
				BigInteger biPow = lFactors.Aggregate(BigInteger.One,
					(prod, f) => prod * f);
				biPow /= lFactors[p];
				if(eq.Equals(grp.Pow(el, biPow), grp.Identity))
					lFactors.RemoveAt(p);

				--p;
			}

			return lFactors.Aggregate(BigInteger.One, (prod, f) => prod * f);
		}

		private static DmLogSigBlock<DmMultiZnGroupElement> GetCyclicSpan(
			DmMultiZnGroup grp, NormalChainEq eq, DmMultiZnGroupElement g)
		{
			var blkSubgroup = new DmLogSigBlock<DmMultiZnGroupElement>();
			DmMultiZnGroupElement s = grp.Identity;
			do
			{
				blkSubgroup.Elements.Add(s);
				s = grp.Multiply(s, g);
			}
			while(!eq.Equals(s, grp.Identity));
			return blkSubgroup;
		}

		private static NormalChainEq AppendSubgroup(DmMultiZnGroup grp,
			NormalChainEq eq, DmMultiZnGroupElement g)
		{
			DmBlockSeq<DmMultiZnGroupElement> lsNew = eq.NormalChain.Clone();
			lsNew.Blocks.Add(GetCyclicSpan(grp, eq, g));
			return new NormalChainEq(grp, lsNew);
		}

		private static List<BigInteger> EstSeqComplexity(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			NormalChainEq eq)
		{
			var bsSmp = SimplifySeq(bs, grp, eq);
			if(bsSmp.Blocks.Count <= 1)
				return new List<BigInteger>(new BigInteger[] {
					BigInteger.One, BigInteger.One });

			BigInteger u = bsSmp.Blocks.Aggregate(BigInteger.One, (prod, blk) =>
				(prod * blk.Elements.Count));
			BigInteger s = bsSmp.Blocks.Aggregate(BigInteger.Zero, (so, blk) =>
				(so + blk.Elements.Aggregate(BigInteger.Zero, (si, e) =>
				(si + ElementOrder(e, grp, eq)))));

			var eC = GetLowOrderElements(bsSmp, grp, eq);
			foreach(DmMultiZnGroupElement g in eC)
			{
				NormalChainEq eqSub = AppendSubgroup(grp, eq, g);
				var bsSub = SimplifySeq(bsSmp, grp, eqSub);
				if((ElementOrder(g, grp, eq) * bsSub.Blocks.Aggregate(
					BigInteger.One, (prod, blk) => (prod *
					blk.Elements.Count))) <= u)
					return EstSeqComplexity(bsSmp, grp, eqSub);
			}

			List<BigInteger> lRes = new List<BigInteger>(2);
			lRes.Add(u);
			lRes.Add(s);
			return lRes;
		}

		private static DmBlockSeq<DmMultiZnGroupElement> SimplifySeq(
			DmBlockSeq<DmMultiZnGroupElement> bs, DmMultiZnGroup grp,
			NormalChainEq eq)
		{
			var bsNew = new DmBlockSeq<DmMultiZnGroupElement>();
			for(int x = 0; x < bs.Blocks.Count; ++x)
				bsNew.AddBlock(bs.Blocks[x].Elements.Distinct(eq).ToArray());
			for(int x = bs.Blocks.Count - 1; x >= 0; --x)
			{
				if(bsNew.Blocks[x].Elements.Count == 1)
					bsNew.Blocks.RemoveAt(x);
			}
			return bsNew;
		}

		private sealed class NormalChainEq :
			IEqualityComparer<DmMultiZnGroupElement>
		{
			private readonly DmMultiZnGroup m_grp;
			private readonly DmMultiZnGroupElement[] m_v;

			private readonly DmBlockSeq<DmMultiZnGroupElement> m_ls;
			public DmBlockSeq<DmMultiZnGroupElement> NormalChain
			{
				get { return m_ls; }
			}

			public NormalChainEq(DmMultiZnGroup grp,
				DmBlockSeq<DmMultiZnGroupElement> lsNormalChain)
			{
				if(grp == null) throw new ArgumentNullException("grp");
				m_grp = grp;

				m_v = new DmMultiZnGroupElement[grp.Mods.Length];
				m_ls = Simplify(grp, lsNormalChain);
			}

			private DmBlockSeq<DmMultiZnGroupElement> Simplify(DmMultiZnGroup grp,
				DmBlockSeq<DmMultiZnGroupElement> lsChain)
			{
				if(lsChain == null) throw new ArgumentNullException("lsChain");

				DmBlockSeq<DmMultiZnGroupElement> ls = lsChain.Clone();

				for(int c = 0; c < grp.Mods.Length; ++c)
				{
					for(int x = 0; x < ls.Blocks.Count; ++x)
					{
						DmMultiZnGroupElement sMin = null;
						foreach(DmMultiZnGroupElement el in ls.Blocks[x].Elements)
						{
							if(el.Values[c].IsZero) continue;
							if(!el.Values.Take(c).All(bi => bi.IsZero)) continue;

							if((sMin == null) || (el.Values[c] < sMin.Values[c]))
								sMin = el;
						}
						if(sMin == null) continue;

						m_v[c] = sMin;

						BigInteger sc = sMin.Values[c];
						Debug.Assert((sc >= 0) && (sc < grp.Mods[c]));
						for(int xr = x + 1; xr < ls.Blocks.Count; ++xr)
						{
							for(int yr = 0; yr < ls.Blocks[xr].Elements.Count;
								++yr)
							{
								var h = ls.Blocks[xr].Elements[yr];
								BigInteger hc = h.Values[c];
								Debug.Assert((hc >= 0) && (hc < grp.Mods[c]));

								if(sc <= hc)
								{
									var sMulti = grp.Pow(sMin, hc / sc);
									Debug.Assert(!sMulti.Equals(grp.Identity));

									var gRed = grp.Multiply(h, grp.Invert(sMulti));
									ls.Blocks[xr].Elements[yr] = gRed;
									Debug.Assert(gRed.Values[c] < sc);
								}
							}
						}
					}
				}

				return ls;
			}

			public bool Equals(DmMultiZnGroupElement x, DmMultiZnGroupElement y)
			{
				DmMultiZnGroupElement d = m_grp.Multiply(x, m_grp.Invert(y));

				for(int c = 0; c < m_grp.Mods.Length; ++c)
				{
					if(d.Values[c] == 0) continue;

					DmMultiZnGroupElement s = m_v[c];
					if(s == null) return false;

					BigInteger q = d.Values[c] / s.Values[c];

					DmMultiZnGroupElement sMul = m_grp.Pow(s, q);
					d = m_grp.Multiply(d, m_grp.Invert(sMul));

					if(!d.Values[c].IsZero) return false;
				}

				Debug.Assert(d.Values.All(bi => bi.IsZero));
				return true;
			}

			public int GetHashCode(DmMultiZnGroupElement obj)
			{
				return 0; // Force that the Equals method is used
			}
		}
	}
}
