﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Linq;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM.ZnGroup;
using LogSig.Interfaces;
using LogSig.Libs.LpSolve;

namespace LogSig.DM.MultiZnGroup
{
	public sealed partial class DmMultiZnGroup : IDmGroup<DmMultiZnGroupElement>
	{
		private readonly string m_strName;
		public string Name { get { return m_strName; } }
		// public string ShortName { get { return "DP"; } }
		// public string ShortNameAlt { get { return "x"; } }

		private readonly BigInteger[] m_vMods;
		public BigInteger[] Mods
		{
			get { return m_vMods; }
		}

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private List<BigInteger> m_lFactors = null;
		public List<BigInteger> OrderFactors
		{
			get
			{
				if(m_lFactors == null) m_lFactors = NumUtil.Factorize(m_biOrder);
				return m_lFactors;
			}
		}

		private readonly DmMultiZnGroupElement m_id;
		public DmMultiZnGroupElement Identity { get { return m_id; } }

		public DmMultiZnGroup(BigInteger[] vMods)
		{
			if(vMods == null) throw new ArgumentNullException("vMods");
			if(vMods.Length < 1) throw new ArgumentOutOfRangeException("vMods");

			m_vMods = new BigInteger[vMods.Length];
			Array.Copy(vMods, m_vMods, m_vMods.Length);

			m_biOrder = BigInteger.One;

			StringBuilder sbName = new StringBuilder();
			for(int i = 0; i < m_vMods.Length; ++i)
			{
				if(i > 0)
					sbName.Append((Program.OutputFormat == OutputFmt.TeX) ?
						" \\timesAb " : "x");

				sbName.Append((Program.OutputFormat == OutputFmt.TeX) ?
					"\\mathbb{Z}_{" : "Z");

				string strFactors = NumUtil.FactorsToString(m_vMods[i], true);
				if(Program.OutputFormat == OutputFmt.TeX)
				{
					strFactors = strFactors.Replace("*", " \\cdot ");
					strFactors += "}";
				}
				sbName.Append(strFactors);

				m_biOrder *= m_vMods[i];
			}
			m_strName = sbName.ToString();

			BigInteger[] vId = new BigInteger[m_vMods.Length];
			m_id = new DmMultiZnGroupElement(vId);
		}

		public override string ToString()
		{
			return this.Name;
		}

		public DmMultiZnGroupElement ParseElement(string str)
		{
			return DmMultiZnGroupElement.Parse(str, this);
		}

		public DmMultiZnGroupElement Multiply(DmMultiZnGroupElement a,
			DmMultiZnGroupElement b)
		{
			BigInteger[] vProd = new BigInteger[m_vMods.Length];

			for(int i = 0; i < m_vMods.Length; ++i)
				vProd[i] = ((a.Values[i] + b.Values[i]) % m_vMods[i]);

			return new DmMultiZnGroupElement(vProd);
		}

		public DmMultiZnGroupElement Pow(DmMultiZnGroupElement el,
			BigInteger e)
		{
			BigInteger[] v = new BigInteger[el.Values.Length];
			for(int i = 0; i < el.Values.Length; ++i)
				v[i] = ((el.Values[i] * e) % m_vMods[i]);
			return new DmMultiZnGroupElement(v);
		}

		public DmMultiZnGroupElement Invert(DmMultiZnGroupElement t)
		{
			BigInteger[] vInv = new BigInteger[m_vMods.Length];

			for(int i = 0; i < m_vMods.Length; ++i)
			{
				if(t.Values[i].IsZero) { Debug.Assert(vInv[i].IsZero); }
				else vInv[i] = m_vMods[i] - t.Values[i];
			}

			return new DmMultiZnGroupElement(vInv);
		}

		public DmMultiZnGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			BigInteger[] vRand = new BigInteger[m_vMods.Length];

			for(int i = 0; i < m_vMods.Length; ++i)
			{
				vRand[i] = rand.GenerateBigInt(m_vMods[i]);
			}

			return new DmMultiZnGroupElement(vRand);
		}

		public DmMultiZnGroupElement NextElement(DmMultiZnGroupElement t)
		{
			BigInteger[] vNew = new BigInteger[m_vMods.Length];
			Array.Copy(t.Values, vNew, m_vMods.Length);

			for(int i = vNew.Length - 1; i >= 0; --i)
			{
				++vNew[i];
				if(vNew[i] >= m_vMods[i]) vNew[i] -= m_vMods[i];
				else break;
			}

			return new DmMultiZnGroupElement(vNew);
		}

		public BigInteger ElementToBigInt(DmMultiZnGroupElement t)
		{
			throw new NotImplementedException();
		}

		public DmMultiZnGroupElement BigIntToElement(BigInteger biElem)
		{
			throw new NotImplementedException();
		}

		public DmBlockSeq<DmMultiZnGroupElement> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmMultiZnGroupElement>();

			for(int m = 0; m < m_vMods.Length; ++m)
			{
				DmZnGroup grp = new DmZnGroup(m_vMods[m]);
				var lsSub = grp.CreateDefaultSig();

				for(int x = 0; x < lsSub.Blocks.Count; ++x)
				{
					var blk = ls.AddBlock(m_id);

					for(int y = 1; y < lsSub.Blocks[x].Elements.Count; ++y)
					{
						BigInteger[] vEl = new BigInteger[m_vMods.Length];
						vEl[m] = lsSub.Blocks[x].Elements[y].Value;

						blk.Elements.Add(new DmMultiZnGroupElement(vEl));
					}
				}
			}

			return ls;
		}

		public int[] FactorInDefaultSig(DmMultiZnGroupElement t)
		{
			List<int> vFact = new List<int>();

			for(int m = 0; m < m_vMods.Length; ++m)
			{
				DmZnGroup grp = new DmZnGroup(m_vMods[m]);
				int[] vSubFact = grp.FactorInDefaultSig(new DmZnGroupElement(
					t.Values[m]));

				vFact.AddRange(vSubFact);
			}

			return vFact.ToArray();
		}

		private static void TraceParams(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmBlockSeq<DmMultiZnGroupElement> lsDivChain,
			DmBlockSeq<DmMultiZnGroupElement> eq, uint uLevel)
		{
			Program.TraceLine("\\begin{center} \\footnotesize");
			Program.TraceLine(ls, "$\\alpha_{" + uLevel.ToString() + "} := ",
				"\\text{,}$ \\\\");

			if(lsDivChain.Blocks.Count > 0)
				Program.TraceLine(lsDivChain, "$N_{" + uLevel.ToString() + "} := ",
					"\\text{,}$ \\\\");
			else
				Program.TraceLine("$N_{" + uLevel.ToString() +
					"} := \\{ \\id \\}\\text{,}$ \\\\");

			if(eq.Blocks.Count > 0)
				Program.TraceLine(eq, "$N_{" + uLevel.ToString() + "}' := ",
					"\\text{.}$");
			else
				Program.TraceLine("$N_{" + uLevel.ToString() +
					"}' := \\{ \\id \\}\\text{.}$");

			Program.TraceLine("\\end{center}");
			Program.TraceLine("\\noindent");
		}

		private static void TraceFac(DmBlockSeq<DmMultiZnGroupElement> ls,
			int[] vFactorization, uint uLevel)
		{
			Program.TraceLine("Factorization of $g + N_{" + uLevel.ToString() +
				"}$ in $\\alpha_{" + uLevel.ToString() + "}$:");
			Program.TraceLine("\\begin{center} \\footnotesize");
			Program.TraceLine("$\\alpha_{" + uLevel.ToString() + "} =");
			Program.TraceLine(ls.ToString(vFactorization) + "\\text{.}$");
			Program.TraceLine("\\end{center}");
			Program.TraceLine("\\noindent");
		}

		private static string GetBlockNumber(int iIndex)
		{
			if(iIndex == 0) return "first";
			if(iIndex == 1) return "second";
			if(iIndex == 2) return "third";
			return (iIndex + 1).ToString() + "th";
		}

		private static void TraceSubGroup(int iSubBlock, uint uLevel)
		{
			string str = "The " + GetBlockNumber(iSubBlock);
			str += " block in $\\alpha_{" + uLevel.ToString() +
				"}$ is a subgroup (in $G/N_{" + uLevel.ToString() +
				"}$).";
			Program.TraceLine(str);
		}

		private static void TraceSplitPeriodic(int iPeriodicBlock, uint uLevel)
		{
			string str = "The " + GetBlockNumber(iPeriodicBlock) +
				" block is periodic (in $G/N_{" + uLevel.ToString() +
				"}$) and can be splitted into two blocks.";
			Program.TraceLine(str);
		}

		// See LpFileGenerator class
		/* public static string FacPrbToILP(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement elFac)
		{
			StringBuilder sb = new StringBuilder();

			int nElements = ls.Blocks.Aggregate(0, (acc, blk) =>
				acc + blk.Elements.Count);
			int nCmp = grp.Mods.Length;

			sb.Append("max: ");
			for(int i = 1; i <= nElements; ++i)
			{
				if(i > 1) sb.Append("+");
				sb.Append("x");
				sb.Append(i);
			}
			for(int c = 0; c < nCmp; ++c)
			{
				sb.Append("-x");
				sb.Append(nElements + c + 1);
			}
			sb.AppendLine(";");
			sb.AppendLine();

			int iVar = 1;
			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				for(int y = 0; y < ls.Blocks[x].Elements.Count; ++y)
				{
					if(y > 0) sb.Append("+");
					sb.Append("x"); sb.Append(iVar); ++iVar;
				}
				sb.AppendLine("=1;");
			}
			sb.AppendLine();

			for(int c = 0; c < nCmp; ++c)
			{
				iVar = 1;
				for(int x = 0; x < ls.Blocks.Count; ++x)
				{
					for(int y = 0; y < ls.Blocks[x].Elements.Count; ++y)
					{
						if(y > 0) sb.Append("+");
						sb.Append(ls.Blocks[x].Elements[y].Values[c].ToString());
						sb.Append("x"); sb.Append(iVar); ++iVar;
					}
				}
				sb.Append("-");
				sb.Append(grp.Mods[c].ToString());
				sb.Append("x");
				sb.Append(nElements + c + 1);
				sb.Append("=");
				sb.Append(elFac.Values[c].ToString());
				sb.AppendLine(";");
			}

			return sb.ToString();
		} */

		public static int[] FactorByILP(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el)
		{
			int nLsLen = ls.Blocks.Sum(blk => blk.Elements.Count);
			int nCmp = grp.Mods.Length;
			int nVars = nLsLen + nCmp;

			int[] r = null;
			using(LpSolve lp = new LpSolve(nVars))
			{
				r = FactorByILP(ls, grp, el, lp);
			}
			return r;
		}

		public static int[] FactorByILP(DmBlockSeq<DmMultiZnGroupElement> ls,
			DmMultiZnGroup grp, DmMultiZnGroupElement el, IIlpSolver lp)
		{
			int nLsLen = ls.Blocks.Sum(blk => blk.Elements.Count);
			int nCmp = grp.Mods.Length;
			int nVars = nLsLen + nCmp;

			for(int l = 0; l < nLsLen; ++l)
				lp.SetTypeBinary(l);
			for(int c = nLsLen; c < (nLsLen + nCmp); ++c)
				lp.SetTypeInt(c);

			// Select exactly one element from each block
			int i = 0;
			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				var blk = ls.Blocks[x];
				double[] vRow = new double[nVars];

				for(int y = 0; y < blk.Elements.Count; ++y)
				{
					vRow[i] = 1.0;
					++i;
				}

				lp.AddConstraint(vRow, ConstraintType.EQ, 1.0);
			}
			/* for(int x = 0; x < nLsLen; ++x)
			{
				double[] vRow = new double[nVars];
				vRow[x] = 1.0;
				lp.AddConstraint(vRow, LpSolve.ConstraintType.GE, 0.0);
			} */

			// Require solution
			for(int c = 0; c < nCmp; ++c)
			{
				double[] vRow = new double[nVars];

				i = 0;
				for(int x = 0; x < ls.Blocks.Count; ++x)
				{
					var blk = ls.Blocks[x];

					for(int y = 0; y < blk.Elements.Count; ++y)
					{
						vRow[i] = (double)blk.Elements[y].Values[c];
						++i;
					}
				}

				vRow[nLsLen + c] = -((double)grp.Mods[c]);
				lp.AddConstraint(vRow, ConstraintType.EQ,
					(double)el.Values[c]);
			}

			int[] v = lp.SolveInt();
			int[] r = new int[ls.Blocks.Count];
			i = 0;
			for(int x = 0; x < ls.Blocks.Count; ++x)
			{
				var blk = ls.Blocks[x];
				for(int y = 0; y < blk.Elements.Count; ++y)
				{
					if(v[i] != 0) r[x] = y;
					++i;
				}
			}
			return r;
		}
	}
}
