﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Numerics;
using System.Diagnostics;

using LogSig.Util;

namespace LogSig.DM.MultiZnGroup
{
	public sealed class DmMultiZnGroupElement : IDmGroupElement<DmMultiZnGroupElement>
	{
		private int? m_hash = null;

		private readonly BigInteger[] m_vValues;
		public BigInteger[] Values
		{
			get { return m_vValues; }
			// set { m_vValues = value; } // Immutable
		}

		public DmMultiZnGroupElement(BigInteger[] vValues)
		{
			if(vValues == null) throw new ArgumentNullException("vValues");
			if(vValues.Length < 1) throw new ArgumentOutOfRangeException("vValues");

			m_vValues = new BigInteger[vValues.Length];
			Array.Copy(vValues, m_vValues, m_vValues.Length);
		}

		public DmMultiZnGroupElement Clone()
		{
			return (DmMultiZnGroupElement)MemberwiseClone();
		}

		public override string ToString()
		{
			// if(m_vValues.All(bi => bi.IsZero))
			//	return ((Program.OutputFormat == OutputFmt.TeX) ? "\\id" : "id");

			string strSep = ((Program.OutputFormat == OutputFmt.TeX) ? ", " : ",");

			StringBuilder sb = new StringBuilder();
			sb.Append('(');
			for(int i = 0; i < m_vValues.Length; ++i)
			{
				if(i > 0) sb.Append(strSep);
				sb.Append(NumUtil.FactorsToString(m_vValues[i], true));
			}

			sb.Append(')');
			return sb.ToString();
		}

		private static readonly char[] m_vSplit = new char[] { 'x', 'X', ',', ';', ' ' };
		private static readonly char[] m_vTrimWrap = new char[] { '(', ')' };
		public static DmMultiZnGroupElement Parse(string str,
			IDmGroup<DmMultiZnGroupElement> grp)
		{
			if(str == null) throw new ArgumentNullException("str");

			str = str.Trim(m_vTrimWrap);
			if(str.Equals("id", StrUtil.CaseIgnoreCmp))
				return grp.Identity;

			string[] vValues = str.Split(m_vSplit, StringSplitOptions.RemoveEmptyEntries);
			BigInteger[] vNums = new BigInteger[vValues.Length];

			for(int i = 0; i < vValues.Length; ++i)
				vNums[i] = NumUtil.Parse(vValues[i]);

			return new DmMultiZnGroupElement(vNums);
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			int h = m_vValues[0].GetHashCode();
			for(int i = 1; i < m_vValues.Length; ++i)
			{
				h += (h << 10);
				h ^= (h >> 6);
				h += m_vValues[i].GetHashCode();
			}

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			var other = (obj as DmMultiZnGroupElement);
			return (CompareTo(other) == 0);
		}

		public bool Equals(DmMultiZnGroupElement other)
		{
			return (CompareTo(other) == 0);
		}

		public int CompareTo(DmMultiZnGroupElement other)
		{
			if(other == null) return 1;

			BigInteger[] vOther = other.m_vValues;

			if(m_vValues.Length != vOther.Length)
			{
				Debug.Assert(false);
				return ((m_vValues.Length > vOther.Length) ? 1 : -1);
			}

			for(int i = 0; i < m_vValues.Length; ++i)
			{
				int c = m_vValues[i].CompareTo(vOther[i]);
				if(c != 0) return c;
			}

			return 0;
		}
	}
}
