﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Util;

namespace LogSig.DM
{
	public static class NumUtil
	{
		public static int GetPow(BigInteger n, int p)
		{
			if(n < BigInteger.Zero) throw new ArgumentOutOfRangeException("n");
			if(p <= 1) throw new ArgumentOutOfRangeException("p");

			if(n.IsZero) return -1;

			BigInteger bp = p;

			int c = 0;
			while((n % bp).IsZero)
			{
				++c;
				n /= bp;
			}

			return c;
		}

		public static int CountOneBits(BigInteger n)
		{
			if(n < BigInteger.Zero) throw new ArgumentOutOfRangeException("n");

			int b = 0;
			while(n != BigInteger.Zero)
			{
				if((n & 1) != BigInteger.Zero) ++b;
				n >>= 1;
			}

			return b;
		}

		public static List<int> GetOneBitIndices(int n)
		{
			if(n < 0) throw new ArgumentOutOfRangeException("n");

			List<int> v = new List<int>();
			int p = 0;

			while(n != 0)
			{
				if((n & 1) != 0) v.Add(p);

				++p;
				n >>= 1;
			}

			return v;
		}

		public static List<BigInteger> Factorize(BigInteger u)
		{
			if(u < 0) throw new ArgumentNullException("u");

			List<BigInteger> l = new List<BigInteger>();
			if(u <= 3) { l.Add(u); return l; }

			BigInteger d = 2;
			while(u != 1)
			{
				if((u % d).IsZero)
				{
					l.Add(d);
					u /= d;
				}
				else ++d;
			}

			return l;
		}

		public static string FactorsToString(List<BigInteger> vFactors)
		{
			if((vFactors == null) || (vFactors.Count == 0)) return string.Empty;

			StringBuilder sb = new StringBuilder();

			int i = 0, c = 0;
			BigInteger l = -1;
			while(i <= vFactors.Count)
			{
				BigInteger p = ((i < vFactors.Count) ? vFactors[i] : -1);
				if((p != l) && (l != -1))
				{
					if(sb.Length > 0) sb.Append('*');
					sb.Append(l.ToString());
					if(c > 1) sb.Append(@"^" + c.ToString());

					c = 0;
				}

				++c;
				l = p;
				++i;
			}

			return sb.ToString();
		}

		public static string FactorsToString(BigInteger bi)
		{
			return FactorsToString(bi, false);
		}

		public static string FactorsToString(BigInteger bi, bool bOnlyIfCmdLineOpt)
		{
			if(bOnlyIfCmdLineOpt && !Program.FactorizeNumbers)
				return bi.ToString();

			List<BigInteger> v = Factorize(bi);
			return FactorsToString(v);
		}

		public static bool IsNumber(string str)
		{
			if(string.IsNullOrEmpty(str)) return false;

			for(int i = 0; i < str.Length; ++i)
			{
				char ch = str[i];
				if(((ch >= '0') && (ch <= '9')) || (ch == '*') ||
					(ch == '^')) { }
				else return false;
			}

			return true;
		}

		public static BigInteger? GetPrefixedNumber(string str, string strPrefix)
		{
			if(!str.StartsWith(strPrefix, StrUtil.CaseIgnoreCmp))
				return null;

			string strNumber = str.Substring(strPrefix.Length);
			if(!NumUtil.IsNumber(strNumber)) return null;

			return NumUtil.Parse(strNumber);
		}

		public static BigInteger Parse(string str)
		{
			if(string.IsNullOrEmpty(str)) { Debug.Assert(false); return BigInteger.Zero; }

			str += "*"; // Append sentinel

			BigInteger biNumber = BigInteger.One;
			StringBuilder sbFactor = new StringBuilder();
			StringBuilder sbPow = new StringBuilder();
			bool bPow = false;
			for(int i = 0; i < str.Length; ++i)
			{
				char ch = str[i];

				if((ch >= '0') && (ch <= '9'))
				{
					if(bPow) sbPow.Append(ch);
					else sbFactor.Append(ch);
				}
				else if(ch == '*')
				{
					if(bPow)
					{
						BigInteger biFactor = BigInteger.Parse(sbFactor.ToString());
						int nPow = int.Parse(sbPow.ToString());

						biNumber *= BigInteger.Pow(biFactor, nPow);
					}
					else biNumber *= BigInteger.Parse(sbFactor.ToString());

					sbFactor.Clear();
					sbPow.Clear();
					bPow = false;
				}
				else if(ch == '^') bPow = true;
				else throw new FormatException("Cannot parse number!");
			}

			return biNumber;
		}

		public static void NextPermutation(int[] v)
		{
			int n = v.Length;

			int i = n - 1;
			while(true)
			{
				if(i <= 0) // Initial permutation follows last one
				{
					for(int a = 0; a < n; ++a) v[a] = a;
					return;
				}

				if(v[i - 1] < v[i]) break;
				--i;
			}

			int j = n;
			while(v[j - 1] <= v[i - 1]) --j;

			int t = v[i - 1];
			v[i - 1] = v[j - 1];
			v[j - 1] = t;

			++i;
			j = n;
			while(i < j) // Reverse
			{
				t = v[i - 1];
				v[i - 1] = v[j - 1];
				v[j - 1] = t;

				++i;
				--j;
			}
		}

		public static BigInteger Factorial(BigInteger n)
		{
			if(n < 0) throw new ArgumentOutOfRangeException("n");
			if(n <= 1) return BigInteger.One;

			BigInteger p = BigInteger.One;
			while(!n.IsOne) { p *= n; --n; }

			return p;
		}

		public static int RotL32(int x, int b)
		{
			unchecked
			{
				return ((x << b) | (x >> (32 - b)));
			}
		}

		// Compute \\mathfrak{d}(n, m)
		public static KeyValuePair<int, int> CountDiffModSets(int n, int m)
		{
			if(m <= 1) return new KeyValuePair<int, int>(1, 1);

			int[] v = new int[m];
			Array.Clear(v, 0, v.Length);

			HashSet<string> hsSets = new HashSet<string>();
			HashSet<string> hsSetsWithZ = new HashSet<string>();
			CountDiffModSetsRec(n, v, 1, hsSets, hsSetsWithZ);
			return new KeyValuePair<int, int>(hsSets.Count, hsSetsWithZ.Count);
		}

		private static void CountDiffModSetsRec(int n, int[] v, int y,
			HashSet<string> hsSets, HashSet<string> hsSetsWithZ)
		{
			int m = v.Length;
			for(int t = 1; t < n; ++t)
			{
				int tr = (t % m);
				if(v.Any(el => ((el % m) == tr))) continue;

				v[y] = t;
				if(y == (m - 1))
					CountDiffModSetsValidateAdd(n, v, hsSets, hsSetsWithZ);
				else CountDiffModSetsRec(n, v, y + 1, hsSets, hsSetsWithZ);
			}

			v[y] = 0; // Restore
		}

		private static void CountDiffModSetsValidateAdd(int n, int[] v,
			HashSet<string> hsSets, HashSet<string> hsSetsWithZ)
		{
			Debug.Assert(v.Select(x => (x % v.Length)).Distinct().Count() ==
				v.Length);

			foreach(int z in v)
			{
				bool bValid = true;
				foreach(int x in v)
				{
					if(Array.IndexOf<int>(v, (z - x + n) % n) < 0)
					{
						bValid = false;
						break;
					}
				}

				if(bValid)
				{
					int[] vs = new int[v.Length];
					Array.Copy(v, vs, v.Length);
					Array.Sort<int>(vs);
					string str = StrUtil.ToString(vs);
					hsSets.Add(str);
					hsSetsWithZ.Add(z.ToString() + " : " + str);
				}
			}
		}
	}
}
