﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;
using LogSig.DM.LsGen.GroupImplCaps;

namespace LogSig.DM.QuaternionGroup
{
	/// <summary>
	/// Generalized quaternion group.
	/// </summary>
	public sealed partial class DmQuaternionGroup : IDmGroup<DmQuaternionGroupElement>,
		IFacPermShuffleTest<DmQuaternionGroupElement>
	{
		public string Name
		{
			get { return ("Q2*" + NumUtil.FactorsToString(m_nRotMod, true)); }
		}

		private readonly BigInteger m_nOrder;
		private readonly BigInteger m_nRotMod; // = m_nOrder / 2
		private readonly BigInteger m_nHalfMod; // = m_nOrder / 4
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_nOrder; }
		}

		private static readonly DmQuaternionGroupElement m_id =
			new DmQuaternionGroupElement(BigInteger.Zero, false);
		public DmQuaternionGroupElement Identity
		{
			get { return m_id; }
		}

		private static readonly DmQuaternionGroupElement m_im =
			new DmQuaternionGroupElement(BigInteger.Zero, true);

		public DmQuaternionGroup(BigInteger nOrder)
		{
			if((nOrder < 4) || ((nOrder % 4) != 0))
				throw new ArgumentOutOfRangeException("nOrder");

			m_nOrder = nOrder;
			m_nRotMod = (nOrder / 2);
			m_nHalfMod = (nOrder / 4);
		}

		public DmQuaternionGroupElement ParseElement(string str)
		{
			return DmQuaternionGroupElement.Parse(str);
		}

		public DmQuaternionGroupElement Multiply(DmQuaternionGroupElement a,
			DmQuaternionGroupElement b)
		{
			BigInteger aRot = a.Rotation, bRot = b.Rotation;
			bool aIm = a.Imaginary, bIm = b.Imaginary;

			if(aIm)
			{
				if(bIm)
					return new DmQuaternionGroupElement((aRot - bRot +
						m_nRotMod + m_nHalfMod) % m_nRotMod, false);

				return new DmQuaternionGroupElement((aRot - bRot +
					m_nRotMod) % m_nRotMod, true);
			}

			return new DmQuaternionGroupElement((aRot + bRot) % m_nRotMod, bIm);
		}

		public DmQuaternionGroupElement Invert(DmQuaternionGroupElement t)
		{
			if(!t.Imaginary)
			{
				if(t.Rotation.IsZero) return t;
				return new DmQuaternionGroupElement(m_nRotMod - t.Rotation, false);
			}

			return new DmQuaternionGroupElement((t.Rotation + m_nHalfMod) %
				m_nRotMod, true);
		}

		public DmQuaternionGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			BigInteger nRot = rand.GenerateBigInt(m_nRotMod);
			bool bIm = rand.GenerateBool();

			return new DmQuaternionGroupElement(nRot, bIm);
		}

		public DmQuaternionGroupElement NextElement(DmQuaternionGroupElement t)
		{
			BigInteger nNewRot = ((t.Rotation + 1) % m_nRotMod);
			if(nNewRot != 0) return new DmQuaternionGroupElement(nNewRot, t.Imaginary);

			return new DmQuaternionGroupElement(0, !t.Imaginary);
		}

		public BigInteger ElementToBigInt(DmQuaternionGroupElement t)
		{
			return ((t.Rotation << 1) + (t.Imaginary ? 1 : 0));
		}

		public DmQuaternionGroupElement BigIntToElement(BigInteger biElem)
		{
			return new DmQuaternionGroupElement(biElem >> 1, ((biElem & 1) != 0));
		}

		public DmBlockSeq<DmQuaternionGroupElement> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmQuaternionGroupElement>();

			BigInteger n = m_nRotMod, p = 2, pp = 1;
			while(n != 1)
			{
				if((n % p).IsZero)
				{
					var blk = ls.AddBlock();
					for(int i = 0; i < p; ++i)
						blk.Elements.Add(new DmQuaternionGroupElement(pp * i, false));

					pp *= p;
					n /= p;
					continue;
				}

				++p;
			}

			ls.AddBlock(m_id, m_im);
			return ls;
		}

		public int[] FactorInDefaultSig(DmQuaternionGroupElement t)
		{
			DmBlockSeq<DmQuaternionGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			BigInteger biRot = t.Rotation;
			for(int i = 0; i < vFactorization.Length - 1; ++i)
			{
				int p = ls.Blocks[i].Elements.Count;
				vFactorization[i] = (int)(biRot % p);
				biRot /= p;
			}

			vFactorization[vFactorization.Length - 1] = (t.Imaginary ? 1 : 0);
			return vFactorization;
		}

		public bool IsFacPermBlockSeq(List<DmLogSigBlock<
			DmQuaternionGroupElement>> lBlocks)
		{
			for(int x = 0; x < lBlocks.Count; ++x)
			{
				foreach(DmQuaternionGroupElement el in lBlocks[x].Elements)
				{
					if(el.Imaginary) return false;
				}
			}

			return true;
		}
	}
}
