﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;

using LogSig.Interfaces;

namespace LogSig.DM.QuaternionGroup
{
	/// <summary>
	/// Immutable element of a generalized quaternion group.
	/// </summary>
	public sealed class DmQuaternionGroupElement : IDmGroupElement<DmQuaternionGroupElement>
	{
		private int? m_hash = null;

		private readonly BigInteger m_biRot;
		public BigInteger Rotation
		{
			get { return m_biRot; }
			// set { m_biRot = value; } // Immutable
		}

		private readonly bool m_bIm;
		public bool Imaginary
		{
			get { return m_bIm; }
			// set { m_bIm = value; } // Immutable
		}

		public DmQuaternionGroupElement()
		{
			m_biRot = 0;
			m_bIm = false;
		}

		public DmQuaternionGroupElement(BigInteger iRotation, bool bImaginary)
		{
			m_biRot = iRotation;
			m_bIm = bImaginary;
		}

		public DmQuaternionGroupElement Clone()
		{
			return (DmQuaternionGroupElement)MemberwiseClone();
		}

		public override string ToString()
		{
			if(m_biRot == 0) return (m_bIm ? "0t" : "id");

			string str = NumUtil.FactorsToString(m_biRot, true);
			return (str + (m_bIm ? "t" : string.Empty));
		}

		public static DmQuaternionGroupElement Parse(string str)
		{
			if(str == null) throw new ArgumentNullException("str");

			str = str.Trim().ToLower();
			if(str.Equals("id")) return new DmQuaternionGroupElement(BigInteger.Zero, false);

			bool bIm = false;
			if(str.EndsWith("t"))
			{
				bIm = true;
				str = str.Substring(0, str.Length - 1);
			}

			return new DmQuaternionGroupElement(NumUtil.Parse(str), bIm);
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			int h = m_biRot.GetHashCode();
			if(m_bIm) h ^= 0x5E1E7079;

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			DmQuaternionGroupElement other = (obj as DmQuaternionGroupElement);
			return Equals(other);
		}

		public bool Equals(DmQuaternionGroupElement other)
		{
			if(other == null) return false;
			return ((m_bIm == other.m_bIm) && (m_biRot == other.m_biRot));
		}

		public int CompareTo(DmQuaternionGroupElement other)
		{
			if(other == null) return 1;

			if(m_bIm && !other.m_bIm) return 1;
			if(!m_bIm && other.m_bIm) return -1;
			if(m_biRot == other.m_biRot) return 0;

			return ((m_biRot < other.m_biRot) ? -1 : 1);
		}
	}
}
