﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;

namespace LogSig.DM.SnGroup
{
	public sealed class DmSnGroup : IDmGroup<DmSnGroupElement>
	{
		public string Name
		{
			get { return ("S" + m_nDegree.ToString()); }
		}
		// public string ShortName { get { return "Sym"; } }
		// public string ShortNameAlt { get { return "Sn"; } }

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private readonly int m_nDegree;
		/// <summary>
		/// Degree of the group (size of the set that the permutation
		/// group acts on).
		/// </summary>
		public int Degree
		{
			get { return m_nDegree; }
		}

		private readonly DmSnGroupElement m_id;
		public DmSnGroupElement Identity { get { return m_id; } }

		public DmSnGroup(BigInteger biOrder)
		{
			if(biOrder <= 0) throw new ArgumentOutOfRangeException("biOrder");

			m_biOrder = biOrder;
			m_nDegree = 1;

			BigInteger p = BigInteger.One;
			while(true)
			{
				if(p > biOrder) throw new ArgumentOutOfRangeException("biOrder");
				if(p == biOrder) break;

				++m_nDegree;
				p *= m_nDegree;
			}

			m_id = new DmSnGroupElement(m_nDegree);
		}

		public DmSnGroupElement ParseElement(string str)
		{
			return DmSnGroupElement.Parse(str, m_nDegree);
		}

		public DmSnGroupElement Multiply(DmSnGroupElement a, DmSnGroupElement b)
		{
			int[] vA = a.Permutation;
			int[] vB = b.Permutation;
			int[] v = new int[m_nDegree];
			for(int i = 0; i < m_nDegree; ++i) v[i] = vA[vB[i]];

			return new DmSnGroupElement(v);
		}

		public DmSnGroupElement Invert(DmSnGroupElement t)
		{
			int[] vT = t.Permutation;
			int[] v = new int[m_nDegree];
			for(int i = 0; i < m_nDegree; ++i) v[vT[i]] = i;

			return new DmSnGroupElement(v);
		}

		public DmSnGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			int[] v = new int[m_nDegree];
			for(int a = 0; a < m_nDegree; ++a) v[a] = a;

			for(int i = (m_nDegree - 1); i >= 1; --i)
			{
				int p = rand.GenerateInt32(i + 1);

				int t = v[p];
				v[p] = v[i];
				v[i] = t;
			}

			return new DmSnGroupElement(v);
		}

		public DmSnGroupElement NextElement(DmSnGroupElement t)
		{
			int[] v = new int[m_nDegree];
			Array.Copy(t.Permutation, v, m_nDegree);
			NumUtil.NextPermutation(v);

			return new DmSnGroupElement(v);
		}

		public static BigInteger ElementToBigInt(int[] vPerm)
		{
			BigInteger bi = BigInteger.Zero;

			List<int> l = new List<int>(vPerm);
			for(int i = 0; ; ++i)
			{
				if(l.Count <= 1) break;

				int p = l.IndexOf(i);
				Debug.Assert(p >= 0);

				bi *= l.Count;
				bi += p;

				l.RemoveAt(p);
			}

			return bi;
		}

		public BigInteger ElementToBigInt(DmSnGroupElement t)
		{
			return ElementToBigInt(t.Permutation);
		}

		public DmSnGroupElement BigIntToElement(BigInteger biElem)
		{
			List<int> vIndices = new List<int>();
			for(int i = 0; i < m_nDegree; ++i)
			{
				vIndices.Add((int)(biElem % (i + 1)));
				biElem /= (i + 1);
			}

			int[] v = new int[m_nDegree];
			for(int a = 0; a < m_nDegree; ++a) v[a] = -1;
			for(int j = 0; j < m_nDegree; ++j)
			{
				int vp = 0;
				for(int s = 0; s < m_nDegree; ++s)
				{
					if(v[s] < 0)
					{
						if(vp == vIndices[m_nDegree - j - 1])
						{
							v[s] = j;
							break;
						}

						++vp;
					}
				}
			}

			return new DmSnGroupElement(v);
		}

		public DmBlockSeq<DmSnGroupElement> CreateDefaultSig()
		{
			DmBlockSeq<DmSnGroupElement> ls = new DmBlockSeq<DmSnGroupElement>();

			for(int i = 1; i < m_nDegree; ++i)
			{
				var blk = ls.AddBlock(m_id);

				for(int j = 0; j < i; ++j)
				{
					int[] v = new int[m_nDegree];
					for(int a = 0; a < m_nDegree; ++a) v[a] = a;
					int t = v[i];
					v[i] = v[j];
					v[j] = t;

					blk.Elements.Add(new DmSnGroupElement(v));
				}
			}

			return ls;
		}

		public int[] FactorInDefaultSig(DmSnGroupElement t)
		{
			DmBlockSeq<DmSnGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			DmSnGroupElement elToFactor = t;
			for(int i = (vFactorization.Length - 1); i >= 0; --i)
			{
				int[] v = elToFactor.Permutation;
				int p = Array.IndexOf<int>(v, i + 1);
				Debug.Assert(p >= 0);

				if(p == (i + 1)) vFactorization[i] = 0;
				else
				{
					vFactorization[i] = p + 1;
					elToFactor = Multiply(elToFactor,
						Invert(ls.Blocks[i].Elements[p + 1]));
				}
			}

			return vFactorization;
		}
	}
}
