﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;

using LogSig.Util;

namespace LogSig.DM.SnGroup
{
	public class DmSnGroupElement : IDmGroupElement<DmSnGroupElement>
	{
		private int? m_hash = null;

		private readonly int[] m_vPerm;
		public int[] Permutation
		{
			get { return m_vPerm; }
		}

		private int m_iSign = 0;
		/// <summary>
		/// Sign of the permutation (-1 for odd, 1 for even).
		/// </summary>
		public int Sign
		{
			get
			{
				if(m_iSign == 0) m_iSign = ComputeSign(m_vPerm);
				return m_iSign;
			}
		}

		public DmSnGroupElement()
		{
			throw new InvalidOperationException(); // Order unknown
		}

		public DmSnGroupElement(int nDegree)
		{
			if(nDegree <= 0) throw new ArgumentOutOfRangeException("nDegree");

			m_vPerm = new int[nDegree];
			for(int i = 0; i < nDegree; ++i) m_vPerm[i] = i; // Identity
		}

		public DmSnGroupElement(int[] vPerm)
		{
			if(vPerm == null) throw new ArgumentNullException("vPerm");
			if(vPerm.Length == 0) throw new ArgumentOutOfRangeException("vPerm");

			m_vPerm = new int[vPerm.Length];
			Array.Copy(vPerm, m_vPerm, vPerm.Length);
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			// The following hashing method would be nice, but
			// unfortunately is too slow...
			// 2147483647 is the largest prime fitting into 31 bits
			// return (int)(DmSnGroup.ElementToBigInt(m_vPerm) % 2147483647);

			int h = 0;
			for(int i = 0; i < m_vPerm.Length; ++i)
			{
				h += m_vPerm[i];
				h += (h << 10);
				h ^= (h >> 6);
			}

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			return (CompareTo(obj as DmSnGroupElement) == 0);
		}

		public bool Equals(DmSnGroupElement other)
		{
			return (CompareTo(other) == 0);
		}

		public int CompareTo(DmSnGroupElement other)
		{
			if(other == null) return 1;

			if(other.m_vPerm.Length != m_vPerm.Length)
				throw new ArgumentOutOfRangeException("other");

			for(int i = 0; i < m_vPerm.Length; ++i)
			{
				if(m_vPerm[i] > other.m_vPerm[i]) return 1;
				if(m_vPerm[i] < other.m_vPerm[i]) return -1;
			}

			return 0;
		}

		public DmSnGroupElement Clone()
		{
			return new DmSnGroupElement(m_vPerm);
		}

		public override string ToString()
		{
			bool[] vDone = new bool[m_vPerm.Length];
			Array.Clear(vDone, 0, vDone.Length);

			StringBuilder sb = new StringBuilder();
			while(true)
			{
				int s = Array.IndexOf<bool>(vDone, false);
				if(s < 0) break;

				vDone[s] = true;
				if(m_vPerm[s] == s) continue;

				sb.Append('(');
				sb.Append(s);

				int p = s;
				while(true)
				{
					p = m_vPerm[p];
					if(p == s) break;

					Debug.Assert(!vDone[p]);
					vDone[p] = true;

					sb.Append(' ');
					sb.Append(p);
				}

				sb.Append(')');
			}

			if(sb.Length == 0) return "id";
			return sb.ToString();
		}

		public static DmSnGroupElement Parse(string str, int nDegree)
		{
			if(nDegree <= 0) throw new ArgumentOutOfRangeException("nDegree");

			if(str.Equals("id", StrUtil.CaseIgnoreCmp))
				return new DmSnGroupElement(nDegree);

			int[] vPerm = new int[nDegree];
			for(int a = 0; a < vPerm.Length; ++a) vPerm[a] = a;

			string[] vCycles = str.Split(new char[] { '(', ')' },
				StringSplitOptions.RemoveEmptyEntries);
			foreach(string strCycle in vCycles)
			{
				string[] vElements = strCycle.Split(new char[] { ' ' });

				int s = -1, p = -1;
				foreach(string strElement in vElements)
				{
					int e = int.Parse(strElement);

					if(s < 0) s = e;
					if(p >= 0) vPerm[p] = e;
					p = e;
				}

				if((p >= 0) && (s >= 0)) vPerm[p] = s;
				else { Debug.Assert(false); }
			}

			return new DmSnGroupElement(vPerm);
		}

		public static int ComputeSign(int[] vPerm)
		{
			bool[] vDone = new bool[vPerm.Length];
			// Array.Clear(vDone, 0, vDone.Length);
			Debug.Assert(Array.TrueForAll<bool>(vDone, el => !el));

			int iSign = 1;
			while(true)
			{
				int s = Array.IndexOf<bool>(vDone, false);
				if(s < 0) break;

				vDone[s] = true;
				if(vPerm[s] == s) continue;

				int p = s, l = 0;
				while(true)
				{
					p = vPerm[p];
					++l;
					if(p == s) break;

					Debug.Assert(!vDone[p]);
					vDone[p] = true;
				}

				if((l % 2) == 0) iSign *= -1;
			}

			return iSign;
		}
	}
}
