﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Linq;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;

namespace LogSig.DM.WreathProductGroup
{
	public sealed class DmWreathProductGroup<H, P> :
		IDmGroup<DmWreathProductGroupElement<H, P>>
		where H : IDmGroupElement<H>
		where P : IDmGroupElement<P>
	{
		public string Name
		{
			get { return (m_h.Name + "w" + m_p.Name); }
		}
		// public string ShortName { get { return "WP"; } }
		// public string ShortNameAlt { get { return "w"; } }

		private readonly IDmGroup<H> m_h;
		private readonly IDmGroup<P> m_p;

		public IDmGroup<H> BaseGroup { get { return m_h; } }
		public IDmGroup<P> PermGroup { get { return m_p; } }

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private readonly DmWreathProductGroupElement<H, P> m_id;
		public DmWreathProductGroupElement<H, P> Identity { get { return m_id; } }

		private SortedDictionary<P, int> m_dPToIndex =
			new SortedDictionary<P, int>();
		private SortedDictionary<P, int[]> m_dPToPerm =
			new SortedDictionary<P, int[]>();

		public DmWreathProductGroup(IDmGroup<H> h, IDmGroup<P> p)
		{
			if(h == null) throw new ArgumentNullException("h");
			if(p == null) throw new ArgumentNullException("p");

			m_h = h;
			m_p = p;
			m_biOrder = BigInteger.Pow(m_h.Order, (int)m_p.Order) * m_p.Order;

			H[] v = new H[(int)m_p.Order];
			for(int i = 0; i < v.Length; ++i) v[i] = m_h.Identity;
			m_id = new DmWreathProductGroupElement<H, P>(v, m_p.Identity);

			P pIt = m_p.Identity;
			for(int i = 0; i < (int)m_p.Order; ++i)
			{
				m_dPToIndex[pIt] = i;
				pIt = m_p.NextElement(pIt);
			}
			Debug.Assert(pIt.Equals(m_p.Identity));

			foreach(KeyValuePair<P, int> kvp in m_dPToIndex)
			{
				int[] vPerm = new int[(int)m_p.Order];

				foreach(KeyValuePair<P, int> kvpIt in m_dPToIndex)
				{
					P pProd = m_p.Multiply(kvp.Key, kvpIt.Key);
					vPerm[m_dPToIndex[pProd]] = m_dPToIndex[kvpIt.Key]; // Inverse
				}
				Debug.Assert(vPerm.Sum() == (((vPerm.Length - 1) * vPerm.Length) / 2));

				m_dPToPerm[kvp.Key] = vPerm;
			}
		}

		public DmWreathProductGroupElement<H, P> ParseElement(string str)
		{
			return DmWreathProductGroupElement<H, P>.Parse(str, m_h, m_p);
		}

		public DmWreathProductGroupElement<H, P> Multiply(DmWreathProductGroupElement<H, P> a,
			DmWreathProductGroupElement<H, P> b)
		{
			int[] vPerm = m_dPToPerm[a.Perm];
			H[] v = new H[a.Base.Length];
			for(int i = 0; i < v.Length; ++i)
				v[i] = m_h.Multiply(a.Base[i], b.Base[vPerm[i]]);

			return new DmWreathProductGroupElement<H, P>(v, m_p.Multiply(a.Perm, b.Perm));
		}

		public DmWreathProductGroupElement<H, P> Invert(DmWreathProductGroupElement<H, P> t)
		{
			int[] vPerm = m_dPToPerm[t.Perm];
			H[] v = new H[t.Base.Length];
			for(int i = 0; i < v.Length; ++i)
				v[vPerm[i]] = m_h.Invert(t.Base[i]);

			return new DmWreathProductGroupElement<H, P>(v, m_p.Invert(t.Perm));
		}

		public DmWreathProductGroupElement<H, P> GenerateRandomElement(CryptoRandomGen rand)
		{
			H[] v = new H[(int)m_p.Order];
			for(int i = 0; i < v.Length; ++i)
				v[i] = m_h.GenerateRandomElement(rand);
			return new DmWreathProductGroupElement<H, P>(v,
				m_p.GenerateRandomElement(rand));
		}

		public DmWreathProductGroupElement<H, P> NextElement(DmWreathProductGroupElement<H, P> t)
		{
			H[] v = (H[])t.Base.Clone();
			for(int i = 0; i < v.Length; ++i)
			{
				v[i] = m_h.NextElement(v[i]);
				if(!v[i].Equals(m_h.Identity))
					return new DmWreathProductGroupElement<H, P>(v, t.Perm);
			}

			return new DmWreathProductGroupElement<H, P>(v, m_p.NextElement(t.Perm));
		}

		public BigInteger ElementToBigInt(DmWreathProductGroupElement<H, P> t)
		{
			BigInteger bi = BigInteger.Zero;

			for(int i = 0; i < t.Base.Length; ++i)
			{
				bi *= m_h.Order;
				bi += m_h.ElementToBigInt(t.Base[i]);
			}

			bi *= m_p.Order;
			bi += m_p.ElementToBigInt(t.Perm);

			return bi;
		}

		public DmWreathProductGroupElement<H, P> BigIntToElement(BigInteger biElem)
		{
			P p = m_p.BigIntToElement(biElem % m_p.Order);
			biElem /= m_p.Order;

			H[] v = new H[(int)m_p.Order];
			for(int i = (v.Length - 1); i >= 0; --i)
			{
				v[i] = m_h.BigIntToElement(biElem % m_h.Order);
				biElem /= m_h.Order;
			}
			Debug.Assert(biElem.IsZero);

			return new DmWreathProductGroupElement<H, P>(v, p);
		}

		public DmBlockSeq<DmWreathProductGroupElement<H, P>> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmWreathProductGroupElement<H, P>>();

			H[] vId = new H[(int)m_p.Order];
			for(int e = 0; e < vId.Length; ++e) vId[e] = m_h.Identity;

			var lsH = m_h.CreateDefaultSig();
			for(int i = 0; i < (int)m_p.Order; ++i)
			{
				for(int x = 0; x < lsH.Blocks.Count; ++x)
				{
					var blk = lsH.Blocks[x].Elements;
					ls.Blocks.Add(new DmLogSigBlock<DmWreathProductGroupElement<H, P>>());
					for(int y = 0; y < blk.Count; ++y)
					{
						H[] v = (H[])vId.Clone();
						v[i] = blk[y];
						ls.Blocks[ls.Blocks.Count - 1].Elements.Add(
							new DmWreathProductGroupElement<H, P>(
							v, m_p.Identity));
					}
				}
			}

			var lsP = m_p.CreateDefaultSig();
			for(int x = 0; x < lsP.Blocks.Count; ++x)
			{
				var blk = lsP.Blocks[x].Elements;
				ls.Blocks.Add(new DmLogSigBlock<DmWreathProductGroupElement<H, P>>());
				for(int y = 0; y < blk.Count; ++y)
				{
					ls.Blocks[ls.Blocks.Count - 1].Elements.Add(
						new DmWreathProductGroupElement<H, P>(vId, blk[y]));
				}
			}

			return ls;
		}

		public int[] FactorInDefaultSig(DmWreathProductGroupElement<H, P> t)
		{
			throw new NotImplementedException();
		}
	}
}
