﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Util;

namespace LogSig.DM.WreathProductGroup
{
	public sealed class DmWreathProductGroupElement<H, P> :
		IDmGroupElement<DmWreathProductGroupElement<H, P>>
		where H : IDmGroupElement<H>
		where P : IDmGroupElement<P>
	{
		private int? m_hash = null;

		private readonly H[] m_v;
		public H[] Base
		{
			get { return m_v; }
			// set { m_v = value; } // Immutable
		}

		private readonly P m_p;
		public P Perm
		{
			get { return m_p; }
			// set { m_p = value; } // Immutable
		}

		public DmWreathProductGroupElement(H[] v, P p)
		{
			if(v == null) throw new ArgumentNullException("v");
			if(v.Length < 1) throw new ArgumentOutOfRangeException("v");

			m_v = new H[v.Length];
			Array.Copy(v, m_v, v.Length);

			m_p = p;
		}

		public DmWreathProductGroupElement<H, P> Clone()
		{
			return (DmWreathProductGroupElement<H, P>)MemberwiseClone();
		}

		public override string ToString()
		{
			string strSep = ((Program.OutputFormat == OutputFmt.TeX) ? ", " : ",");

			StringBuilder sb = new StringBuilder();
			sb.Append('(');
			for(int i = 0; i < m_v.Length; ++i)
			{
				if(i > 0) sb.Append(strSep);
				sb.Append(m_v[i].ToString());
			}

			sb.Append(strSep);
			sb.Append(m_p.ToString());

			sb.Append(')');
			return sb.ToString();
		}

		private static readonly char[] m_vSplit = new char[] { ',', ';', ' ' };
		public static DmWreathProductGroupElement<H, P> Parse(string str,
			IDmGroup<H> grpH, IDmGroup<P> grpP)
		{
			if(str == null) throw new ArgumentNullException("str");

			str = str.Trim();
			if(!str.StartsWith("(") || !str.EndsWith(")"))
				throw new FormatException("Failed to parse group element!");
			str = str.Substring(1, str.Length - 2);

			List<string> lParts = StrUtil.SplitLevel0(str, m_vSplit);
			if(lParts.Count < 2)
				throw new FormatException("Failed to parse group element!");

			H[] v = new H[lParts.Count - 1];
			for(int i = 0; i < v.Length; ++i)
				v[i] = grpH.ParseElement(lParts[i]);

			return new DmWreathProductGroupElement<H, P>(v,
				grpP.ParseElement(lParts[lParts.Count - 1]));
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			int h = m_p.GetHashCode();
			for(int i = 0; i < m_v.Length; ++i)
			{
				h += (h << 10);
				h ^= (h >> 6);
				h += m_v[i].GetHashCode();
			}

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			var other = (obj as DmWreathProductGroupElement<H, P>);
			return (CompareTo(other) == 0);
		}

		public bool Equals(DmWreathProductGroupElement<H, P> other)
		{
			return (CompareTo(other) == 0);
		}

		public int CompareTo(DmWreathProductGroupElement<H, P> other)
		{
			if(other == null) return 1;

			if(m_v.Length != other.m_v.Length)
			{
				Debug.Assert(false);
				return ((m_v.Length > other.m_v.Length) ? 1 : -1);
			}

			for(int i = 0; i < m_v.Length; ++i)
			{
				int c = m_v[i].CompareTo(other.m_v[i]);
				if(c != 0) return c;
			}

			return m_p.CompareTo(other.m_p);
		}
	}
}
