﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.Crypto;

namespace LogSig.DM.ZnGroup
{
	public sealed class DmZnGroup : IDmGroup<DmZnGroupElement>
	{
		public string Name
		{
			get { return ("Z" + NumUtil.FactorsToString(m_biOrder, true)); }
		}
		// public string ShortName { get { return "Res"; } }
		// public string ShortNameAlt { get { return "Zn"; } }

		private readonly BigInteger m_biOrder;
		/// <summary>
		/// Order of the group (number of elements).
		/// </summary>
		public BigInteger Order
		{
			get { return m_biOrder; }
		}

		private readonly DmZnGroupElement m_id = new DmZnGroupElement();
		public DmZnGroupElement Identity { get { return m_id; } }

		public DmZnGroup(BigInteger biOrder)
		{
			if(biOrder <= 0) throw new ArgumentOutOfRangeException("biOrder");

			m_biOrder = biOrder;
		}

		public DmZnGroupElement ParseElement(string str)
		{
			return DmZnGroupElement.Parse(str);
		}

		public DmZnGroupElement Multiply(DmZnGroupElement a, DmZnGroupElement b)
		{
			return new DmZnGroupElement((a.Value + b.Value) % m_biOrder);
		}

		public DmZnGroupElement Invert(DmZnGroupElement t)
		{
			if(t.Value.IsZero) return t;
			return new DmZnGroupElement(m_biOrder - t.Value);
		}

		public DmZnGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			return new DmZnGroupElement(rand.GenerateBigInt(m_biOrder));
		}

		public DmZnGroupElement NextElement(DmZnGroupElement t)
		{
			return new DmZnGroupElement((t.Value + BigInteger.One) % m_biOrder);
		}

		public BigInteger ElementToBigInt(DmZnGroupElement t)
		{
			return t.Value;
		}

		public DmZnGroupElement BigIntToElement(BigInteger biElem)
		{
			return new DmZnGroupElement(biElem);
		}

		public DmBlockSeq<DmZnGroupElement> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<DmZnGroupElement>();

			BigInteger n = m_biOrder, p = 2, pp = 1;
			while(n != 1)
			{
				if((n % p).IsZero)
				{
					var blk = ls.AddBlock();
					for(int i = 0; i < p; ++i)
						blk.Elements.Add(new DmZnGroupElement(pp * i));

					pp *= p;
					n /= p;
					continue;
				}

				++p;
			}

			return ls;
		}

		public int[] FactorInDefaultSig(DmZnGroupElement t)
		{
			DmBlockSeq<DmZnGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			BigInteger biRot = t.Value;
			for(int i = 0; i < vFactorization.Length; ++i)
			{
				int p = ls.Blocks[i].Elements.Count;
				vFactorization[i] = (int)(biRot % p);
				biRot /= p;
			}

			return vFactorization;
		}
	}
}
