﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;

using LogSig.Interfaces;

namespace LogSig.Libs.LpSolve
{
	public sealed class LpSolve : IIlpSolver, IDisposable
	{
		private enum lpsolve_return : int
		{
			NOMEMORY = -2,
			OPTIMAL = 0,
			SUBOPTIMAL = 1,
			INFEASIBLE = 2,
			UNBOUNDED = 3,
			DEGENERATE = 4,
			NUMFAILURE = 5,
			USERABORT = 6,
			TIMEOUT = 7,
			PRESOLVED = 9,
			PROCFAIL = 10,
			PROCBREAK = 11,
			FEASFOUND = 12,
			NOFEASFOUND = 13
		}

		[DllImport("lpsolve55.dll")]
		private static extern IntPtr make_lp(int rows, int columns);

		[DllImport("lpsolve55.dll")]
		private static extern void delete_lp(IntPtr lp);

		[DllImport("lpsolve55.dll")]
		private static extern byte add_constraint(IntPtr lp, double[] row,
			ConstraintType constr_type, double rh);

		[DllImport("lpsolve55.dll")]
		private static extern lpsolve_return solve(IntPtr lp);

		[DllImport("lpsolve55.dll")]
		private static extern void set_timeout(IntPtr lp, int sectimeout);

		[DllImport("lpsolve55.dll")]
		private static extern int get_Ncolumns(IntPtr lp);

		[DllImport("lpsolve55.dll")]
		private static extern byte get_variables(IntPtr lp, double[] var);

		[DllImport("lpsolve55.dll")]
		private static extern byte set_int(IntPtr lp, int column, byte must_be_int);

		[DllImport("lpsolve55.dll")]
		private static extern byte set_binary(IntPtr lp, int column, byte must_be_bin);

		private IntPtr m_h;
		private int m_nVariables;

		public LpSolve(int nVariables)
		{
			m_h = make_lp(0, nVariables);
			if(m_h == IntPtr.Zero)
				throw new OutOfMemoryException();

			m_nVariables = nVariables;
		}

		~LpSolve()
		{
			Dispose();
		}

		public void Dispose()
		{
			if(m_h != IntPtr.Zero)
			{
				delete_lp(m_h);
				m_h = IntPtr.Zero;
			}
		}

		public bool AddConstraint(double[] vRow, ConstraintType t,
			double rh)
		{
			double[] v = new double[vRow.Length + 1];
			Array.Copy(vRow, 0, v, 1, vRow.Length);

			return (add_constraint(m_h, v, t, rh) != 0);
		}

		public bool SetTypeBinary(int iCol)
		{
			// Debug.Assert(set_binary(m_h, 0, 1) == 0);
			return (set_binary(m_h, iCol + 1, 1) != 0);
		}

		public bool SetTypeInt(int iCol)
		{
			// Debug.Assert(set_int(m_h, 0, 1) == 0);
			return (set_int(m_h, iCol + 1, 1) != 0);
		}

		public double[] Solve()
		{
			set_timeout(m_h, 0);

			lpsolve_return r = solve(m_h);
			if((r != lpsolve_return.OPTIMAL) && (r != lpsolve_return.FEASFOUND))
				throw new Exception("LPSolve error: " + ((int)r).ToString());

			if(get_Ncolumns(m_h) != m_nVariables)
				throw new InvalidOperationException();

			double[] v = new double[m_nVariables];
			if(get_variables(m_h, v) == 0)
				throw new InvalidOperationException();
			return v;
		}

		public int[] SolveInt()
		{
			double[] v = Solve();

			int[] r = new int[v.Length];
			for(int i = 0; i < v.Length; ++i)
				r[i] = (int)Math.Round(v[i]);

			return r;
		}
	}
}
