// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Numerics;

using LogSig.Crypto;
using LogSig.DM;
using LogSig.Plugins;

namespace LsZnGroup
{
	public sealed class LsZnGroupExt : LsPlugin
	{
		public override bool Initialize(ILsPluginHost host)
		{
			if(host == null) return false;

			host.GroupFactories.Add(new ZnGroupFactory());
			return true;
		}
	}

	public sealed class ZnGroupFactory : IDmGroupFactory
	{
		public object CreateInstance(string strGroup)
		{
			BigInteger? biOrder = NumUtil.GetPrefixedNumber(strGroup, "Res");
			if(!biOrder.HasValue) return null;

			return new ZnGroup(biOrder.Value);
		}
	}

	public sealed class ZnGroupElement : IDmGroupElement<ZnGroupElement>
	{
		private int? m_hash = null;

		private readonly BigInteger m_n;
		public BigInteger Value { get { return m_n; } }

		public ZnGroupElement(BigInteger biValue)
		{
			m_n = biValue;
		}

		public ZnGroupElement Clone()
		{
			return (ZnGroupElement)MemberwiseClone();
		}

		public override string ToString()
		{
			return NumUtil.FactorsToString(m_n, true);
		}

		public static ZnGroupElement Parse(string str)
		{
			return new ZnGroupElement(NumUtil.Parse(str));
		}

		public override int GetHashCode()
		{
			if(m_hash.HasValue) return m_hash.Value;

			int h = m_n.GetHashCode();

			m_hash = h;
			return h;
		}

		public override bool Equals(object obj)
		{
			return Equals(obj as ZnGroupElement);
		}

		public bool Equals(ZnGroupElement other)
		{
			if(other == null) return false;
			return (m_n == other.m_n);
		}

		public int CompareTo(ZnGroupElement other)
		{
			if(other == null) return 1;
			return m_n.CompareTo(other.m_n);
		}
	}

	public sealed class ZnGroup : IDmGroup<ZnGroupElement>
	{
		public string Name
		{
			get { return ("Res" + NumUtil.FactorsToString(m_biOrder, true)); }
		}

		private readonly BigInteger m_biOrder;
		public BigInteger Order { get { return m_biOrder; } }

		private ZnGroupElement m_id = new ZnGroupElement(0);
		public ZnGroupElement Identity { get { return m_id; } }

		public ZnGroup(BigInteger biOrder) { m_biOrder = biOrder; }

		public ZnGroupElement ParseElement(string str)
		{
			return ZnGroupElement.Parse(str);
		}

		public ZnGroupElement Multiply(ZnGroupElement a, ZnGroupElement b)
		{
			return new ZnGroupElement((a.Value + b.Value) % m_biOrder);
		}

		public ZnGroupElement Invert(ZnGroupElement t)
		{
			if(t.Value.IsZero) return t;
			return new ZnGroupElement(m_biOrder - t.Value);
		}

		public ZnGroupElement GenerateRandomElement(CryptoRandomGen rand)
		{
			return new ZnGroupElement(rand.GenerateBigInt(m_biOrder));
		}

		public ZnGroupElement NextElement(ZnGroupElement t)
		{
			return new ZnGroupElement((t.Value + BigInteger.One) % m_biOrder);
		}

		public BigInteger ElementToBigInt(ZnGroupElement t)
		{
			return t.Value;
		}

		public ZnGroupElement BigIntToElement(BigInteger biElem)
		{
			return new ZnGroupElement(biElem);
		}

		public DmBlockSeq<ZnGroupElement> CreateDefaultSig()
		{
			var ls = new DmBlockSeq<ZnGroupElement>();

			BigInteger n = m_biOrder, p = 2, pp = 1;
			while(n != 1)
			{
				if((n % p).IsZero)
				{
					var blk = ls.AddBlock();
					for(int i = 0; i < p; ++i)
						blk.Elements.Add(new ZnGroupElement(pp * i));

					pp *= p;
					n /= p;
					continue;
				}

				++p;
			}

			return ls;
		}

		public int[] FactorInDefaultSig(ZnGroupElement t)
		{
			DmBlockSeq<ZnGroupElement> ls = CreateDefaultSig();

			int[] vFactorization = new int[ls.Blocks.Count];

			BigInteger biRot = t.Value;
			for(int i = 0; i < vFactorization.Length; ++i)
			{
				int p = ls.Blocks[i].Elements.Count;
				vFactorization[i] = (int)(biRot % p);
				biRot /= p;
			}

			return vFactorization;
		}
	}
}
