﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.Security.Cryptography;

using LogSig.DM;
using LogSig.Util;

namespace LogSig.Pkc
{
	public sealed class PkcPrivateKey : PkcPublicKey
	{
		private string m_strKey = string.Empty;
		/// <summary>
		/// Base-64-encoded key for the AES-CTR CSPRNG used by LS-Gen;
		/// encrypted using the master password.
		/// </summary>
		public string GeneratorKey
		{
			get { return m_strKey; }
			set
			{
				if(value == null) throw new ArgumentNullException("value");
				m_strKey = value;
			}
		}

		private string m_strCtr = string.Empty;
		/// <summary>
		/// Base-64-encoded counter for the AES-CTR CSPRNG used by LS-Gen;
		/// encrypted using the master password.
		/// </summary>
		public string GeneratorCtr
		{
			get { return m_strCtr; }
			set
			{
				if(value == null) throw new ArgumentNullException("value");
				m_strCtr = value;
			}
		}

		private ulong m_uRounds = 5000;
		public ulong GeneratorRounds
		{
			get { return m_uRounds; }
			set { m_uRounds = value; }
		}

		private string m_strTrfIORels = string.Empty;
		public string GeneratorTrfIORels
		{
			get { return m_strTrfIORels; }
			set
			{
				if(value == null) throw new ArgumentNullException("value");
				m_strTrfIORels = value;
			}
		}

		public PkcPrivateKey() { }

		public static PkcPrivateKey CreatePrivate<T>(IDmGroup<T> grp, DmBlockSeq<T> ls,
			byte[] pbKey32, byte[] pbCtr16, ulong uRounds, string strTrfIORels,
			string strPassword)
			where T : IDmGroupElement<T>
		{
			if(grp == null) throw new ArgumentNullException("grp");
			if(ls == null) throw new ArgumentNullException("ls");
			if(pbKey32 == null) throw new ArgumentNullException("pbKey32");
			if(pbCtr16 == null) throw new ArgumentNullException("pbCtr16");

			PkcPrivateKey pk = new PkcPrivateKey();
			pk.Group = grp.Name;
			pk.LogSig = ls.ToString();
			pk.GeneratorKey = EncryptKey(pbKey32, strPassword);
			pk.GeneratorCtr = EncryptKey(pbCtr16, strPassword);
			pk.GeneratorRounds = uRounds;
			pk.GeneratorTrfIORels = (strTrfIORels ?? string.Empty);
			return pk;
		}

		private static string EncryptKey(byte[] pbBuf, string strPassword)
		{
			UTF8Encoding utf8 = new UTF8Encoding(false);
			SHA256Managed sha256 = new SHA256Managed();
			byte[] pbKey32 = sha256.ComputeHash(utf8.GetBytes(strPassword));

			return Convert.ToBase64String(MemUtil.EncryptEcb(pbBuf, pbKey32,
				true), Base64FormattingOptions.None);
		}

		public static byte[] DecryptKey(string strKey, string strPassword)
		{
			UTF8Encoding utf8 = new UTF8Encoding(false);
			SHA256Managed sha256 = new SHA256Managed();
			byte[] pbKey32 = sha256.ComputeHash(utf8.GetBytes(strPassword));

			byte[] pbData = Convert.FromBase64String(strKey);
			return MemUtil.EncryptEcb(pbData, pbKey32, false);
		}
	}
}
