﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Linq;
using System.Threading.Tasks;
using System.Security.Cryptography;
using System.Diagnostics;

using LogSig.AppModules;
using LogSig.Crypto;
using LogSig.DM;
using LogSig.DM.DihedralGroup;
using LogSig.Interfaces;
using LogSig.Plugins;
using LogSig.Util;

namespace LogSig
{
	public enum OutputFmt
	{
		Plain = 0,
		TeX = 1
	}

	public static class Program
	{
		private static CommandLineArgs m_cmdLineArgs = null;

		private const string CmdOptVerbose = "verbose";
		public static bool Verbose
		{
			get { return (m_cmdLineArgs[CmdOptVerbose] != null); }
		}

		private const string CmdOptFactorize = "fct";
		public static bool FactorizeNumbers
		{
			get { return (m_cmdLineArgs[CmdOptFactorize] != null); }
		}

		private const string CmdOptOutFmt = "outfmt";
		private static OutputFmt? m_fmtOut = null;
		public static OutputFmt OutputFormat
		{
			get
			{
				if(m_fmtOut.HasValue) return m_fmtOut.Value;

				string strFmt = m_cmdLineArgs[CmdOptOutFmt];
				if(strFmt == null) m_fmtOut = OutputFmt.Plain;
				else if(strFmt.Equals("TeX", StrUtil.CaseIgnoreCmp))
					m_fmtOut = OutputFmt.TeX;
				else m_fmtOut = OutputFmt.Plain;

				return m_fmtOut.Value;
			}
		}

		private static CryptoRandomGen m_crg = null;
		public static CryptoRandomGen CryptoRandom
		{
			get
			{
				if(m_crg != null) return m_crg;

				MemoryStream ms = new MemoryStream();

				byte[] pb = MemUtil.Int64ToBytes(DateTime.UtcNow.ToBinary());
				ms.Write(pb, 0, pb.Length);

				pb = MemUtil.Int64ToBytes(Environment.TickCount);
				ms.Write(pb, 0, pb.Length);

				pb = Guid.NewGuid().ToByteArray();
				if(pb != null) ms.Write(pb, 0, pb.Length);
				else { Debug.Assert(false); }

				pb = new byte[32];
				using(RNGCryptoServiceProvider rngCsp = new RNGCryptoServiceProvider())
				{
					rngCsp.GetBytes(pb);
					ms.Write(pb, 0, pb.Length);
				}

				byte[] pbData = ms.ToArray();
				ms.Close();
				using(SHA512Managed sha512 = new SHA512Managed())
				{
					pb = sha512.ComputeHash(pbData);
				}

				byte[] pb32 = new byte[32];
				Array.Copy(pb, 0, pb32, 0, 32);

				byte[] pb16 = new byte[16];
				Array.Copy(pb, 32, pb16, 0, 16);

				var crgAes = new AesCryptoRandomGen(pb32, pb16);
				m_crg = crgAes;
				return crgAes;
			}
		}

		private static List<AppMod> m_vModules = new List<AppMod>();
		public static List<AppMod> AppModules
		{
			get { return m_vModules; }
		}

		private static DmGroupFactoryPool m_poolGrp = new DmGroupFactoryPool();
		public static DmGroupFactoryPool GroupFactories
		{
			get { return m_poolGrp; }
		}

		public static void Main(string[] args)
		{
#if DEBUG
			MainEx(args);
#else
			try { MainEx(args); }
			catch(Exception ex) { Console.WriteLine(ex.Message); }
#endif
		}

		private static void MainEx(string[] args)
		{
			m_cmdLineArgs = new CommandLineArgs(args);

			// Add default modules; can be extended by plugins
			m_vModules.Add(new AppModCore());
			m_vModules.Add(new AppModExp());
			m_vModules.Add(new AppModGen());
			m_vModules.Add(new AppModFac());
			m_vModules.Add(new AppModRepCases());
			m_vModules.Add(new AppModPgm());
			m_vModules.Add(new AppModMst1());

			LsPluginManager.LoadAllPlugins();

			AppModEventArgs e = new AppModEventArgs(m_cmdLineArgs);
			foreach(AppMod m in m_vModules)
			{
				foreach(KeyValuePair<string, EventHandler<AppModEventArgs>> kvp
					in m.CommandHandlers)
				{
					if(m_cmdLineArgs[kvp.Key] != null) kvp.Value(null, e);
				}
			}

			LsPluginManager.TerminateAllPlugins();
		}

		public static void TraceLine(string str)
		{
			Console.WriteLine(str);
		}

		public static void TraceLine<T>(DmBlockSeq<T> ls, string strPrefix,
			string strSuffix)
			where T : IDmGroupElement<T>
		{
			string str = ls.ToString();
			if(!string.IsNullOrEmpty(strPrefix)) str = strPrefix + str;
			if(!string.IsNullOrEmpty(strSuffix)) str += strSuffix;

			Console.WriteLine(str);
		}
	}
}
