﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Numerics;
using System.Diagnostics;

using LogSig.DM;
using LogSig.DM.DihedralGroup;
using LogSig.DM.WreathProductGroup;
using LogSig.DM.ZnGroup;
using LogSig.DM.MultiZnGroup;

namespace LogSig.Test
{
	public static class DmWreathProductGroupTests
	{
		public static void Run()
		{
			DmDihedralGroup grpDih1 = new DmDihedralGroup(10);
			DmDihedralGroup grpDih2 = new DmDihedralGroup(14);
			DmZnGroup grpZ11 = new DmZnGroup(11);
			DmMultiZnGroup grpMZ = new DmMultiZnGroup(new BigInteger[] { 13, 17 });

			RunTests(grpZ11, grpMZ);
			RunTests(grpMZ, grpZ11);
			RunTests(grpDih1, grpDih2);
			RunTests(grpDih1, grpZ11);
		}

		private static void RunTests<H, P>(IDmGroup<H> grpH, IDmGroup<P> grpP)
			where H : IDmGroupElement<H>
			where P : IDmGroupElement<P>
		{
			var grp = new DmWreathProductGroup<H, P>(grpH, grpP);

			bool bBigIntSupported = false;
			try
			{
				H tH = grpH.GenerateRandomElement(Program.CryptoRandom);
				P tP = grpP.GenerateRandomElement(Program.CryptoRandom);
				BigInteger biH = grpH.ElementToBigInt(tH);
				BigInteger biP = grpP.ElementToBigInt(tP);
				H tH2 = grpH.BigIntToElement(biH);
				P tP2 = grpP.BigIntToElement(biP);
				if(tH.Equals(tH2) && tP.Equals(tP2))
					bBigIntSupported = true;
				else { Debug.Assert(false); }
			}
			catch(Exception) { } // BigInt methods might be unsupported

			for(int r = 0; r < 100; ++r)
			{
				var a = grp.GenerateRandomElement(Program.CryptoRandom);
				var b = grp.GenerateRandomElement(Program.CryptoRandom);
				var c = grp.GenerateRandomElement(Program.CryptoRandom);

				if(!grp.Multiply(grp.Multiply(a, b), c).Equals(
					grp.Multiply(a, grp.Multiply(b, c))))
					throw new Exception("WP-Assoc");

				if(!grp.Multiply(a, grp.Identity).Equals(a) ||
					!grp.Multiply(grp.Identity, a).Equals(a))
					throw new Exception("WP-Id");

				var ai = grp.Invert(a);
				if(!grp.Multiply(a, ai).Equals(grp.Identity) ||
					!grp.Multiply(ai, a).Equals(grp.Identity))
					throw new Exception("WP-Inv");

				var bi = grp.Invert(b);
				if(!grp.Invert(grp.Multiply(a, b)).Equals(grp.Multiply(bi, ai)))
					throw new Exception("WP-InvPart");

				if(!a.Equals(grp.Identity) && !b.Equals(grp.Identity))
				{
					if(grp.Multiply(a, b).Equals(a) ||
						grp.Multiply(a, b).Equals(b))
						throw new Exception("WP-NonIdMul");
				}

				string strA = a.ToString();
				var ap = grp.ParseElement(strA);
				if((ap == null) || !a.Equals(ap))
					throw new Exception("WP-Str");

				if(bBigIntSupported)
				{
					BigInteger biA = grp.ElementToBigInt(a);
					var an = grp.BigIntToElement(biA);
					if((an == null) || !a.Equals(an))
						throw new Exception("WP-BigInt");
				}
			}
		}
	}
}
