﻿// LogSig - Logarithmic Signature Utility
// Copyright (C) 2011-2015 Dominik Reichl <info@dominik-reichl.de>

using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Security.Cryptography;
using System.Diagnostics;

namespace LogSig.Util
{
	public static class MemUtil
	{
		/// <summary>
		/// Convert 4 bytes to a 32-bit signed integer (little-endian).
		/// </summary>
		public static int BytesToInt32(byte[] pb)
		{
			Debug.Assert((pb != null) && (pb.Length == 4));
			if(pb == null) throw new ArgumentNullException("pb");
			if(pb.Length != 4) throw new ArgumentException();

			return ((int)pb[0] | ((int)pb[1] << 8) | ((int)pb[2] << 16) |
				((int)pb[3] << 24));
		}

		/// <summary>
		/// Convert 4 bytes to a 32-bit unsigned integer (little-endian).
		/// </summary>
		public static uint BytesToUInt32(byte[] pb)
		{
			Debug.Assert((pb != null) && (pb.Length == 4));
			if(pb == null) throw new ArgumentNullException("pb");
			if(pb.Length != 4) throw new ArgumentException();

			return ((uint)pb[0] | ((uint)pb[1] << 8) | ((uint)pb[2] << 16) |
				((uint)pb[3] << 24));
		}

		/// <summary>
		/// Convert 8 bytes to a 64-bit signed integer (little-endian).
		/// </summary>
		public static long BytesToInt64(byte[] pb)
		{
			Debug.Assert((pb != null) && (pb.Length == 8));
			if(pb == null) throw new ArgumentNullException("pb");
			if(pb.Length != 8) throw new ArgumentException();

			return ((long)pb[0] | ((long)pb[1] << 8) | ((long)pb[2] << 16) |
				((long)pb[3] << 24) | ((long)pb[4] << 32) | ((long)pb[5] << 40) |
				((long)pb[6] << 48) | ((long)pb[7] << 56));
		}

		/// <summary>
		/// Convert 8 bytes to a 64-bit unsigned integer (little-endian).
		/// </summary>
		public static ulong BytesToUInt64(byte[] pb)
		{
			Debug.Assert((pb != null) && (pb.Length == 8));
			if(pb == null) throw new ArgumentNullException("pb");
			if(pb.Length != 8) throw new ArgumentException();

			return ((ulong)pb[0] | ((ulong)pb[1] << 8) | ((ulong)pb[2] << 16) |
				((ulong)pb[3] << 24) | ((ulong)pb[4] << 32) | ((ulong)pb[5] << 40) |
				((ulong)pb[6] << 48) | ((ulong)pb[7] << 56));
		}

		/// <summary>
		/// Convert 8 bytes to a 64-bit unsigned integer (little-endian).
		/// </summary>
		public static ulong BytesToUInt64(byte[] pb, int iOffset)
		{
			if(iOffset < 0) throw new ArgumentOutOfRangeException("iOffset");
			Debug.Assert((pb != null) && (pb.Length >= (iOffset + 8)));
			if(pb == null) throw new ArgumentNullException("pb");
			if(pb.Length < (iOffset + 8)) throw new ArgumentException();

			return ((ulong)pb[iOffset] | ((ulong)pb[iOffset + 1] << 8) |
				((ulong)pb[iOffset + 2] << 16) | ((ulong)pb[iOffset + 3] << 24) |
				((ulong)pb[iOffset + 4] << 32) | ((ulong)pb[iOffset + 5] << 40) |
				((ulong)pb[iOffset + 6] << 48) | ((ulong)pb[iOffset + 7] << 56));
		}

		/// <summary>
		/// Convert a 64-bit signed integer to 8 bytes (little-endian).
		/// </summary>
		public static byte[] Int64ToBytes(long nValue)
		{
			byte[] pb = new byte[8];

			unchecked
			{
				pb[0] = (byte)nValue;
				pb[1] = (byte)(nValue >> 8);
				pb[2] = (byte)(nValue >> 16);
				pb[3] = (byte)(nValue >> 24);
				pb[4] = (byte)(nValue >> 32);
				pb[5] = (byte)(nValue >> 40);
				pb[6] = (byte)(nValue >> 48);
				pb[7] = (byte)(nValue >> 56);
			}

			return pb;
		}

		public static void CopyStream(Stream sSource, Stream sTarget)
		{
			Debug.Assert((sSource != null) && (sTarget != null));
			if(sSource == null) throw new ArgumentNullException("sSource");
			if(sTarget == null) throw new ArgumentNullException("sTarget");

			const int nBufSize = 4096;
			byte[] pbBuf = new byte[nBufSize];
			int nRead;

			while((nRead = sSource.Read(pbBuf, 0, nBufSize)) > 0)
				sTarget.Write(pbBuf, 0, nRead);
		}

		public static byte[] EncryptEcb(byte[] pbBuf, byte[] pbKey32, bool bEncrypt)
		{
			if(pbBuf == null) throw new ArgumentNullException("pbBuf");
			if(pbKey32 == null) throw new ArgumentNullException("pbKey32");
			if(pbKey32.Length != 32) throw new ArgumentOutOfRangeException("pbKey32");

			RijndaelManaged r = new RijndaelManaged();
			r.KeySize = 256;
			r.Key = pbKey32;
			r.Mode = CipherMode.ECB;
			r.Padding = PaddingMode.None;

			if(((pbBuf.LongLength * 8) % (long)r.BlockSize) != 0)
				throw new ArgumentOutOfRangeException("pbBuf");

			ICryptoTransform t = (bEncrypt ? r.CreateEncryptor() : r.CreateDecryptor());
			Debug.Assert(t.CanTransformMultipleBlocks);

			if(bEncrypt)
			{
				byte[] pbEnc = null;
				using(MemoryStream msEnc = new MemoryStream())
				{
					using(CryptoStream csEnc = new CryptoStream(msEnc, t, CryptoStreamMode.Write))
					{
						csEnc.Write(pbBuf, 0, pbBuf.Length);
					}

					pbEnc = msEnc.ToArray();
				}

				Debug.Assert((pbEnc != null) && (pbEnc.Length == pbBuf.Length));
				return pbEnc;
			}

			// Decrypt
			byte[] pbDec = new byte[pbBuf.Length];
			using(MemoryStream msBuf = new MemoryStream(pbBuf, false))
			{
				using(CryptoStream csDec = new CryptoStream(msBuf, t, CryptoStreamMode.Read))
				{
					csDec.Read(pbDec, 0, pbDec.Length);
				}
			}

			return pbDec;
		}

		public static Stream EncryptCbc(Stream s, byte[] pbKey32, byte[] pbIV16,
			bool bEncrypt)
		{
			if(s == null) throw new ArgumentNullException("s");
			if(pbKey32 == null) throw new ArgumentNullException("pbKey32");
			if(pbKey32.Length != 32) throw new ArgumentOutOfRangeException("pbKey32");
			if(pbIV16 == null) throw new ArgumentNullException("pbIV16");
			if(pbIV16.Length != 16) throw new ArgumentOutOfRangeException("pbIV16");

			RijndaelManaged r = new RijndaelManaged();
			r.KeySize = 256;
			r.Key = pbKey32;
			r.Mode = CipherMode.CBC;
			r.IV = pbIV16;
			r.Padding = PaddingMode.PKCS7;

			ICryptoTransform t = (bEncrypt ? r.CreateEncryptor() : r.CreateDecryptor());
			return new CryptoStream(s, t, (bEncrypt ? CryptoStreamMode.Write :
				CryptoStreamMode.Read));
		}

		/* public static void EncryptCbc(string strFileIn, string strFileOut,
			byte[] pbKey32, byte[] pbIV16, bool bEncrypt)
		{
			if(pbKey32 == null) throw new ArgumentNullException("pbKey32");
			if(pbKey32.Length != 32) throw new ArgumentOutOfRangeException("pbKey32");
			if(pbIV16 == null) throw new ArgumentNullException("pbIV16");
			if(pbIV16.Length != 16) throw new ArgumentOutOfRangeException("pbIV16");

			FileStream fsIn = new FileStream(strFileIn, FileMode.Open, FileAccess.Read,
				FileShare.Read);
			FileStream fsOut = new FileStream(strFileOut, FileMode.Create, FileAccess.Write,
				FileShare.None);

			if(bEncrypt)
			{
				fsOut.Write(pbIV16, 0, pbIV16.Length);

				Stream sEnc = EncryptCbc(fsOut, pbKey32, pbIV16, true);
				MemUtil.CopyStream(fsIn, sEnc);
				sEnc.Close();
			}
			else
			{
				fsIn.Read(pbIV16, 0, pbIV16.Length);

				Stream sDec = EncryptCbc(fsIn, pbKey32, pbIV16, false);
				MemUtil.CopyStream(sDec, fsOut);
				sDec.Close();
			}

			fsIn.Close();
			fsOut.Close();
		} */
	}
}
